/*============================================================================*/
/*                        L a b W i n d o w s / C V I                         */
/*----------------------------------------------------------------------------*/
/*    Copyright (c) National Instruments 1987-1996.  All Rights Reserved.     */
/*----------------------------------------------------------------------------*/
/*                                                                            */
/* Title:       toolbox.c                                                     */
/* Purpose:     provides commonly useful functions                            */
/*                                                                            */
/*============================================================================*/

#include <ansi_c.h>
#include <userint.h>
#include <utility.h>
#include <formatio.h>
#include <easyio.h>
#include "toolbox.h"

typedef struct ListStructTag
    {
    int signature;              /* debugging aid */
    int percentIncrease;        /* %of current size to increase by when list is out of space */
    int minNumItemsIncrease;    /* fixed number of items to increase by when list is out of space */
    int listSize;               /* number of items than can fit in the currently allocated memory */
    int itemSize;               /* the size of each item in the list (same for every item) */
    int numItems;               /* number of items currently in the list */
    unsigned char itemList[1];  /* resizable array of list elements */
    } ListStruct;

#define kDefaultAllocationPercentIncrease   10  /* increase list size by 10% every time it is full */
#define kDefaultAllocationminNumItemsIncrease   4   /* always increase list size by 4 items when it is full */

     /* how many items to expand the list by when it becomes full = current listSize (in items) + (hiword percent of list size) +   loword */
#define NUMITEMSPERALLOC(list) Max(((*list)->listSize * ((*list)->percentIncrease + 100)) / 100 , (*list)->minNumItemsIncrease)

#define ITEMPTR(list, item)     &(((char *)(void *)&(*list)->itemList)[ (*(list))->itemSize * (item)])

#define LIST_SIGNATURE          CAT4CHARS('L', 'I', 'S', 'T');

#define MOUSE_EVENTS_SIGNATURE  CAT4CHARS('M', 'E', 'V', 'T');

typedef struct
    {
    void            *ptr;
    unsigned int    size;
    } HandleRecord;


typedef struct
    {
    int             signature;                  /* debug help, makes this structure identifiable in Dynamic Memory Display */
    char            *type;                      /* copy of type string passed to ChainCtrlCallback(), example "DAQChart", or "DAQNumeric", or "Button With Right Click Help", etc... */
    void            *callbackData;              /* callback data passed in to ChainCtrlCallback */
    CtrlCallbackPtr callback;                   /* callback function passed in to ChainCtrlCallback */
    void            *previousCallbackData;      /* if previously linked, this a ptr to the previous link structure */
    CtrlCallbackPtr previousCallback;           /* if previously linked, this is LinkedCtrlCallback() */
    }   *CtrlCallbackChainLink;

typedef struct
    {
    int             signature;                  /* debug help, makes this structure identifiable in Dynamic Memory Display */
    char            *type;                      /* copy of type string passed to ChainPanelCallback() */
    void            *callbackData;              /* callback data passed in to ChainPanelCallback */
    PanelCallbackPtr callback;                  /* callback function passed in to ChainPanelCallback */
    void            *previousCallbackData;      /* if previously linked, this a ptr to the previous link structure */
    PanelCallbackPtr previousCallback;          /* if previously linked, this is LinkedPanelCallback() */
    }   *PanelCallbackChainLink;


    /*  For adding extra mouse events to a control (see EnableExtendedMouseEvents) */
typedef struct
    {
    int     signature;          /*  debug help, makes this structure identifiable in Dynamic Memory Display */
    int     panel;              /*  owning panel of the ctrl receiving extended mouse events */
    int     ctrl;               /*  the control receiving extended mouse events */
    int     rightButtonDown;    /*  was the right button down the last time we checked? */
    int     leftButtonDown;     /*  was the left button down the last time we checked? */ 
    int     dispose;            /*  set to TRUE when the control is discarded, causes this struct to be discarded on the next delayed callback */
    int     lastX, lastY;       /*  mouse position the last time we checked */
    double  minPeriod;          /*  min time between mouse events */
    int     enabled;            /*  TRUE if extended mouse events are currently enabled for the control */
    } *MouseEventsInfo;
    

 /********************/
 /* Static Functions */
 /********************/

static void     CVIFUNC SetListMemBlockCompareSize(unsigned int size);
static int      CVICALLBACK ListMemBlockCompare(void *a, void *b);
static void     CVIFUNC SetDefaultCompareFunctionIfNeeded(ListType list, CompareFunction *compareFunction);

static void     CVIFUNC FilterDownToCorrectHeapPosition(void *array, int numElements, int elementSize, CompareFunction compareFunction);
static void     CVIFUNC FilterUpToCorrectHeapLocation(void *array, int index, int elementSize, CompareFunction compareFunction);
static void     CVIFUNC BuildHeap(void *array, int numElements, int elementSize, CompareFunction compareFunction);

static int      CVIFUNC SetIntAttributeForCtrlsInCtrlList(int panel, int attribute, int value, int numCtrls, ListType ctrlList);
static int      CVIFUNC IsTBBoundsAttr(int attribute);

static void     CVIFUNC InterleavePartialLastScan(void *list, int itemSize, int numChannels, int numScans, int numItems);
static void     CVIFUNC TransposeDataInPlace(void *list, int itemSize, int numPoints, int numChannels);
static int      CVIFUNC ConvertNumStrToUint (char *str, unsigned long *n, int *negative);

static int      CVICALLBACK ProgressCancelCallback(int panel, int ctrl, int event, void *callbackData, int eventData1, int eventData2);
static int      CVIFUNC InternationalTimeString(time_t rawTime, char timeString[], int bufferSize);
static int      CVIFUNC InternationalDateString(time_t rawTime, char dateString[], int bufferSize);

static void     CVIANSI ToolBoxAtExitFunction(void);
static double   CVIFUNC GetRand(void);

static int      CVICALLBACK LinkedCtrlCallback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2);
static int      CVIFUNC FindCtrlCallbackLink(int panel, int ctrl, char *type,  CtrlCallbackChainLink *link);

static void     CVIFUNC InstallObjectDeletionProcedure(void);
static void     CVIANSI ObjectDeletionProcedure(void);

static int      CVICALLBACK LinkedPanelCallback(int panel, int event, void *callbackData, int eventData1, int eventData2);
static int      CVIFUNC FindPanelCallbackLink(int panel, char *type,  PanelCallbackChainLink *link);

static int      CVIFUNC LastCtrl(int panel);
static int      CVIFUNC NextCtrl(int panel, int ctrl, int wrap);

static void     CVIFUNC MouseEvents_Check(MouseEventsInfo info);
static int      CVICALLBACK MouseEventsCallback (int panel, int control, int event, void *callbackData, int eventData1, int eventData2);
static void     CVICALLBACK MouseEvents_CheckCallback(void *callbackData);
static char *   CVIFUNC ToolBox_GetDAQErrorString (short int error);
static char *   CVIFUNC ToolBox_InternalGetDAQErrorString (short int error);

/********************************************************************/
#pragma warning(disable: 4761)
    /*  floating point comparison with a tolerance to compensate for the inability to exactly represent some numbers */
    /*  returns -1 if a < b, 0 if a == b, and 1 if a > b */
int CVIFUNC FP_Compare(double a, double b)
{
    if (FP_EQ(a,b))
        return 0;
    else
    if (FP_GT(a,b))
        return 1;
    else
        return -1;
}

/********************************************************************/

    /*  floating point comparison with a tolerance to compensate for the inability to exactly represent some numbers */
int CVIFUNC FP_EQ(double a, double b)       /* is a == b    */
{
    double tmp;

    if(b == 0)
        tmp= a;
    else
        tmp= a/b - 1;

    if(tmp < FP_CompareEpsilon && tmp > -FP_CompareEpsilon)
        if (ABS_VAL(a-b) < FP_CompareEpsilon)
            return TRUE;

    return FALSE;
}

/********************************************************************/

    /*  floating point comparison with a tolerance to compensate for the inability to exactly represent some numbers */
int CVIFUNC FP_GT(double a, double b)       /* is a > b */
{
    double tmp;
    int reply;

    if(b == 0)
        reply= (a > 0);
    else if(a == 0)
        reply= (b < 0);
    else
        {
        tmp= a/b;
        if(tmp > 0 && (ABS_VAL(a-b) <= FP_CompareEpsilon))                      /* both neg, or both pos */
            {
            if(b > 0)
                reply= (tmp - 1 >= FP_CompareEpsilon);
            else
                reply= (tmp - 1 <= -FP_CompareEpsilon);
            }
        else
            reply= a > b;
        }

    return reply;
}

/********************************************************************/

    /*  floating point comparison with a tolerance to compensate for the inability to exactly represent some numbers */
int CVIFUNC FP_GE(double a, double b)       /* is a >= b    */
{
    double tmp;
    int reply;

    if(b == 0)
        reply= (a >= 0);
    else if(a == 0)
        reply= (b <= 0);
    else
        {
        tmp= a/b;
        if(tmp > 0 && (ABS_VAL(a-b) <= FP_CompareEpsilon))                      /* both neg, or both pos */
            {
            if(b > 0)
                reply= (tmp - 1 > -FP_CompareEpsilon);
            else
                reply= (tmp - 1 < FP_CompareEpsilon);
            }
        else
            reply= a > b;
        }

    return reply;
}

/********************************************************************/

    /*  floating point comparison with a tolerance to compensate for the inability to exactly represent some numbers */
int CVIFUNC FP_LT(double a, double b)       /* is a < b */
{
    return FP_GT(b, a);
}

/********************************************************************/

    /*  floating point comparison with a tolerance to compensate for the inability to exactly represent some numbers */
int CVIFUNC FP_LE(double a, double b)       /* is a <= b    */
{
    return FP_GE(b, a);
}

/********************************************************************/

double CVIFUNC Pin(double value, double low, double high)
{
    if (value < low)
        return low;
    else
    if (value > high)
        return high;
    else
        return value;
}

/********************************************************************/

static unsigned gRandomSeed = 0x38589391;
static int      gReSeed = TRUE;

double CVIFUNC Random(double minimum, double maximum)
{
    double  range = maximum - minimum;

    if (range < 0)
        return minimum;
    else
        return minimum + GetRand() * range;
}

/*******************************************************/

void CVIFUNC SetRandomSeed(unsigned long seed)
{
    if (seed)
        gRandomSeed = seed;
    else
    do
        {
        gRandomSeed = clock();
        } while (!gRandomSeed);     /* make sure we don't somehow get zero, since that would cause an infinite loop */

    gReSeed = TRUE;
}

/*******************************************************/

static double CVIFUNC GetRand(void)
{
    double          randNum;
    static double   numJ, numK, numH;
    static long     seedX, seedY, seedZ;

    if (gReSeed)
        {
        gReSeed = FALSE;
        seedX = 0;
        }

    while(!seedX)
        {
        seedX = gRandomSeed & 0x3FFF;
        seedY = (seedX * 8191) & 0x3FFF;
        seedZ = (seedY * 8191) & 0x3FFF;
        numJ = 1.0 / 30269.0;
        numK = 1.0 / 30307.0;
        numH = 1.0 / 30323.0;
        }

    seedX = (171 * seedX) % 30269;
    seedY = (172 * seedY) % 30307;
    seedZ = (170 * seedZ) % 30323;

    randNum = seedX * numJ + seedY * numK + seedZ * numH;
    randNum -= floor(randNum);
    if(randNum < 0) randNum += 1.0; /* if floor does a Round To Nearest */
    return randNum;
}

/********************************************************************/

void CVIFUNC ConvertArrayType(void * sourceArray, int sourceDataType, void * targetArray,
                    int targetDataType, int numberOfPoints)
{
    int     i;
    double  temp;

    for (i=0; i < numberOfPoints; i++)  /* watcom will optimize out the indexing overhead */
        {
        switch (sourceDataType)
            {
            case VAL_CHAR:
                temp = ((char *)sourceArray)[i];
                break;
            case VAL_INTEGER:
                temp = ((int *)sourceArray)[i];
                break;
            case VAL_SHORT_INTEGER:
                temp = ((short int *)sourceArray)[i];
                break;
            case VAL_FLOAT:
                temp = ((float *)sourceArray)[i];
                break;
            case VAL_DOUBLE:
                temp = ((double *)sourceArray)[i];
                break;
            case VAL_UNSIGNED_SHORT_INTEGER:
                temp = ((unsigned short *)sourceArray)[i];
                break;
            case VAL_UNSIGNED_INTEGER:
                temp = ((unsigned int *)sourceArray)[i];
                break;
            case VAL_UNSIGNED_CHAR:
                temp = ((unsigned char *)sourceArray)[i];
                break;
            }

        switch (targetDataType)
            {
            case VAL_CHAR:
                ((char *)targetArray)[i] = (char)temp;
                break;
            case VAL_INTEGER:
                ((int *)targetArray)[i] = (int)temp;
                break;
            case VAL_SHORT_INTEGER:
                ((short *)targetArray)[i] = (short)temp;
                break;
            case VAL_FLOAT:
                ((float *)targetArray)[i] = (float)temp;
                break;
            case VAL_DOUBLE:
                ((double *)targetArray)[i] = temp;
                break;
            case VAL_UNSIGNED_SHORT_INTEGER:
                ((unsigned short *)targetArray)[i] = (unsigned short)temp;
                break;
            case VAL_UNSIGNED_INTEGER:
                ((unsigned int *)targetArray)[i] = (unsigned int)temp;
                break;
            case VAL_UNSIGNED_CHAR:
                ((unsigned char *)targetArray)[i] = (unsigned char)temp;
                break;
            }
        }
}

/********************************************************************/

void CVIFUNC DoAssert(int passed, char *fileName, int line, char *msg)
{
    extern void  CVIFUNC_C _UPLibReportError(long, long, ...);

    if (!passed)
        {
        char buf[256];

            sprintf(buf, "Assert!\n\n%s\n\nThe error was detected on line %d of file '%s'.\n\nProgram aborting.", msg ? msg : "Programming error detected.", line, fileName);

            MessagePopup("Assert", buf);
            abort();
        }
}

/********************************************************************/

    /*  Binary search numElements of size elementSize in array for a match
        to the. item. Return the index of the element that matches (0 - numElements - 1).
        If no match is found return the -i-1 where i is the index (0 - numElements)
        where the item should be placed. (*theCmp)(a,b) should return  <0 if a<b,
        0 if a==b, >0 if a>b.

        This function is like the C-Library function bsearch() except that this
        function returns the index where the item should be placed if it is not
        found.
    */
int CVIFUNC BinSearch(void *array, int numElements, int elementSize, void *itemPtr, CompareFunction compareFunction)
{
    int low, high, mid, cmp;

    for(low=0, high= numElements-1, mid=0, cmp= -1; low <= high; )
        {
        mid= (low + high) >> 1;
        if((cmp= (*compareFunction)(itemPtr, (char *)array + mid*elementSize)) == 0)
            return mid;
        if(cmp < 0)
            high= mid - 1;
        else
            low= mid + 1;
        }
    if (cmp > 0)
        mid++;

    return -mid-1;
}

/********************************************************************/

    /*  Swap numBytes between block1 and block2.
    */
void CVIFUNC SwapBlock(void *block1, void *block2, int numBytes)
{
    if(((int)block1 | (int)block2 | numBytes) & 3)
        {
        char *p, *q;
        char x;
        for(p= (char *)block1, q= (char *)block2; --numBytes >= 0; p++, q++){ x= *p; *p= *q; *q= x; }
        }
    else
        {
        long int    *p, *q, x;
        for(p= (long int*)block1, q= (long int*)block2, numBytes >>= 2; --numBytes >= 0; p++, q++){ x= *p; *p= *q; *q= x; }
        }
}

/********************************************************************/

    /*  returns negative UIL error,  0 if time conversion/formatting problem, or 1 if successful */
int CVIFUNC GetFileCLibTime(char filePath[], time_t *time)
{
    int         error = 0;
    int         month;
    int         day;
    int         year;
    int         hours;
    int         minutes;
    int         seconds;
    struct tm   fileTM;
    int         savedBole = SetBOLE(FALSE); /* don't break on GetFileDate() errors */

    switch (GetFileDate (filePath, &month, &day, &year))
        {
        case 0:
            break;
        case -1:
            errChk( UIEFileWasNotFound);
            break;
        case -4:
            errChk( UIEOutOfMemory);
            break;
        case -3:
        default:
            errChk( UIEIOError);
            break;
        }

    switch (GetFileTime (filePath, &hours, &minutes, &seconds))
        {
        case 0:
            break;
        case -1:
            errChk( UIEFileWasNotFound);
            break;
        case -4:
            errChk( UIEOutOfMemory);
            break;
        case -3:
        default:
            errChk( UIEIOError);
            break;
        }

    fileTM.tm_sec = seconds;
    fileTM.tm_min = minutes;
    fileTM.tm_hour = hours;
    fileTM.tm_mday = day;
    fileTM.tm_mon = month - 1;
    fileTM.tm_year = year - 1900;
    fileTM.tm_isdst = -1;       /* -1 means "I don't know, you tell me..." */

    *time = mktime(&fileTM);

    return *time != (time_t)-1;

Error:
    SetBOLE(savedBole);
    return error;
}

/********************************************************************/

    /*  returns negative UIL error,  # of characters formatted if success, or 0 if time conversion/formatting problem */
int CVIFUNC InternationalFileTime(char filePath[], char timeString[], int bufferSize)
{
    time_t  time;
    int     error = 0;
    int     success;

    errChk( success = GetFileCLibTime(filePath, &time));

    if (success)
        return InternationalTimeString(time, timeString, bufferSize);
    else
        return 0;

Error:
    return error;
}

/********************************************************************/

    /*  returns negative UIL error, # of characters formatted if success, or 0 if time conversion/formatting problem */
int CVIFUNC InternationalFileDate(char filePath[], char dateString[], int bufferSize)
{
    time_t  time;
    int     error = 0;
    int     success;

    errChk( success = GetFileCLibTime(filePath, &time));

    if (success)
        return InternationalDateString(time, dateString, bufferSize);
    else
        return 0;

Error:
    return error;
}

/********************************************************************/

    /*  returns # of characters formatted if success, or 0 if time conversion/formatting problem */
int CVIFUNC InternationalTime(char timeString[], int bufferSize)
{
    time_t      rawTime;

    time (&rawTime);
    return InternationalTimeString(rawTime, timeString, bufferSize);
}

/********************************************************************/

    /*  returns # of characters formatted if success, or 0 if time conversion/formatting problem */
int CVIFUNC InternationalDate(char dateString[], int bufferSize)
{
    time_t      rawTime;

    time (&rawTime);
    return InternationalDateString(rawTime, dateString, bufferSize);
}

/********************************************************************/

static int CVIFUNC InternationalTimeString(time_t rawTime, char timeString[], int bufferSize)
{
    char        *oldLocale;
    struct tm   *localTime;
    int         result;

    oldLocale = setlocale (LC_ALL, NULL);
    setlocale (LC_ALL, "");

    localTime = localtime (&rawTime);
    result = strftime (timeString, bufferSize, "%X", localTime); 

    setlocale (LC_ALL, oldLocale);
    return result;
}

/********************************************************************/

static int CVIFUNC InternationalDateString(time_t rawTime, char dateString[], int bufferSize)
{
    char        *oldLocale;
    struct tm   *localTime;
    int         result;

    oldLocale = setlocale (LC_ALL, NULL);
    setlocale (LC_ALL, "");

    localTime = localtime (&rawTime);
    result = strftime (dateString, bufferSize, "%x", localTime);

    setlocale (LC_ALL, oldLocale);
    return result;
}

/********************************************************************/

char * CVIFUNC StrDup(char *stringToDuplicate)
{
    char    *duplicatedString = 0;
    int     length;

    if (stringToDuplicate)
        {
        length = strlen(stringToDuplicate);
        duplicatedString = (char *)malloc(length+1);
        if (duplicatedString)
            strcpy(duplicatedString, stringToDuplicate);
        }

    return duplicatedString;
}

/**************************************************************************/

    /*  This function adds stringToAdd to the end of *string. */
int CVIFUNC AppendString(char **string, char *stringToAdd, int lengthToAdd)
{
    int     currLen;
    char    *newString;

    AssertMsg(string, "NULL passed to first parameter of AppendString()");

    if (!stringToAdd)
        return TRUE;

    if (lengthToAdd < 0)
        lengthToAdd = strlen(stringToAdd);
    else
        {
        int index;
                    /* If the stringToAdd contains a terminator, we have to stop there */
        for (index = 0; index < lengthToAdd && stringToAdd[index] != 0; index++);
        lengthToAdd = index;
        }

    if (lengthToAdd == 0)
        return TRUE;

    if (!*string)
        {
        newString = (char *)malloc(lengthToAdd + 1);
        currLen = 0;
        }
    else
        {
        currLen = strlen(*string);
        newString = realloc(*string, currLen + lengthToAdd + 1);
        }

    if (newString)
        {
        memmove(newString + currLen, stringToAdd, lengthToAdd);
        newString[currLen + lengthToAdd] = 0;
        *string = newString;
        return TRUE;
        }
    else
        return FALSE;
}

/*****************************************************************************************/

/*
   Creates a dynamically allocated copy of a string without
   any leading or trailing white space.
*/
char * CVIFUNC StrDupWithoutSurrWhiteSpace(char *string)
{
    char    *start, *copy;
    int   len;

    start = SkipWhiteSpace(string);

    len = strlen(start);
    while ((len > 0) && isspace(start[len-1]))   /* look backwards for last non white space */
        len--;

    copy = malloc (len+1);
    if (copy)
        {
        memmove (copy, start, len);
        copy[len] = '\0';
        }

    return copy;
}

/**************************************************************************/

char * CVIFUNC SkipWhiteSpace(char *string)
{
    char    *p=string;

    if (p)
        while (*p && isspace(*p))
            p++;
    return p;
}

/**************************************************************************/

char * CVIFUNC SkipNonWhiteSpace(char *string)
{
    char    *p=string;

    if (p)
        while (*p && !isspace(*p))
            p++;
    return p;
}

/**************************************************************************/

int CVIFUNC HasNonWhiteSpace (char *s)
{
    if (s == NULL)
        return FALSE;
    else
        return *SkipWhiteSpace(s) != '\0';
}

/**************************************************************************/

void CVIFUNC StringCopyMax(char *dest, char *source, int destBufferSize)
{
    int destSize, sourceSize;

    AssertMsg(dest, "Null string passed to first parameter of StringCopyMax");
    if (source && destBufferSize > 0)
        {
        sourceSize = strlen(source) + 1;
        destSize = Min(sourceSize, destBufferSize);
        memmove(dest, source, Min(destBufferSize, destSize));
        if (destSize < sourceSize)
            dest[destSize - 1] = 0;     /* add null terminator since we didn't copy it */
        }
}

/***********************************************************************************/

char * CVIFUNC StringInsert(char *destinationString, char *stringToInsert, int destinationBufferSize)
{
    int destinationLength, lengthOfStringToInsert;

 /* insert stringToInsert before destinationString in destinationString but don't exceed destinationBufferSize in destinationString
    if destinationString is a pointer in a larger string then destinationBufferSize applies only to the section pointed to by destinationString */

    AssertMsg(stringToInsert, "NULL string passed to first parameter of StringInsert");
    AssertMsg(destinationString, "NULL string passed to second parameter of StringInsert");

    destinationLength = strlen(destinationString);
    lengthOfStringToInsert = strlen(stringToInsert);

    if ((destinationLength + lengthOfStringToInsert) > destinationBufferSize)
        lengthOfStringToInsert = destinationBufferSize - destinationLength;

    if (lengthOfStringToInsert > 0)
        {
        memmove(destinationString + lengthOfStringToInsert, destinationString, destinationLength+1);    /* slide the original string forward, +1 to keep NULL byte */
        memmove(destinationString, stringToInsert, lengthOfStringToInsert);         /* copy new string into the new space in front */
        }

    return destinationString+lengthOfStringToInsert;
}

/**************************************************************************/

int CVIFUNC StrICmp (char *s1, char *s2)
{
    int ch1, ch2;

    if (s1 == NULL)
        if (s2 == NULL)
            return 0;
        else
            return -1;
    else
        if (s2 == NULL)
            return 1;
        else
            for (;; s1++,s2++)
                {
                ch1 = toupper(*s1);
                ch2 = toupper(*s2);
                if (ch1 == 0)
                    if (ch2 == 0)
                        return 0;
                    else
                        return -1;
                else
                    if (ch2 == 0)
                        return 1;
                    else
                        if (ch1 < ch2)
                            return -1;
                        else if (ch1 > ch2)
                            return 1;
                }
}

/**************************************************************************/

void CVIFUNC RemoveSurroundingWhiteSpace(char string[])
{
    char *start, *end;
    int lengthWithoutLeadingWS;

    if (string)
        {
        start = SkipWhiteSpace(string);
        lengthWithoutLeadingWS = strlen(start);
        memmove (string, start, lengthWithoutLeadingWS);
        end = string + lengthWithoutLeadingWS;
        while (end > string && isspace(*(end - 1)))
            end--;

        *end = 0;
        }
}


/**************************************************************************/

int CVIFUNC StrICmpWithoutSurrWhiteSpace (char *s1, char *s2)
{
    int ch1, ch2;

    s1 = SkipWhiteSpace(s1);
    s2 = SkipWhiteSpace(s2);

    if (s1 == NULL)
        if (s2 == NULL)
            return 0;
        else
            return -1;
    else
        if (s2 == NULL)
            return 1;
        else
            for (;; s1++,s2++)
                {
                ch1 = toupper(*s1);
                ch2 = toupper(*s2);

                if (ch1 == 0)
                    {
                    if (*SkipWhiteSpace(s2) == 0)   /* check for trailing whitespace */
                        return 0;
                    else
                        return -1;
                    }
                else
                    {
                    if (ch2 == 0)
                        {
                        if (*SkipWhiteSpace(s1) == 0)  /* check for trailing whitespace */
                            return 0;
                        else
                            return 1;
                        }
                    else
                        {
                        if (ch1 < ch2)
                            return -1;
                        else if (ch1 > ch2)
                            return 1;
                        }
                    }
                }
}

/****************************************************************************/

    /*  This function splices list2 into list1 and stores the result in the newly
        allocated string splicedList.  All lists are strings of characters with commas separating
        the items.  The resulting splicedList will have as many items as the longest list.  Each item
        in the result splicedList will be the same as the corresponding item in list1 unless list1
        does not have a corresponding item.  In that case, list2's corresponding item is used.

        Example:
        char    *list1 = "Time,,Pressure, Temperature";
        char    *list2 = "Time, Column 1, Column 2, Column 3, Column 4, Column 5";
        char    *splicedList = 0;

        SpliceCommaSeparatedList(list1, list2, &splicedList);
        printf("Spliced list = %s\n", splicedList);

        This prints: Spliced list = Time, Column 1,Pressure, Temperature, Column 4, Column 5

        This function is used to fill in any missing items in a partially specified list with
        default items.

        The function returns 0 if successful, or -12 (UIEOutOfMemory) if there is not
        enough memory to create the splicedList.
    */
int CVIFUNC SpliceCommaSeparatedList(char *list1, char *list2, char **splicedList)
{
    int     error = 0;
    char    *newList = 0;
    char    *currentList1;
    char    *currentList2;
    char    *list1Comma;
    char    *list2Comma;
    int     item1Length;
    int     item2Length;

    currentList1 = list1;
    currentList2 = list2;
    *splicedList = 0;

    while (currentList1 || currentList2)
        {
            /*  Find next header item */
        if (currentList1)
            {
            list1Comma = strchr(currentList1, ','); 

            if (list1Comma)
                item1Length = list1Comma - currentList1;
            else
                item1Length = strlen(currentList1);
            }
        else
            item1Length = 0;

            /*  Find next default header item */
        if (currentList2)
            {
            list2Comma = strchr(currentList2, ','); 

            if (list2Comma)
                item2Length = list2Comma - currentList2;
            else
                item2Length = strlen(currentList2);
            }
        else
            item2Length = 0;

        if (item1Length > 0)    /* if there is a header item, use it */
            nullChk( AppendString(&newList, currentList1, item1Length));
        else
        if (item2Length > 0)   /* else, use the default header item, if any */
            nullChk( AppendString(&newList, currentList2, item2Length));

        if (list1Comma || list2Comma)               /* add trailing comma if we aren't on the last item */   
            nullChk( AppendString(&newList, ",", -1));


        if (list1Comma) /* skip to next header */
            currentList1 = list1Comma + 1;
        else
            currentList1 = 0;

        if (list2Comma) /* skip to next header */
            currentList2 = list2Comma + 1;
        else
            currentList2 = 0;
        }

    if (!newList)       /* if list1 and list2 had no items to contribute to list3, then make it an empty list */
        nullChk( AppendString(&newList, "", -1));

Error:
    if (error < 0)
        {
        free(newList);
        *splicedList = 0;
        return error;
        }
    else
        {
        *splicedList = newList;
        return 0;
        }
}

/***********************************************************************/

int CVIFUNC StrToInt (char *str, int *n)
{
    unsigned long   ulval;
    int             negative;

    if (!ConvertNumStrToUint (str, &ulval, &negative))
        return FALSE;
    if (negative)
        {
        if (ulval > 0x80000000L)     /* absolute value is too large */
            return FALSE;
        *n = -(long)ulval;
        }
    else
        {
        if (ulval > 0x7fffffffL)     /* absolute value is too large */
            return FALSE;
        *n = ulval;
        }
    return TRUE;
}

/**************************************************************************/

int CVIFUNC StrToUInt (char *str, unsigned int *n)
{
    unsigned long   ulval;
    int             negative;

    if (!ConvertNumStrToUint (str, &ulval, &negative))
        return FALSE;
    if (negative)
        return FALSE;
    *n = ulval;
    return TRUE;
}

/***********************************************************************/

static int CVIFUNC ConvertNumStrToUint (char *str, unsigned long *n, int *negative)
{
    int             ch;
    int             numDigits = 0;
    int             signSeen = FALSE;
    int             endOfNonWhiteSpace = FALSE;
    unsigned long   ulval;
    char            *s;

    *negative = FALSE;
    ulval = 0;
    for (s=str; ch=*s; s++)
        {
        if (endOfNonWhiteSpace)     /* should be all blanks now */
            if (isspace (ch))
                continue;
            else            /* there was an embedded blank */
                return FALSE;
        if (isspace(ch))
            {
            if (signSeen || numDigits > 0)
                endOfNonWhiteSpace = TRUE;
            continue;
            }
        if ((ch == '+') || (ch == '-')) 
            {
            if (signSeen || numDigits > 0)   /* embedded sign */
                return FALSE;
            if (ch == '-')
                *negative = TRUE;
            signSeen = TRUE;
            continue;
            }
        if (isdigit(ch))
            {
            if (((ulval == 429496729L) && (ch > '5')) ||
                                            (ulval > 429496729L))
                return FALSE;     /* would overlflow */
            ulval = (ulval * 10) + (ch - '0');
            numDigits++;
            continue;
            }
        return FALSE;    /* invalid character */
        }
    if (numDigits == 0)
        return FALSE;
    *n = ulval;
    return TRUE;
}

/********************************************************************/

    /*  this is a more convenient Set/Get-BreakOnLibraryErrors() function */
int CVIFUNC SetBOLE(int on)
{
    int wasOn;

    wasOn = GetBreakOnLibraryErrors();

    if (on)
        EnableBreakOnLibraryErrors();
    else
        DisableBreakOnLibraryErrors();

    return wasOn;
}

/********************************************************************/

static void CVIFUNC InterleavePartialLastScan(void *list, int itemSize, int numChannels, int numScans, int numItems)
{
    int     numChannelsLeft;
    int     channel;
    int     destination;
    char    itemBuf[20];    /* should be large enough for one item of any conceivable numeric type */

    Assert (itemSize < 20);

    numChannelsLeft = numItems - numScans * numChannels;
    for (channel = numChannelsLeft - 1; channel >= 0; channel--)
        {
        memmove(itemBuf, (char *)list + (numItems - 1) * itemSize, itemSize);
        destination = (channel + 1) * numScans;
        memmove((char *)list + (destination + 1) * itemSize, (char *)list + destination * itemSize,
                (numItems - destination - 1) * itemSize);
        memmove((char *)list + destination * itemSize, itemBuf, itemSize);
        }
}

/*******************************************************/

static void CVIFUNC TransposeDataInPlace(void *list, int itemSize, int numPoints, int numChannels)
{
    int     scan;
    int     item;
    int     numScans;
    int     numChannelsLeft = numChannels;
    int     *listPtr = list;

    Assert (itemSize < 20);

    numScans = numPoints / numChannels;
    if (numChannels > 1)
        {
        do  {
            for (scan = 1; scan < numScans; scan++)
                {
                item = listPtr[scan*numChannelsLeft];  /* extract relevant channel from start of scan */
                memmove(&listPtr[scan+1], &listPtr[scan], sizeof(int)* (scan*numChannelsLeft - scan));
                listPtr[scan] = item;
                }

            listPtr = listPtr + numScans;
            numChannelsLeft--;
            } while (numChannelsLeft > 1);

        InterleavePartialLastScan(list, itemSize, numChannels, numScans, numPoints);
        }
}

/*******************************************************/

void CVIFUNC TransposeData(void *data, int dataType, int numPoints, int numChannels)
{
    char    *reOrderedData = 0;
    int     dataSize;
    int     numScans;
    int     oldIndex;
    int     newIndex;
    int     scan = 0;
    int     channel = 0;
    int     numScansForThisChannel;
    int     extraChannelsSampled;   /* > 0 if the last scan is incomplete */

    if (numChannels < 2)
        return;

    dataSize = GetDataTypeSize(dataType);
    numScans = numPoints / numChannels;

    extraChannelsSampled = numPoints - numScans * numChannels;

    if (numScans < 2 &&  extraChannelsSampled == 0)
        return;

    if (dataSize > 0 && numPoints > 0)
        {
        reOrderedData = (char *)malloc(numPoints * dataSize);

        if (reOrderedData)
            {
            newIndex = 0;
            for (channel = 0; channel < numChannels; channel++)
                {
                numScansForThisChannel = numScans;
                if (channel < extraChannelsSampled)
                    numScansForThisChannel++;

                for (scan = 0; scan < numScansForThisChannel; scan++)
                    {
                    oldIndex = scan * numChannels + channel;
                    memmove(reOrderedData + newIndex * dataSize, (char *)data + oldIndex * dataSize, dataSize);
                    newIndex++;
                    }
                }
            memmove(data, reOrderedData, numPoints * dataSize);
            }
        else    /* this is slower, but takes no memory */
            TransposeDataInPlace(data, dataSize, numPoints, numChannels);
        }

    free(reOrderedData);
}

/********************************************************************/

    /*  converts from 1-based heap index to zero based array index */
#define HeapItemPtr(array, oneBasedIndex, elementSize)  ((void *)((char *)(array) + (elementSize) * ((oneBasedIndex)-1)))

    /*  Expects the array to be a heap of index - 1 elements.
        the index'th element (1-based) will be inserted into the heap at the correct
        spot, making array a heap of index elements.
    */
static void CVIFUNC FilterUpToCorrectHeapLocation(void *array, int index, int elementSize, CompareFunction compareFunction)
{
    int     parent;
    void    *indexPtr, *parentPtr;

    for (parent = index >> 1; parent > 0; parent = parent >> 1)
        {
        indexPtr = HeapItemPtr(array, index, elementSize);
        parentPtr = HeapItemPtr(array, parent, elementSize);
        if (compareFunction(indexPtr, parentPtr) <= 0)
            break;  /* element is in correct spot */

        SwapBlock(indexPtr, parentPtr, elementSize);
        index = parent;
        }
}

/**************************************************************************/

    /*  Expects array to be a heap except for the root element (element 1).
        The root element is moved to it correct location so that the array is a heap.
    */
static void CVIFUNC FilterDownToCorrectHeapPosition(void *array, int numElements, int elementSize, CompareFunction compareFunction)
{
    int     leftChild, rightChild, largestChild;
    int     root;
    void    *rootPtr, *leftChildPtr, *rightChildPtr, *largestChildPtr;

    /* use 1-based indexing inside function for array representation of binary tree */

    for (root = 1;;)
        {
        leftChild = root << 1;
        rightChild = leftChild + 1;

        rootPtr = HeapItemPtr(array, root, elementSize);

        if (leftChild <= numElements)
            leftChildPtr = HeapItemPtr(array, leftChild, elementSize);
        else
            leftChildPtr = 0;

        if (rightChild <= numElements)
            rightChildPtr = HeapItemPtr(array, rightChild, elementSize);
        else
            rightChildPtr = 0;

        if (leftChildPtr == 0)
            break;                  /* no more children, we are done */

        if (rightChildPtr == 0)
            largestChild = leftChild;
        else
        if (compareFunction(leftChildPtr, rightChildPtr) >= 0)
            largestChild = leftChild;
        else
            largestChild = rightChild;

        largestChildPtr = HeapItemPtr(array, largestChild, elementSize);

        if (compareFunction(rootPtr, largestChildPtr) >= 0)
            break;  /* element is in correct spot */

        SwapBlock(rootPtr, largestChildPtr, elementSize);
        root = largestChild;
        }
}

/**************************************************************************/

static void CVIFUNC BuildHeap(void *array, int numElements, int elementSize, CompareFunction compareFunction)
{
    int     index;

    for (index = 2; index <= numElements; index++)
        FilterUpToCorrectHeapLocation(array, index, elementSize, compareFunction);
}

/**************************************************************************/

void CVIFUNC HeapSort(void *array, int numElements, int elementSize, CompareFunction compareFunction)
{
    int index;

    BuildHeap(array, numElements, elementSize, compareFunction);

    for (index = numElements; index > 1; index--)
        {
            /*  move greatest item for the root to last position that is still part of the heap
                and move the item there to root of the heap.
            */
        SwapBlock(array, HeapItemPtr(array, index, elementSize), elementSize);

            /*  filter the item at the top of the heap into its correct position in
                the now one element smaller heap.
            */
        FilterDownToCorrectHeapPosition(array, index - 1, elementSize, compareFunction);
        }
}

/********************************************************************/

    /* This function provides a stable alternative to quicksort */
void CVIFUNC InsertionSort(void *array, int numElements, int elementSize, CompareFunction compareFunction)
{
    int     start, item;
    char    *topItem, *bottomItem;

    for (start = 1; start < numElements; start++)
        {
        topItem = (char *)array + start * elementSize;
        bottomItem = topItem - elementSize;

        for (item = start; item > 0; item--)
            {
            if ((*compareFunction)(bottomItem, topItem) > 0)
                {
                SwapBlock(bottomItem, topItem, elementSize);
                topItem -= elementSize;
                bottomItem -= elementSize;
                }
            else
                break;
            }
        }
}

/********************************************************************/

    /* starts a tinny sounding whine on the built PC speaker */
void CVIFUNC StartPCSound(unsigned int frequency)
{
#if defined(_NI_mswin_)  /* _NI_mswin_ is defined automatically when compiling with the Windows version of CVI. */
    outp(0x43, (char)0xb6);         /* prepare timer */

    if (frequency > 0)          /* divide number of cycles per second by the input frequency */
        frequency = (unsigned) (1193180L/frequency);

    outp(0x42, (char)frequency);
    outp(0x42, (char)(frequency >> 8));
    outp(0x61, (char)(inp(0x61) | 0x3));    /* turn on sound with bits 0 and 1 */
#endif /* _NI_mswin_ */
}

/****************************************************************************/

    /* A function you would pay money for after calling StartPCSound(). */
void CVIFUNC StopPCSound(void)
{
#if defined(_NI_mswin_)  /* _NI_mswin_ is defined automatically when compiling with the Windows version of CVI. */
    outp(0x61, (char)(inp(0x61) & ~0x3));   /* turn off sound with bits 0 and 1 */
#endif /* _NI_mswin_ */
}

/********************************************************************/

Handle CVIFUNC NewHandle(unsigned int numBytes)
{
    void            *memPtr;
    HandleRecord    *hanPtr;

    memPtr = calloc(numBytes, 1);
    hanPtr = (HandleRecord*)calloc(sizeof(HandleRecord), 1);
    if (hanPtr && (memPtr || numBytes == 0))
        {
        hanPtr->ptr = memPtr;
        hanPtr->size = numBytes;
        return (Handle)hanPtr;
        }
    else
        {
        free(memPtr);
        free(hanPtr);
        return NULL;
        }
}

/********************************************************************/

void CVIFUNC DisposeHandle(Handle handle)
{
    if (handle)
        {
        free(*handle);
        free((void *)handle);
        }
}

/********************************************************************/

unsigned int CVIFUNC GetHandleSize(Handle handle)
{
    AssertMsg(handle, "Parameter 1 to GetHandleSize was NULL.");
    return ((HandleRecord *)handle)->size;
}

/********************************************************************/

int CVIFUNC SetHandleSize(Handle handle, unsigned int newSize)
{
    HandleRecord    *hanRecPtr = (HandleRecord *)handle;
    void            *newPtr, *oldPtr;
    unsigned int    oldSize;

    AssertMsg(hanRecPtr, "Parameter 1 to SetHandleSize was NULL.");

    oldPtr = hanRecPtr->ptr;
    oldSize = hanRecPtr->size;
    if (oldSize == newSize)
        return TRUE;

    if (oldPtr == NULL)
        newPtr = malloc(newSize);
    else
        newPtr = realloc(oldPtr, newSize);

    if (newPtr || (newSize == 0))
        {
        hanRecPtr->ptr = newPtr;
        hanRecPtr->size = newSize;
        if (newSize > oldSize)
            memset ((char *)newPtr+oldSize, 0, newSize-oldSize);
        return TRUE;
        }
    else
        return FALSE;
}

/*******************************************************************************/

    /* BEGIN PATCH to 4.0.1 - jag 8-29-96 */
    /*  If numNewItems == 0 then expand the list by the number of items indicated by its allocation policy.
        If numNewItems > 0 then expand the list by exactly the number of items indicated.
        If numNewItems < 0 then expand the list by the absolute value of numNewItems plus the number of items indicated by its allocation policy.
        Returns TRUE for success, FALSE if out of memory
    */
static int CVIFUNC ExpandListSpace (ListType list, int numNewItems)
{
    if (numNewItems == 0)
        numNewItems = NUMITEMSPERALLOC(list);
    else
    if (numNewItems < 0)
        numNewItems = (-numNewItems) + NUMITEMSPERALLOC(list);

    /* END PATCH to 4.0.1 - jag 8-29-96 */
    if (SetHandleSize ((Handle)list,
        sizeof (ListStruct) +
        ((*list)->listSize + numNewItems) * (*list)->itemSize)) {
        (*list)->listSize += numNewItems;
        return TRUE;
        }
    else
        return FALSE;
}

/*******************************/

     /* This function reallocate the list, minus any currently unused portion of its allotted memory. */
void CVIFUNC ListCompact (ListType list)
{
    if (!list)
        return;

    if (!SetHandleSize ((Handle)list,
        sizeof (ListStruct) + (*list)->numItems * (*list)->itemSize))
        return;

    (*list)->listSize = (*list)->numItems;
}

/*******************************/

ListType CVIFUNC ListCreate (int elementSize)
{
    ListType    list;

     list = (ListType)(NewHandle (sizeof(ListStruct)));   /* create empty list */
     if (list) {
        (*list)->signature = LIST_SIGNATURE;
        (*list)->numItems = 0;
        (*list)->listSize = 0;
        (*list)->itemSize = elementSize;
        (*list)->percentIncrease = kDefaultAllocationPercentIncrease;
        (*list)->minNumItemsIncrease = kDefaultAllocationminNumItemsIncrease;
        }

    return list;
}

/*******************************/

void CVIFUNC ListSetAllocationPolicy (ListType list, int minItemsPerAlloc, int percentIncreasePerAlloc)
{
    if (!list)
        return;
        
    (*list)->percentIncrease = percentIncreasePerAlloc;
    (*list)->minNumItemsIncrease = minItemsPerAlloc;
}

/*******************************/

void CVIFUNC ListDispose (ListType list)
{
    DisposeHandle ((Handle)list);
}

/*******************************/

void CVIFUNC ListDisposePtrList(ListType list)
{
    int index; 
    int numItems;
    
    if (list)
        {
        numItems = ListNumItems(list);
        
        for (index = 1; index <= numItems; index++)
            free(*(void **)ListGetPtrToItem(list, index));
            
        ListDispose(list);
        }
}

/*******************************/

void  CVIFUNC ListClear (ListType list)  /* keeps memory, resets the number of items to 0 */
{
    if (!list)
        return;
    (*list)->numItems = 0;
}

/*******************************/

ListType CVIFUNC ListCopy (ListType originalList)   /* copy is only as large as necessary */
{
    ListType    tempList = NULL;
    int         numItems;

    if (!originalList)
        return NULL;

    tempList = ListCreate ((*originalList)->itemSize);
    if (tempList) {
        numItems = ListNumItems (originalList);

        if (!SetHandleSize ((Handle)tempList,
            sizeof (ListStruct) + numItems * (*tempList)->itemSize)) {
            ListDispose (tempList);
            return NULL;
            }

        (*tempList)->numItems = (*originalList)->numItems;
        (*tempList)->listSize = (*originalList)->numItems;
        (*tempList)->itemSize = (*originalList)->itemSize;
        (*tempList)->percentIncrease = (*originalList)->percentIncrease;
        (*tempList)->minNumItemsIncrease = (*originalList)->minNumItemsIncrease;

        memmove (ITEMPTR(tempList,0), ITEMPTR(originalList,0),
            numItems * (*tempList)->itemSize);
        }

    return tempList;
}

/********************************/

int CVIFUNC ListAppend (ListType list1, ListType list2)  /* list1 = list1 + list2 */
{
    int numItemsL1, numItemsL2;

    if (!list2)
        return TRUE;

    if (!list1)
        return FALSE;
    if ((*list1)->itemSize != (*list2)->itemSize)
        return FALSE;

    numItemsL1 = ListNumItems (list1);
    numItemsL2 = ListNumItems (list2);

    if (numItemsL2 == 0)
        return TRUE;

    if (!SetHandleSize ((Handle)list1,
        sizeof (ListStruct) + (numItemsL1 + numItemsL2) * (*list1)->itemSize))
        return FALSE;

    (*list1)->numItems = numItemsL1 + numItemsL2;
    (*list1)->listSize = numItemsL1 + numItemsL2;

    memmove (ITEMPTR(list1,numItemsL1), ITEMPTR(list2,0),
        numItemsL2 * (*list2)->itemSize);

    return TRUE;
}

/*******************************/

    /*  returns TRUE if the item is inserted, returns FALSE if out of memory or
        bad arguments were passed.
    */
int CVIFUNC ListInsertItem (ListType list, void *ptrToItem, int itemPosition)
{
    return ListInsertItems (list, ptrToItem, itemPosition, 1);
}

/*******************************/

int CVIFUNC ListInsertItems (ListType list, void *ptrToItems, int firstItemPosition, int numItemsToInsert)
{
    if (!list)
        return FALSE;
    if (numItemsToInsert <= 0)
        return FALSE;

    if (firstItemPosition == (*list)->numItems + 1)
        firstItemPosition = END_OF_LIST;

    if (firstItemPosition > (*list)->numItems)
        return FALSE;

    if ((*list)->numItems >= (*list)->listSize)
        if (!ExpandListSpace (list, -numItemsToInsert)) /* PATCH to 4.0.1 - jag 8-29-96 */
            return FALSE;

    if (firstItemPosition == FRONT_OF_LIST)
        if ((*list)->numItems == 0)
           firstItemPosition = END_OF_LIST; /* special case for empty list */
        else
           firstItemPosition = 1;

    if (firstItemPosition == END_OF_LIST) { /* add at the end of the list */
        if (ptrToItems)
            memmove (ITEMPTR(list,(*list)->numItems), ptrToItems,
                (*list)->itemSize * numItemsToInsert);
        else
            memset (ITEMPTR(list,(*list)->numItems), 0,
                (*list)->itemSize * numItemsToInsert);

        (*list)->numItems += numItemsToInsert;
        }
    else {                                  /* move part of list up to make room for new item */
        memmove (ITEMPTR(list,firstItemPosition-1+numItemsToInsert),
            ITEMPTR(list,firstItemPosition-1),
            ((*list)->numItems + 1 - firstItemPosition) * (*list)->itemSize);

        if (ptrToItems)
            memmove (ITEMPTR(list,firstItemPosition-1), ptrToItems,
                (*list)->itemSize * numItemsToInsert);
        else
            memset (ITEMPTR(list,firstItemPosition-1), 0,
                (*list)->itemSize * numItemsToInsert);

        (*list)->numItems += numItemsToInsert;
        }

    return TRUE;
}

/*******************************/

int CVIFUNC ListEqual (ListType list1, ListType list2)
{
    if (list1 == list2)
        return TRUE;

    if (list1 == NULL || list2 == NULL)
        return FALSE;

    if ((*list1)->itemSize == (*list1)->itemSize)
        if ((*list1)->numItems == (*list2)->numItems)
            return (memcmp (ITEMPTR(list1,0), ITEMPTR(list2,0),
                (*list1)->itemSize * (*list1)->numItems) == 0);

    return FALSE;
}

/*******************************/

     /* The item pointed to by ptrToItem is copied over the current item at itemPosition */
void CVIFUNC ListReplaceItem (ListType list, void *ptrToItem, int itemPosition)
{
    ListReplaceItems (list, ptrToItem, itemPosition, 1);
}

/*******************************/

     /* The item pointed to by ptrToItems is copied over the current item at itemPosition */
void CVIFUNC ListReplaceItems (ListType list, void *ptrToItems, int firstItemPosition, int numItemsToReplace)
{
    if (!list || !ptrToItems || numItemsToReplace <= 0)
        return;
    if ((*list)->numItems == 0)
        return;
        
    if (firstItemPosition == END_OF_LIST)
        firstItemPosition = (*list)->numItems;

    if (firstItemPosition == FRONT_OF_LIST)
        firstItemPosition = 1;

    if (firstItemPosition <= 0  ||
        firstItemPosition + numItemsToReplace - 1 > (*list)->numItems)
        return;

    memmove (ITEMPTR(list,firstItemPosition-1), ptrToItems,
        (*list)->itemSize * numItemsToReplace);
}

/*******************************/

void CVIFUNC ListRemoveItem (ListType list, void *itemDestination, int itemPosition)
{
    ListRemoveItems (list, itemDestination, itemPosition, 1);
}

/*******************************/

void CVIFUNC ListRemoveItems (ListType list, void *itemsDestination, int firstItemPosition, int numItemsToRemove)
{
    if (!list || numItemsToRemove <= 0 || numItemsToRemove > (*list)->numItems)
        return;

    if (firstItemPosition == FRONT_OF_LIST)
        firstItemPosition = 1;

    if (firstItemPosition == END_OF_LIST)
        firstItemPosition = (*list)->numItems;

    if (firstItemPosition <= 0 ||
        firstItemPosition + numItemsToRemove - 1 > (*list)->numItems)   /* can't remove more than are there */
        return;

    if (itemsDestination != NULL)
         memmove (itemsDestination, ITEMPTR(list,firstItemPosition-1),
            (*list)->itemSize * numItemsToRemove);

           /* move part of list down to cover hole left by removed item */
    memmove (ITEMPTR(list,firstItemPosition-1),
        ITEMPTR(list,firstItemPosition-1+numItemsToRemove),
        ((*list)->numItems - firstItemPosition) * (*list)->itemSize * numItemsToRemove);

    (*list)->numItems -= numItemsToRemove;
}

/*******************************/

void CVIFUNC ListGetItem (ListType list, void *itemDestination, int itemPosition)
{
    ListGetItems (list, itemDestination, itemPosition, 1);
}

/*******************************/

void CVIFUNC ListGetItems(ListType list, void *itemsDestination, int firstItemPosition, int numItemsToGet)
{
    if (!list || !itemsDestination || numItemsToGet <= 0)
        return;
    if ((*list)->numItems == 0)
        return;

    if (firstItemPosition == FRONT_OF_LIST)
        firstItemPosition = 1;

    if (firstItemPosition == END_OF_LIST)
        firstItemPosition = (*list)->numItems;

    if (firstItemPosition <= 0  ||
        firstItemPosition + numItemsToGet - 1 > (*list)->numItems)
        return;

    memmove (itemsDestination, ITEMPTR(list,firstItemPosition-1),
        (*list)->itemSize * numItemsToGet);
}

/*******************************/

    /*  Returns a pointer to the item at itemPosition. returns null if an errors occurred.
    */
void * CVIFUNC ListGetPtrToItem (ListType list, int itemPosition)
{
    if (!list)
        return NULL;
    if ((*list)->numItems == 0)
        return NULL;

    if (itemPosition == FRONT_OF_LIST)
        itemPosition = 1;

    if (itemPosition == END_OF_LIST)
        itemPosition = (*list)->numItems;

    if (itemPosition <= 0 || itemPosition > (*list)->numItems)
        return NULL;

    return ITEMPTR(list,itemPosition-1);
}

/*******************************/

     /* returns a pointer the lists data (abstraction violation for optimization) */
void * CVIFUNC ListGetDataPtr (ListType list)
{
    if (!list)
        return NULL;

    return &((*list)->itemList[0]);
}

/********************************/

int CVIFUNC ListApplyToEach (ListType list, int ascending, ListApplicationFunc funcToApply, void *callbackData)
{
    int result = 0, index;

    if (!list || !funcToApply)
        goto Error;

    if (ascending) {
        for (index = 1; index <= ListNumItems (list); index++) {
            result  = funcToApply (index, ListGetPtrToItem (list, index),
                callbackData);
            if (result < 0)
                goto Error;
            }
        }
    else {
        for (index = ListNumItems (list); index > 0 &&
            index <= ListNumItems (list); index--) {
            result  = funcToApply (index, ListGetPtrToItem (list, index),
                callbackData);
            if (result < 0)
                goto Error;
            }
        }

Error:
    return result;
}

/********************************/

int CVIFUNC ListGetItemSize (ListType list)
{
    if (!list)
        return 0;

    return (*list)->itemSize;
}

/********************************/

int CVIFUNC ListNumItems (ListType list)
{
    if (!list)
        return 0;
    else
        return (*list)->numItems;
}

/*******************************/

void CVIFUNC ListRemoveDuplicates (ListType list, CompareFunction compareFunction)
{
    int numItems, index, startIndexForFind, duplicatesIndex;

    numItems = ListNumItems (list);

    for (index = 1; index < numItems; index++) {
        startIndexForFind = index + 1;
        while (startIndexForFind <= numItems) {
            duplicatesIndex = ListFindItem (list, ListGetPtrToItem (list, index),
                startIndexForFind, compareFunction);
            if (duplicatesIndex > 0) {
                ListRemoveItem (list, NULL, duplicatesIndex);
                numItems--;
                startIndexForFind = duplicatesIndex;
                }
             else
                break;
            }
        }
}

/*******************************/

int CVIFUNC ListFindItem (ListType list, void *ptrToItem, int startingPosition, CompareFunction compareFunction)
{
    int index;

    if (!list || !ptrToItem)
        return 0;

    if ((*list)->numItems == 0)
        return 0;

    SetDefaultCompareFunctionIfNeeded (list, &compareFunction);

    if (startingPosition == FRONT_OF_LIST)
        startingPosition = 1;

    if (startingPosition == END_OF_LIST)
        startingPosition = (*list)->numItems;

    if (startingPosition <= 0 || startingPosition > (*list)->numItems)
        return 0;

    for (index = startingPosition; index <= (*list)->numItems; index++)
        if (compareFunction (ptrToItem, ITEMPTR(list,index-1)) == 0)
            return index;

    return 0;
}

/*******************************/

int CVICALLBACK ShortCompare(void *a, void *b)
{
    if (*(short *)a < *(short *)b) return -1;
    if (*(short *)a > *(short *)b) return  1;
    return 0;
}

/*******************************/

int CVICALLBACK IntCompare(void *a, void *b)
{
   if (*(int *)a < *(int *)b) return -1;
   if (*(int *)a > *(int *)b) return  1;
   return 0;
}

/*******************************/

int CVICALLBACK FloatCompare(void *a, void *b)
{
   if (*(float *)a < *(float *)b) return -1;
   if (*(float *)a > *(float *)b) return  1;
   return 0;
}

/*******************************/

int CVICALLBACK DoubleCompare(void *a, void *b)
{
   if (*(double *)a < *(double *)b) return -1;
   if (*(double *)a > *(double *)b) return  1;
   return 0;
}

/*******************************/

int CVICALLBACK CStringCompare(void *a, void *b)
{
    return strcmp(*(char **)a, *(char **)b);
}

/*******************************/

static gListMemBlockCompareSize = 1;

    /*  This function is needed because the prototype for the comparison function
        doesn't have a parameter for size.
    */
static void CVIFUNC SetListMemBlockCompareSize(unsigned int size)
{
    gListMemBlockCompareSize = size;
}

/*******************************/

    /*  Used to compare list elements by their raw data contents */
static int CVICALLBACK ListMemBlockCompare(void *a, void *b)
{
    return memcmp(a, b, gListMemBlockCompareSize);
}

/*******************************/

static void CVIFUNC SetDefaultCompareFunctionIfNeeded (ListType list, CompareFunction *compareFunction)
{
    if (*compareFunction == NULL) {
        SetListMemBlockCompareSize ((*list)->itemSize);
        *compareFunction =  ListMemBlockCompare;
        }
}

/*******************************/

int CVIFUNC ListBinSearch (ListType list, void *ptrToItem, CompareFunction compareFunction)
{
    int index;

    if (!list || !ptrToItem)
        return 0;

    SetDefaultCompareFunctionIfNeeded (list, &compareFunction);

    index = BinSearch (ITEMPTR(list,0), (int)(*list)->numItems,
        (int)(*list)->itemSize, ptrToItem, compareFunction);

    if (index >= 0)
        index++;        /* lists start from 1 */
    else
        index = 0;      /* item not found */

    return index;
}

/*******************************/

int CVIFUNC ListIsSorted (ListType list, CompareFunction compareFunction)
{
    int numItems, numItemsMinusOne, index;

    if (!list)
        return FALSE;

    SetDefaultCompareFunctionIfNeeded (list, &compareFunction);

    if ((numItems = (*list)->numItems) == 0)
        return TRUE;

    numItemsMinusOne = numItems-1;

    for (index = 0; index < numItemsMinusOne; index++)
        if ((*compareFunction) ((int *)ITEMPTR(list,index),
            (int *)ITEMPTR(list,index+1)) > 0)
            return FALSE;

    return TRUE;
}

/**********************************/

void CVIFUNC ListQuickSort (ListType list, CompareFunction compareFunction)
{
    typedef (CVIANSI * qsortCmpFnc)(const void *, const void *);
    
    if (!list)
        return;

    SetDefaultCompareFunctionIfNeeded (list, &compareFunction);

    if ((*list)->numItems == 0)
        return;

    qsort (ITEMPTR(list,0), (*list)->numItems, (*list)->itemSize,
        (qsortCmpFnc)compareFunction);      /* assumes CVICALLBACK and CVIANSI are the same */
}

/*******************************/

void CVIFUNC ListHeapSort (ListType list, CompareFunction compareFunction)
{
    if (!list)
        return;

    SetDefaultCompareFunctionIfNeeded (list, &compareFunction);

    if ((*list)->numItems == 0)
        return;

    HeapSort (ITEMPTR(list,0), (*list)->numItems, (*list)->itemSize,
        compareFunction);
}

/*******************************/

void CVIFUNC ListInsertionSort (ListType list, CompareFunction compareFunction)
{
    if (!list)
        return;

    SetDefaultCompareFunctionIfNeeded (list, &compareFunction);

    if ((*list)->numItems == 0)
        return;

    InsertionSort (ITEMPTR(list,0), (*list)->numItems, (*list)->itemSize,
        compareFunction);
}

/*******************************/

    /*  returns TRUE if insertion is successful, returns FALSE if out of memory or
        if bad arguments are passed.
    */
int CVIFUNC ListInsertInOrder (ListType list, void *ptrToItem, CompareFunction compareFunction)
{
    int numItems, InsertPosition;

    if (!list || !ptrToItem)
        return FALSE;

    SetDefaultCompareFunctionIfNeeded (list, &compareFunction);

    numItems = ListNumItems (list);

    if (numItems == 0)
        InsertPosition = FRONT_OF_LIST;
    else {
        InsertPosition = BinSearch (ITEMPTR(list,0), (int)(*list)->numItems,
            (int)(*list)->itemSize, ptrToItem, compareFunction);
        if (InsertPosition >= 0)
            InsertPosition++;           /* found matching item, list indices start from 1 */
        else {
            InsertPosition = -(InsertPosition+1);   /* BinSearch returns insertion point i in the form -i-1, convert to i: */
            InsertPosition++;           /* list indices start from 1 */

            if (InsertPosition > numItems)
                InsertPosition = END_OF_LIST;
            }
        }

    return ListInsertItem (list, ptrToItem, InsertPosition);
}

/**************************************************************************/

    /*  Reserves memory for numItems in the list. If it succeeds then
        numItems items can be inserted without possibility of an
        out of memory error (useful to simplify error recovery in
        complex functions). Returns TRUE if success, FALSE if
        out of memory.
    */
int CVIFUNC ListPreAllocate (ListType list, int numItems)
{
    if (!list)
        return FALSE;

    if ((*list)->listSize - (*list)->numItems < numItems)
        return ExpandListSpace (list,
            numItems - ((*list)->listSize - (*list)->numItems));
    else
        return TRUE;    /* enough items are already pre-allocated */
}

/**************************************************************************/

int CVICALLBACK OutputIntegerItem(FILE *stream, void *ptrToItem)
{
    return fprintf(stream, "%d", *(int *)ptrToItem);
}

/**************************************************************************/

int CVICALLBACK OutputDoubleItem(FILE *stream, void *ptrToItem)
{
    return fprintf(stream, "%f", *(double *)ptrToItem);
}

/**************************************************************************/

int CVICALLBACK OutputShortItem(FILE *stream, void *ptrToItem)
{
    return fprintf(stream, "%d", (int)*(short int *)ptrToItem);
}

/**************************************************************************/

int CVICALLBACK OutputStringItem(FILE *stream, void *ptrToItem)
{
    return fprintf(stream, "%s", *(char **)ptrToItem);
}

/**************************************************************************/

int CVIFUNC OutputList (ListType list, FILE *stream, char *listPrefix, char *listSuffix, char *itemPrefix, char *itemSuffix, ItemOutputFunction itemOutputFunction)
{
    int index, numItems, totalCharsWritten = 0, charsWritten;

    if (!stream)
        stream = stdout;

    if (!itemOutputFunction)
        return fprintf (stream, "Error, no output function specified\n");

    if (listPrefix)
        if ((charsWritten = fprintf (stream, listPrefix)) < 0)
            return charsWritten;
        else
            totalCharsWritten += charsWritten;

    numItems = ListNumItems (list);

    for (index = 1; index <= numItems; index++) {
        if (itemPrefix)
            if (strstr (itemPrefix, "%d"))
                if ((charsWritten = fprintf (stream, itemPrefix, index)) < 0)
                    return charsWritten;
                else
                    totalCharsWritten += charsWritten;
            else
                if ((charsWritten = fprintf(stream, itemPrefix)) < 0)
                    return charsWritten;
                else
                    totalCharsWritten += charsWritten;

        if ((charsWritten = itemOutputFunction (stream,
            ListGetPtrToItem (list, index))) < 0)
            return charsWritten;
        else
            totalCharsWritten += charsWritten;

        if (itemSuffix && index != numItems)
            if (strstr (itemSuffix, "%d"))
                if ((charsWritten = fprintf (stream, itemSuffix, index)) < 0)
                    return charsWritten;
                else
                    totalCharsWritten += charsWritten;
            else
                if ((charsWritten = fprintf (stream, itemSuffix)) < 0)
                    return charsWritten;
                else
                    totalCharsWritten += charsWritten;
        }

    if (listSuffix)
        if (strstr (itemSuffix, "%d"))
            if ((charsWritten = fprintf (stream, listSuffix, numItems)) < 0)
                return charsWritten;
            else
                totalCharsWritten += charsWritten;
        else
            if ((charsWritten = fprintf (stream, listSuffix)) < 0)
                return charsWritten;
            else
                totalCharsWritten += charsWritten;

    return totalCharsWritten;
}

/**************************************************************************/

int CVIFUNC_C AttributeMaxAndMin(int panel, int attribute, int numCtrls, int *max,
                int *ctrlWithMax, int *min, int *ctrlWithMin, ...)
{
    va_list     parmInfo;
    int         error = 0;
    ListType    ctrlList = 0;

    va_start(parmInfo, ctrlWithMin);
    errChk( GetIntListFromParmInfo(parmInfo, numCtrls, &ctrlList));

    errChk( AttrMaxAndMinForList(panel, ctrlList, attribute, max,
                ctrlWithMax, min, ctrlWithMin));

Error:
    ListDispose(ctrlList);
    va_end(parmInfo);
    if (error >= 0)
        return 0;
    else
        return error;
}

/**************************************************************************/

static int CVIFUNC IsTBBoundsAttr(int attribute)
{
    switch (attribute)
        {
        case ATTR_TB_BOUNDS_TOP:
        case ATTR_TB_BOUNDS_LEFT:
        case ATTR_TB_BOUNDS_HEIGHT:
        case ATTR_TB_BOUNDS_WIDTH:
            return TRUE;
        default:
            return FALSE;
        }
}

/**************************************************************************/

int CVIFUNC AttrMaxAndMinForList(int panel, ListType ctrlList, int attribute, int *max,
                int *ctrlWithMax, int *min, int *ctrlWithMin)
{
    int error = 0;
    int value;
    int ctrl;
    int numCtrls;
    int index;
    int maximum, minimum;
    int maxCtrl, minCtrl;

    numCtrls = ListNumItems(ctrlList);
    for (index = 1; index <= numCtrls; index++)
        {
        ListGetItem(ctrlList, &ctrl, index);

        if (IsTBBoundsAttr(attribute))
            errChk( GetCtrlBoundRectAttribute(panel, ctrl, attribute, &value));
        else
            errChk( GetCtrlAttribute(panel, ctrl, attribute, &value));

        if (index == 1)
            {
            maximum = value;
            minimum = value;
            maxCtrl = ctrl;
            minCtrl = ctrl;
            }
        else
            {
            if (value > maximum)
                {
                maximum = value;
                maxCtrl = ctrl;
                }

            if (value < minimum)
                {
                minimum = value;
                minCtrl = ctrl;
                }
            }
        }

    if (max)
        *max = maximum;

    if (min)
        *min = minimum;

    if (ctrlWithMax)
        *ctrlWithMax = maxCtrl;

    if (ctrlWithMin)
        *ctrlWithMin = minCtrl;

Error:
    if (error >= 0)
        return 0;
    else
        return error;
}

/**************************************************************************/

int CVIFUNC_C SetAttributeForCtrls(int panel, int attribute, int value, int numCtrls, ...)
{
    va_list     parmInfo;
    int         error = 0;
    ListType    ctrlList = 0;

    va_start(parmInfo, numCtrls);
    errChk( GetIntListFromParmInfo(parmInfo, numCtrls, &ctrlList));

    errChk( SetAttributeForList(panel, ctrlList, attribute, value));

Error:
    ListDispose(ctrlList);
    va_end(parmInfo);
    if (error >= 0)
        return 0;
    else
        return error;
}

/**************************************************************************/

int CVIFUNC SetAttributeForList(int panel, ListType ctrlList,int attribute, int value)
{
    int     error = 0;
    int     ctrl;
    int     numCtrls;
    int     index;

    numCtrls = ListNumItems(ctrlList);
    for (index = 1; index <= numCtrls; index++)
        {
        ListGetItem(ctrlList, &ctrl, index);

        errChk( SetCtrlAttribute(panel, ctrl, attribute, value));
        }

Error:
    if (error >= 0)
        return 0;
    else
        return error;
}

/**************************************************************************/

int CVIFUNC PutLabelOnLeft(int panel, int ctrl, int labelToCtrlGap)
{
    int error = 0;
    int ctrlLeft;
    int labelWidth;

    errChk( GetCtrlAttribute(panel, ctrl, ATTR_LEFT, &ctrlLeft));

    errChk( GetCtrlAttribute(panel, ctrl, ATTR_LABEL_WIDTH, &labelWidth));

    errChk( SetCtrlAttribute(panel, ctrl, ATTR_LABEL_LEFT, ctrlLeft - labelWidth - 1));

    errChk( SetCtrlAttribute(panel, ctrl, ATTR_LABEL_TOP, VAL_AUTO_CENTER));

    errChk( SetCtrlAttribute(panel, ctrl, ATTR_LABEL_LEFT, ctrlLeft - labelWidth - labelToCtrlGap));

Error:
    if (error >= 0)
        return 0;
    else
        return error;
}

/**************************************************************************/

    /*  pass numCtrls == 0 to pull off arguments until a zero argument is reached */
int CVIFUNC GetIntListFromParmInfo(va_list parmInfo, int numIntegers, ListType *integerList)
{
    int error = 0;
    int integer;
    int numIntegersSoFar = 0;

    nullChk( *integerList = ListCreate(sizeof(int)));

    for (integer = va_arg(parmInfo, int); integer > 1; numIntegersSoFar++, integer = va_arg(parmInfo,int))
        {
        if (numIntegersSoFar >= numIntegers && numIntegers != 0)
            break;

        if (!ListInsertItem(*integerList, &integer, END_OF_LIST))
            errChk( UIEOutOfMemory);
        }

Error:
    if (error < 0)
        {
        ListDispose(*integerList);
        *integerList = 0;
        return error;
        }
    else
        return 0;
}

/**************************************************************************/

    /*  passed numIntegers == 0 to pull off integers until a zero is found */
ListType CVIFUNC_C BuildIntegerList(int numIntegers, ...)
{
    va_list     parmInfo;
    ListType    integerList = 0;

    va_start(parmInfo, numIntegers);
    GetIntListFromParmInfo(parmInfo, numIntegers, &integerList);
    va_end(parmInfo);
    return integerList;
}

/**************************************************************************/

    /*  passed numIntegers == 0 to pull off integers until a zero is found */
    /*  returns non-zero if success, 0 if out of memory */
int CVIFUNC_C AppendToIntegerList(ListType listToAppendTo, int numIntegers, ...)
{
    int         error = 0;
    va_list     parmInfo;
    ListType    integerList = 0;

    va_start(parmInfo, numIntegers);
    errChk( GetIntListFromParmInfo(parmInfo, numIntegers, &integerList));

    nullChk( ListAppend(listToAppendTo, integerList));

Error:
    va_end(parmInfo);
    ListDispose(integerList);
    return (error == 0);
}

/**************************************************************************/

int CVIFUNC ListInsertAfterItem(ListType list, void *ptrToItem, void *ptrToItemToInsertAfter, CompareFunction compareFunction)
{
    int index;

    index = ListFindItem(list, ptrToItemToInsertAfter, FRONT_OF_LIST, compareFunction);

    if (index == 0)
        index = END_OF_LIST;
    else
        index++;

    return ListInsertItem(list, ptrToItem, index);
}

/**************************************************************************/

int CVIFUNC ListInsertBeforeItem(ListType list, void *ptrToItem, void *ptrToItemToInsertBefore, CompareFunction compareFunction)
{
    int index;

    index = ListFindItem(list, ptrToItemToInsertBefore, FRONT_OF_LIST, compareFunction);

    if (index == 0)
        index = FRONT_OF_LIST;

    return ListInsertItem(list, ptrToItem, index);
}

/**************************************************************************/

int CVIFUNC PutLabelsOnLeftForList(int panel, ListType ctrlList, int labelToCtrlGap)
{
    int     error = 0;
    int     ctrl, index, numCtrls;

    numCtrls = ListNumItems(ctrlList);

    for (index = 1; index <= numCtrls; index++)
        {
        ListGetItem(ctrlList, &ctrl, index);

        errChk( PutLabelOnLeft(panel, ctrl, labelToCtrlGap));
        }

Error:
    if(error >= 0)
        return 0;
    else
        return error;
}

/**************************************************************************/

int CVIFUNC_C PutLabelsOnLeft(int panel, int labelToCtrlGap, int numCtrls, ...)
{
    va_list     parmInfo;
    int         error = 0;
    ListType    ctrlList = 0;

    va_start(parmInfo, numCtrls);
    errChk( GetIntListFromParmInfo(parmInfo, numCtrls, &ctrlList));

    errChk( PutLabelsOnLeftForList(panel, ctrlList, labelToCtrlGap));

Error:
    ListDispose(ctrlList);
    va_end(parmInfo);
    if (error >= 0)
        return 0;
    else
        return error;
}

/**************************************************************************/

int CVIFUNC GetCtrlBoundRectAttribute(int panel, int ctrl, int attribute, int *value)
{
    int error = 0;
    int top, left, width, height;

    errChk( GetCtrlBoundingRect(panel, ctrl, &top, &left, &height, &width));

    switch (attribute)
        {
        case ATTR_LEFT:
        case ATTR_TB_BOUNDS_LEFT:
            *value = left;
            break;
        case ATTR_TOP:
        case ATTR_TB_BOUNDS_TOP:
            *value = top;
            break;
        case ATTR_WIDTH:
        case ATTR_TB_BOUNDS_WIDTH:
            *value = width;
            break;
        case ATTR_HEIGHT:
        case ATTR_TB_BOUNDS_HEIGHT:
            *value = height;
            break;
        default:
            AssertMsg(0, "Bad attribute passed to GetCtrlBoundRectAttribute()");
            break;
        }

Error:
    return error;
}

/**************************************************************************/

int CVIFUNC SetCtrlBoundRectAttribute(int panel, int ctrl, int attribute, int value)
{
    int error = 0;
    int mainCtrlPartPosition;
    int oldPosition;

    errChk( GetCtrlBoundRectAttribute(panel, ctrl, attribute, &oldPosition));
    errChk( GetCtrlAttribute(panel, ctrl, attribute, &mainCtrlPartPosition));

    switch (attribute)
        {
        case ATTR_LEFT:
        case ATTR_TOP:
            errChk( SetCtrlAttribute(panel, ctrl, attribute, (value - oldPosition) + mainCtrlPartPosition));
            break;
        case ATTR_WIDTH:
        case ATTR_HEIGHT:
            errChk( UIEAttributeNotSettable);;
        default:
            AssertMsg(0, "Bad attribute passed to SetCtrlBoundRectAttribute()");
            break;
        }

Error:
    return error;
}

/**************************************************************************/

    /*  Set direction = 0 for horozontal distribution, set direction = 1 for
        vertical distribution.  Set space = 0 for start to end spacing, set
        spacing = 1 for fixed gap spacing.
    */
int CVIFUNC DistributeCtrlsInList(int panel, ListType ctrlList, int direction, int spacing, int *position, int gap)
{
    int     error = 0;
    int     ctrl;
    int     index;
    int     numCtrls;
    int     positionAttribute, sizeAttribute;
    int     totalSize, size;
    int     ctrlGap;

    if (!position)
        errChk( UIENullPointerPassed);

    numCtrls = ListNumItems(ctrlList);

    if (direction == VAL_TB_VERTICAL_DISTRIBUTION)
        {
        positionAttribute = ATTR_TOP;
        sizeAttribute = ATTR_HEIGHT;
        }
    else
        {
        positionAttribute = ATTR_LEFT;
        sizeAttribute = ATTR_WIDTH;
        }

    if (spacing == VAL_TB_FIXED_GAP_SPACING)
        ctrlGap = gap;
    else
    if (numCtrls < 2)
        ctrlGap = 2;
    else
        {
        totalSize = 0;
        for (index = 1; index <= numCtrls; index++)
            {
            ListGetItem(ctrlList, &ctrl, index);

            errChk( GetCtrlBoundRectAttribute(panel, ctrl, sizeAttribute, &size));

            totalSize = totalSize + size;
            }

        ctrlGap = (gap - totalSize)/ (numCtrls - 1);
        }

    for (index = 1; index <= numCtrls; index++)
        {
        ListGetItem(ctrlList, &ctrl, index);

        errChk( GetCtrlBoundRectAttribute(panel, ctrl, sizeAttribute, &size));

        errChk( SetCtrlBoundRectAttribute(panel, ctrl,  positionAttribute, *position));

        *position += size;
        if (index != numCtrls)
            *position += ctrlGap;
        }

Error:
    if (error >= 0)
        return 0;
    else
        return error;
}

/**************************************************************************/

    /*  Set direction = 0 for horozontal distribution, set direction = 1 for
        vertical distribution.  Set space = 0 for start to end spacing, set
        spacing = 1 for fixed gap spacing.
    */
int CVIFUNC_C DistributeCtrls(int panel, int direction, int spacing, int *position, int gap, int numCtrls, ...)
{
    va_list     parmInfo;
    int         error = 0;
    ListType    ctrlList = 0;

    va_start(parmInfo, numCtrls);
    errChk( GetIntListFromParmInfo(parmInfo, numCtrls, &ctrlList));

    errChk( DistributeCtrlsInList(panel, ctrlList, direction, spacing, position, gap));

Error:
    ListDispose(ctrlList);
    va_end(parmInfo);
    if (error >= 0)
        return 0;
    else
        return error;
}

/**************************************************************************/

int CVIFUNC SetCommonDialogShortcutKeys(int panel, int enterCtrl, int escCtrl, int closeCtrl)
{
    int error = 0;

    if (enterCtrl > 0)
        errChk( SetCtrlAttribute(panel, enterCtrl, ATTR_SHORTCUT_KEY, VAL_ENTER_VKEY));

    if (escCtrl > 0)
        errChk( SetCtrlAttribute(panel, escCtrl, ATTR_SHORTCUT_KEY, VAL_ESC_VKEY));

    if (closeCtrl > 0)
        errChk( SetPanelAttribute(panel, ATTR_CLOSE_CTRL, closeCtrl));

Error:
    if (error >= 0)
        return 0;
    else
        return error;
}

/**************************************************************************/

int CVIFUNC SetAttributeToExtremeForList(int panel, ListType ctrlList, int attribute, int whichExtreme)
{
    int     error = 0;
    int     maxValue, minValue, value;

    errChk( AttrMaxAndMinForList(panel, ctrlList, attribute, &maxValue,
                0, &minValue, 0));

    if (whichExtreme == VAL_TB_MAX)
        value = maxValue;
    else
        value = minValue;

    errChk( SetAttributeForList(panel, ctrlList, attribute, value));

Error:
    if (error >= 0)
        return 0;
    else
        return error;
}

/**************************************************************************/

int CVIFUNC_C SetCtrlsToAttributeExtreme(int panel, int attribute,int whichExtreme, int numCtrls, ...)
{
    va_list     parmInfo;
    int         error = 0;
    ListType    ctrlList = 0;

    va_start(parmInfo, numCtrls);

    errChk( GetIntListFromParmInfo(parmInfo, numCtrls, &ctrlList));

    errChk( SetAttributeToExtremeForList(panel, ctrlList, attribute, whichExtreme));

Error:
    ListDispose(ctrlList);
    va_end(parmInfo);

    if (error >= 0)
        return 0;
    else
        return error;
}

/**************************************************************************/

int CVIFUNC GetDataTypeSize(int dataType)
{
    switch (dataType)
        {
        case VAL_CHAR:
        case VAL_UNSIGNED_CHAR:
            return 1;
        case VAL_INTEGER:
        case VAL_UNSIGNED_INTEGER:
        case VAL_FLOAT:
        case VAL_STRING: /* size of a pointer */
            return 4;
        case VAL_UNSIGNED_SHORT_INTEGER:
        case VAL_SHORT_INTEGER:
            return 2;
        case VAL_DOUBLE:
            return 8;
        default:
            AssertMsg(0, "Parameter 1 passed to GetDataTypeSize was not a valid data type.");
            return 0;
        }
}

/**************************************************************************/

double CVIFUNC Pi(void)
{
    return PI;
}

/**************************************************************************/

double CVIFUNC TwoPi(void)
{
    return TWO_PI;
}

/**************************************************************************/

double CVIFUNC HalfPi(void)
{
    return HALF_PI;
}

/**************************************************************************/

double CVIFUNC ThreeHalvesPi(void)
{
    return THREE_HALVES_PI;
}

/**************************************************************************/

double CVIFUNC ReciprocalOfPi(void)
{
    return RECIPROCAL_OF_PI;
}

/**************************************************************************/

double CVIFUNC LnOfPi(void)
{
    return LN_OF_PI;
}

/**************************************************************************/

double CVIFUNC Log10OfPi(void)
{
    return LOG_10_OF_PI;
}

/**************************************************************************/

double CVIFUNC Euler(void)
{
    return EULER;
}

/**************************************************************************/

double CVIFUNC ReciprocalOfEuler(void)
{
    return RECIPROCAL_OF_EULER;
}

/**************************************************************************/

double CVIFUNC Log10OfEuler(void)
{
    return LOG_10_OF_EULER;
}

/**************************************************************************/

double CVIFUNC Ln10(void)
{
    return LN_OF_10;
}

/**************************************************************************/

double CVIFUNC Ln2(void)
{
    return LN_OF_2;
}

/**************************************************************************/

double CVIFUNC PlanckConstant(void)
{
    return PLANCK_CONSTANT;
}

/**************************************************************************/

double CVIFUNC ElementaryCharge(void)
{
    return ELEMENTARY_CHARGE;
}

/**************************************************************************/

double CVIFUNC SpeedOfLight(void)
{
    return SPEED_OF_LIGHT;
}

/**************************************************************************/

double CVIFUNC GravitationalConstant(void)
{
    return GRAVITATIONAL_CONSTANT;
}

/**************************************************************************/

double CVIFUNC AvogadroConstant(void)
{
    return AVOGADRO_CONSTANT;
}

/**************************************************************************/

double CVIFUNC RydbergConstant(void)
{
    return RYDBERG_CONSTANT;
}

/**************************************************************************/

double CVIFUNC MolarGasConstant(void)
{
    return MOLAR_GAS_CONSTANT;
}

/**************************************************************************/

double CVIFUNC DegToRad(double degrees)
{
    return DEG_TO_RAD(degrees);
}

/**************************************************************************/

double CVIFUNC RadToDeg(double radians)
{
    return RAD_TO_DEG(radians);
}

/**************************************************************************/

double CVIFUNC CelsiusToKelvin(double degreesCelsius)
{
    return CELSIUS_TO_KELVIN(degreesCelsius);
}

/**************************************************************************/

double CVIFUNC KelvinToCelsius(double degreesKelvin)
{
    return KELVIN_TO_CELSIUS(degreesKelvin);
}

/**************************************************************************/

double CVIFUNC CelsiusToFahrenheit(double degreesCelsius)
{
    return CELSIUS_TO_FAHRENHEIT(degreesCelsius);
}

/**************************************************************************/

double CVIFUNC FahrenheitToCelsius(double degreesFahrenheit)
{
    return FAHRENHEIT_TO_CELSIUS(degreesFahrenheit);
}

/**************************************************************************/

double CVIFUNC MetersToFeet(double meters)
{
    return METERS_TO_FEET(meters);
}

/**************************************************************************/

double CVIFUNC FeetToMeters(double feet)
{
    return FEET_TO_METERS(feet);
}

/**************************************************************************/

double CVIFUNC KilometersToMiles(double kilometers)
{
    return KILOMETERS_TO_MILES(kilometers);
}

/**************************************************************************/

double CVIFUNC MilesToKilometers(double miles)
{
    return MILES_TO_KILOMETERS(miles);
}

/**************************************************************************/

double CVIFUNC KilogramsToPounds(double kilograms)
{
    return KILOGRAMS_TO_POUNDS(kilograms);
}

/**************************************************************************/

double CVIFUNC PoundsToKilograms(double pounds)
{
    return POUNDS_TO_KILOGRAMS(pounds);
}

/**************************************************************************/

double CVIFUNC LitersToGallons(double liters)
{
    return LITERS_TO_GALLONS(liters);
}

/**************************************************************************/

double CVIFUNC GallonsToLiters(double gallons)
{
    return GALLONS_TO_LITERS(gallons);
}

/**************************************************************************/

    /*** Delay Queues ***/

typedef struct  DelayQueueStructTag *DelayQueue;      /* The DelayQueue abstract data type */

static DelayQueue   CVIFUNC DelayQueueCreate(int panel);
static void         CVIFUNC DelayQueueDispose(DelayQueue delayQueue);
static int          CVIFUNC DelayQueueInsert(DelayQueue delayQueue, DelayedCallbackFunc callback, void *callbackData, double delay);

typedef struct
    {
    DelayedCallbackFunc callback;
    void                *callbackData;
    double              time;
    } DelayQueueEntry;

typedef struct DelayQueueStructTag
    {
    ListType    actionList; /* sorted by latest to soonest expiration times */
    int         panel;
    int         ownsPanel;
    int         timer;
    } DelayQueueStruct;


static int  CVICALLBACK DelayQueueCmpEntriesByTime(void *entry1, void *entry2);
static void CVIFUNC DelayQueueSetNextInterval(DelayQueue delayQueue);
static int  CVICALLBACK DelayQueueHandler(int panelHandle, int controlID, int event, void *callbackData, int eventData1, int eventData2);

/*********************************************************/

static DelayQueue CVIFUNC DelayQueueCreate(int panel)
{
    int         error = 0;
    DelayQueue  newQueue = 0;

    nullChk( newQueue = (DelayQueue)calloc(sizeof(*newQueue), 1));

    if (panel <= 0)
        {
        errChk(panel = NewPanel(0,"", -100, -100, 10, 10));
        newQueue->ownsPanel = TRUE;
        }

    newQueue->panel = panel;

    errChk( newQueue->timer = NewCtrl (panel, CTRL_TIMER, "", 0, 0));
    errChk( InstallCtrlCallback (newQueue->panel, newQueue->timer, DelayQueueHandler, newQueue));
    nullChk( newQueue->actionList = ListCreate(sizeof(DelayQueueEntry)));

Error:
    if (error < 0)
        {
        DelayQueueDispose(newQueue);
        return 0;
        }
    else
        return newQueue;
}

/*********************************************************/

static void CVIFUNC DelayQueueDispose(DelayQueue delayQueue)
{
    if (delayQueue)
        {
        if (delayQueue->panel > 0)
            {
            if (delayQueue->ownsPanel)
                DiscardPanel(delayQueue->panel);
            else
            if (delayQueue->timer > 0)
                DiscardCtrl(delayQueue->panel, delayQueue->timer);
            }

        ListDispose(delayQueue->actionList);

        free(delayQueue);
        }
}

/*********************************************************/

    /* sort in descending order */
static int CVICALLBACK DelayQueueCmpEntriesByTime(void *entry1, void *entry2)
{
    if (((DelayQueueEntry *)entry1)->time > ((DelayQueueEntry *)entry2)->time)
        return -1;
    else
    if (((DelayQueueEntry *)entry1)->time < ((DelayQueueEntry *)entry2)->time)
        return 1;
    else
        return 0;
}

/*********************************************************/

int CVIFUNC PostDelayedCall(DelayedCallbackFunc callback, void *callbackData, double delay)
{
    static  DelayQueue delayQueue = 0;
    int     success = 0;

    if (!delayQueue)
        delayQueue = DelayQueueCreate(0);

    if (delayQueue)
        success = DelayQueueInsert(delayQueue, callback, callbackData, delay);

    return success;
}

/*********************************************************/

static int CVIFUNC DelayQueueInsert(DelayQueue delayQueue, DelayedCallbackFunc callback, void *callbackData, double delay)
{
    DelayQueueEntry entry;
    int             result;

    entry.callback = callback;
    entry.callbackData = callbackData;
    entry.time = Timer() + delay;

    result = ListInsertInOrder(delayQueue->actionList, &entry, DelayQueueCmpEntriesByTime);
    DelayQueueSetNextInterval(delayQueue);

    return result;
}

/*********************************************************/

static void CVIFUNC DelayQueueSetNextInterval(DelayQueue delayQueue)
{
    DelayQueueEntry entry;

    if (ListNumItems(delayQueue->actionList) == 0)
        {
        SetCtrlAttribute(delayQueue->panel, delayQueue->timer, ATTR_ENABLED, 0);
        }
    else
        {
        double  interval;

        ListGetItem(delayQueue->actionList, &entry, END_OF_LIST);

        interval = Max(entry.time - Timer(), 0.0);

        SetCtrlAttribute(delayQueue->panel, delayQueue->timer, ATTR_ENABLED, 1);
        SetCtrlAttribute(delayQueue->panel, delayQueue->timer, ATTR_INTERVAL, interval);
        }
}

/*********************************************************/

static int CVICALLBACK DelayQueueHandler(int panelHandle, int controlID, int event,
            void *callbackData, int eventData1, int eventData2)
{
    DelayQueue      delayQueue = (DelayQueue)callbackData;
    double          time;
    DelayQueueEntry entry;

    UNUSED(panelHandle);    /* avoid warning when compiling with external compiler */
    UNUSED(controlID);      /* avoid warning when compiling with external compiler */
    UNUSED(eventData2);     /* avoid warning when compiling with external compiler */

    switch (event)
        {
        case EVENT_DISCARD:
            delayQueue->timer = 0;
            break;
        case EVENT_TIMER_TICK:
            time = *(double *)eventData1;
            if (ListNumItems(delayQueue->actionList) > 0)
                {
                ListGetItem(delayQueue->actionList, &entry, END_OF_LIST);
                if (entry.time <= time)
                    {
                    ListRemoveItem(delayQueue->actionList, 0, END_OF_LIST);
                    (entry.callback)(entry.callbackData);
                    }

                DelayQueueSetNextInterval(delayQueue);
                }
            break;
        }

    return 0;
}

/*********************************************************/

    /*  Sizing constants for the color change popup */
#define COLOR_CHANGE_MAX_NUM_LIST_BOX_LINES 10
#define COLOR_CHANGE_MIN_NUM_LIST_BOX_LINES 5
#define COLOR_CHANGE_LIST_BOX_WIDTH         550
#define COLOR_CHANGE_SAMPLE_WIDTH           130
#define V_MARGIN                            8
#define H_MARGIN                            8
#define LABEL_H_SEP                         10
#define CTRL_V_SEP                          8
#define MIN_BUTTON_H_SEP         H_MARGIN

typedef struct
    {
    int                     changed;
    int                     panel;
    int                     listBox;
    int                     colorNumeric;
    int                     sampleStart;
    int                     sampleWidth;
    ListType                colorSelectionList;
    void                    *colorChangeCallbackData;
    ShowColorChangeFunction showColorChangeFunc;
    }   ColorChangeDialog;


static int      CVIFUNC FillColorChangeListBox(ColorChangeDialog *colorChangeDialog);
static int      CVICALLBACK ColorListCallback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2);
static void     CVIFUNC GetColorChangeListEntryLabel(ColorChangeEntry *entry, int sampleStart, int sampleWidth, char *labelBuf);
static void     CVIFUNC UpdateColorNumeric(ColorChangeDialog *colorChangeDialog);
static void     CVIFUNC SetColorsToDefaults(ListType colorSelectionList);
static int      CVIFUNC UpdateColorListBoxEntry(ColorChangeDialog *colorChangeDialog, int listIndex);
static int      CVIFUNC UpdateAllColorListBoxEntries(ColorChangeDialog *colorChangeDialog);
static int      CVICALLBACK ColorNumericCallback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2);

static int      CVIFUNC ConformBitmapDataToSystem (unsigned char *bitmap, int *colorTable, 
                                           int depth, int rowBytes, int width, 
                                           int height, int oldPanelColor, 
                                           int newPanelColor, int oldTextColor, 
                                           int newTextColor);
static int      CVIFUNC ConformOneBitmapToPanel (int panel, int control, int index,
                                         int bitmap3dColor, int new3dColor,
                                         int bitmapTextColor, int newTextColor);
static int      CVIFUNC ConformOneCtrlBitmapToPanel (int panel, int control, int index,
                                             int bitmap3dColor, int new3dColor,
                                             int bitmapTextColor, int newTextColor);

/**********************************************************************************/

static void CVIFUNC GetColorChangeListEntryLabel(ColorChangeEntry *entry, int sampleStart, int sampleWidth, char *labelBuf)
{
    sprintf(labelBuf, "%s\033p%dl\033fgCCCCCC\033drawfilledbox001CCCCCC\033fg%6.6x\033drawfilledbox%3.3d%6.6x", 
            entry->name, sampleStart, VAL_WHITE, sampleWidth, entry->color);
}

/**************************************************************************************/

static int CVIFUNC FillColorChangeListBox(ColorChangeDialog *colorChangeDialog)
{
    int                 error = 0;
    int                 index;
    int                 numItems;
    char                buf[512];
    ColorChangeEntry *entry;

    numItems = ListNumItems(colorChangeDialog->colorSelectionList);

    for (index = 1; index <= numItems; index++)
        {
        entry = (ColorChangeEntry *)ListGetPtrToItem(colorChangeDialog->colorSelectionList, index);
        GetColorChangeListEntryLabel(entry, colorChangeDialog->sampleStart, colorChangeDialog->sampleWidth, buf);
        errChk( InsertListItem(colorChangeDialog->panel, colorChangeDialog->listBox, -1, (char *)buf, index));
        }

Error:
    return error;
}

/**************************************************************************************/

static void CVIFUNC SetColorsToDefaults(ListType colorSelectionList)
{
    int                 index;
    int                 numItems;
    ColorChangeEntry *entry;

    numItems = ListNumItems(colorSelectionList);

    for (index = 1; index <= numItems; index++)
        {
        entry = (ColorChangeEntry *)ListGetPtrToItem(colorSelectionList, index);
        entry->color = entry->defaultColor;
        }
}

/*********************************************************************/

static void CVIFUNC UpdateColorNumeric(ColorChangeDialog *colorChangeDialog)
{
    int                 index;
    ColorChangeEntry *entry;

    GetCtrlVal(colorChangeDialog->panel, colorChangeDialog->listBox, &index);
    entry = (ColorChangeEntry *)ListGetPtrToItem(colorChangeDialog->colorSelectionList, index);
    SetCtrlVal(colorChangeDialog->panel, colorChangeDialog->colorNumeric, entry->color);
}

/*********************************************************************/

static int CVICALLBACK ColorListCallback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    UNUSED(panel);          /* avoid warning when compiling with external compiler */
    UNUSED(control);        /* avoid warning when compiling with external compiler */
    UNUSED(eventData1);     /* avoid warning when compiling with external compiler */
    UNUSED(eventData2);     /* avoid warning when compiling with external compiler */

    switch (event)
        {
        case EVENT_VAL_CHANGED:
            UpdateColorNumeric((ColorChangeDialog *)callbackData);
            break;
        }

    return 0;
}

/*********************************************************************/

static int CVIFUNC UpdateColorListBoxEntry(ColorChangeDialog *colorChangeDialog, int listIndex)
{
    int                 error = 0;
    char                buf[512];
    ColorChangeEntry    *entry;
    int                 colorIndex;

    GetValueFromIndex(colorChangeDialog->panel, colorChangeDialog->listBox, listIndex, &colorIndex);
    entry = (ColorChangeEntry *)ListGetPtrToItem(colorChangeDialog->colorSelectionList, colorIndex);
    GetColorChangeListEntryLabel(entry, colorChangeDialog->sampleStart, colorChangeDialog->sampleWidth, buf);
    errChk( ReplaceListItem(colorChangeDialog->panel, colorChangeDialog->listBox, listIndex, (char *)buf, colorIndex));

Error:
    return error;
}

/*********************************************************************/

static int CVIFUNC UpdateAllColorListBoxEntries(ColorChangeDialog *colorChangeDialog)
{
    int error = 0;
    int colorIndex;
    int numItems;

    numItems = ListNumItems(colorChangeDialog->colorSelectionList);

    for (colorIndex = 1; colorIndex <= numItems; colorIndex++)
        errChk( UpdateColorListBoxEntry(colorChangeDialog, colorIndex - 1));

Error:
    return error;
}

/*********************************************************************/

static int CVICALLBACK ColorNumericCallback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    int                 error = 0;
    int                 listIndex, colorIndex;
    int                 color;
    ColorChangeDialog   *colorChangeDialog = (ColorChangeDialog *)callbackData;
    ColorChangeEntry    *entry;


    UNUSED(eventData1);     /* avoid warning when compiling with external compiler */
    UNUSED(eventData2);     /* avoid warning when compiling with external compiler */

    switch (event)
        {
        case EVENT_VAL_CHANGED:
            GetCtrlVal(panel, control, &color);
            GetCtrlIndex(panel, colorChangeDialog->listBox, &listIndex);
            GetValueFromIndex(panel, colorChangeDialog->listBox, listIndex, &colorIndex);
            entry = (ColorChangeEntry *)ListGetPtrToItem(colorChangeDialog->colorSelectionList, colorIndex);
            entry->color = color;
            errChk(UpdateColorListBoxEntry(colorChangeDialog, listIndex));
            break;
        case EVENT_COMMIT:
            colorChangeDialog->changed = TRUE;
            GetCtrlIndex(panel, colorChangeDialog->listBox, &listIndex);
            if (listIndex > -1)
                {
                if (colorChangeDialog->showColorChangeFunc)
                    colorChangeDialog->showColorChangeFunc(colorChangeDialog->colorSelectionList, colorChangeDialog->colorChangeCallbackData, listIndex + 1);
                }
            break;
        }

    return 0;

Error:
    QueueUserEvent(error, 0, 0);
    return 0;
}

/*********************************************************************/

int CVIFUNC ColorChangePopup(char *title, ListType colorList, int allowDefault, ShowColorChangeFunction showColorChangeFunc, void *colorChangeCallbackData)
{
    int             error = 0;
    int             panel = 0;
    int             okButton, cancelButton, listBox, colorHeadingTxtMsg, sampleHeadingTxtMsg;
    int             colorNumeric, defaultButton;
    int             listBoxWidth = COLOR_CHANGE_LIST_BOX_WIDTH;
    int             sampleWidth = COLOR_CHANGE_SAMPLE_WIDTH;
    int             top, height, width;
    int             leftMargin, rightMargin;
    ColorChangeDialog   colorChangeDialog;
    int             done = FALSE;
    int             evtPanel, evtCtrl;
    int             numListLines;
    ListType        colorListCopy = 0;

    nullChk( colorListCopy = ListCopy(colorList));

    top = V_MARGIN;
    leftMargin = H_MARGIN;
    rightMargin = listBoxWidth + H_MARGIN;

    errChk( panel =                 NewPanel(0, title, 30,30, 200,200));
    errChk( SetPanelAttribute(panel, ATTR_CONFORM_TO_SYSTEM, TRUE));
    errChk( okButton =              NewCtrl(panel, CTRL_SQUARE_COMMAND_BUTTON, "__OK", 10, 10));
    errChk( cancelButton =          NewCtrl(panel, CTRL_SQUARE_COMMAND_BUTTON, "  __Cancel  ", 10, 10));
    errChk( defaultButton =         NewCtrl(panel, CTRL_SQUARE_COMMAND_BUTTON, "__Default", 10, 10));
    errChk( colorNumeric =          NewCtrl(panel, CTRL_COLOR_NUMERIC, "Color:" , 0, 0));
    errChk( colorHeadingTxtMsg =    NewCtrl(panel, CTRL_TEXT_MSG, "Color Name:", top, leftMargin));
    errChk( sampleHeadingTxtMsg =   NewCtrl(panel, CTRL_TEXT_MSG, "Sample:", top, 0));
    errChk( listBox =               NewCtrl(panel, CTRL_LIST, 0, 0,H_MARGIN));

        /* setup listbox heading text messages */
    SetCtrlAttribute(panel, sampleHeadingTxtMsg, ATTR_LEFT, rightMargin - sampleWidth);
    GetCtrlAttribute(panel, colorHeadingTxtMsg, ATTR_HEIGHT, &height);
    top += height + CTRL_V_SEP / 2;

        /* setup listbox */
    InstallCtrlCallback(panel, listBox, ColorListCallback, &colorChangeDialog);
    numListLines = (int)Pin(ListNumItems(colorList), COLOR_CHANGE_MIN_NUM_LIST_BOX_LINES, COLOR_CHANGE_MAX_NUM_LIST_BOX_LINES);
    SetCtrlAttribute(panel, listBox, ATTR_VISIBLE_LINES, numListLines);
    SetCtrlAttribute(panel, listBox, ATTR_TOP, top);
    SetCtrlAttribute(panel, listBox, ATTR_WIDTH, listBoxWidth);
    GetCtrlAttribute(panel, listBox, ATTR_HEIGHT, &height);
    top += height + CTRL_V_SEP;

        /* setup color numeric */
    SetCtrlAttribute(panel, colorNumeric, ATTR_TOP, top);
    SetCtrlAttribute(panel, colorNumeric, ATTR_LABEL_TOP, top);
    GetCtrlAttribute(panel,colorNumeric, ATTR_HEIGHT, &height);
    GetCtrlAttribute(panel,colorNumeric, ATTR_LABEL_WIDTH, &width);
    SetCtrlAttribute(panel,colorNumeric, ATTR_LEFT, width + leftMargin + LABEL_H_SEP);
    SetCtrlAttribute(panel,colorNumeric, ATTR_LABEL_LEFT, leftMargin);
    SetCtrlAttribute(panel, colorNumeric, ATTR_LABEL_TOP, VAL_AUTO_CENTER);
    SetCtrlAttribute(panel, colorNumeric, ATTR_SHOW_TRANSPARENT, FALSE);
    InstallCtrlCallback(panel, colorNumeric, ColorNumericCallback, &colorChangeDialog);

        /* position ok, cancel, and default buttons */
    SetCtrlsToAttributeExtreme (panel, ATTR_WIDTH, VAL_TB_MAX, 3, okButton, cancelButton, defaultButton);
    SetAttributeForCtrls (panel, ATTR_TOP, top, 3, okButton, cancelButton, defaultButton);

    GetCtrlAttribute(panel, okButton, ATTR_WIDTH, &width);  /* buttons are now the same width */

    SetCtrlAttribute(panel,cancelButton, ATTR_LEFT, rightMargin - width);
    SetCtrlAttribute(panel,okButton, ATTR_LEFT, rightMargin - width * 2 - 2 * MIN_BUTTON_H_SEP);
    SetCtrlAttribute(panel,defaultButton, ATTR_LEFT, rightMargin - width * 3 - 4 * MIN_BUTTON_H_SEP);

    GetCtrlAttribute(panel,okButton, ATTR_HEIGHT, &height);
    top += CTRL_V_SEP + height;

        /* position panel */
    top += V_MARGIN;
    SetPanelAttribute(panel, ATTR_HEIGHT, top);
    SetPanelAttribute(panel, ATTR_WIDTH, listBoxWidth + 2 * H_MARGIN);
    SetPanelPos(panel, VAL_AUTO_CENTER, VAL_AUTO_CENTER);
    SetCommonDialogShortcutKeys (panel, okButton, cancelButton, cancelButton);

    if (!allowDefault)
        SetCtrlAttribute(panel, defaultButton, ATTR_VISIBLE, FALSE);

    colorChangeDialog.changed = FALSE;
    colorChangeDialog.panel = panel;
    colorChangeDialog.listBox = listBox;
    colorChangeDialog.colorNumeric = colorNumeric;
    colorChangeDialog.sampleStart = listBoxWidth - LIST_BOX_LEFT_PIXEL_POS - sampleWidth;
    colorChangeDialog.sampleWidth = listBoxWidth - colorChangeDialog.sampleStart - LIST_BOX_FRAME_WIDTH - LIST_BOX_RIGHT_MARGIN - LIST_BOX_LEFT_PIXEL_POS - LIST_BOX_SCROLL_BAR_WIDTH;
    colorChangeDialog.showColorChangeFunc = showColorChangeFunc;
    colorChangeDialog.colorChangeCallbackData = colorChangeCallbackData;
    colorChangeDialog.colorSelectionList = colorList;

    errChk(FillColorChangeListBox(&colorChangeDialog));

    SetActiveCtrl(panel, listBox);
    UpdateColorNumeric(&colorChangeDialog);
    InstallPopup(panel);

    while (!done)
        {
        errChk( GetUserEvent(0, &evtPanel, &evtCtrl));

        if (evtPanel == panel)
            {
            if (evtCtrl == cancelButton)
                {
                if (colorChangeDialog.changed)
                    {
                    ListClear(colorList);
                    nullChk( ListAppend (colorList, colorListCopy));
                    if (showColorChangeFunc)
                        showColorChangeFunc(colorChangeDialog.colorSelectionList, colorChangeDialog.colorChangeCallbackData, -1);
                    }

                colorChangeDialog.changed = FALSE;
                done = TRUE;
                }
            else
            if (evtCtrl == okButton)
                {
                done = TRUE;
                }
            else
            if (evtCtrl == defaultButton)
                {
                colorChangeDialog.changed = TRUE;
                SetColorsToDefaults(colorChangeDialog.colorSelectionList);
                errChk( UpdateAllColorListBoxEntries(&colorChangeDialog));
                UpdateColorNumeric(&colorChangeDialog);
                if (showColorChangeFunc)
                    showColorChangeFunc(colorChangeDialog.colorSelectionList, colorChangeDialog.colorChangeCallbackData, -1);
                }
            }
        }

Error:
    if (panel > 0)
        DiscardPanel(panel);

    ListDispose(colorListCopy);
    if (error < 0)
        return error;
    else
        return colorChangeDialog.changed;
}

/******************************************************************/

int CVIFUNC GetNumericAttributeAsType(int panel, int ctrl, int attribute, int destType, void *value)
{
    char            charVal;
    unsigned char   uCharVal;
    int             intVal;
    unsigned int    uIntVal;
    short           shortIntVal;
    unsigned short  uShortIntVal;
    float           floatVal;
    double          doubleVal;
    int             currType;
    int             error = 0;

    errChk( GetCtrlAttribute(panel, ctrl, ATTR_DATA_TYPE, &currType));

    switch (currType)
        {
        case VAL_CHAR:
            errChk( GetCtrlAttribute(panel, ctrl, attribute, &charVal));
            doubleVal = (double)charVal;
            break;
        case VAL_UNSIGNED_CHAR:
            errChk( GetCtrlAttribute(panel, ctrl, attribute, &uCharVal));
            doubleVal = (double)uCharVal;
            break;
        case VAL_INTEGER:
            errChk( GetCtrlAttribute(panel, ctrl, attribute, &intVal));
            doubleVal = (double)intVal;
            break;
        case VAL_UNSIGNED_INTEGER:
            errChk( GetCtrlAttribute(panel, ctrl, attribute, &uIntVal));
            doubleVal = (double)uIntVal;
            break;
        case VAL_FLOAT:
            errChk( GetCtrlAttribute(panel, ctrl, attribute, &floatVal));
            doubleVal = (double)floatVal;
            break;
        case VAL_UNSIGNED_SHORT_INTEGER:
            errChk( GetCtrlAttribute(panel, ctrl, attribute, &uShortIntVal));
            doubleVal = (double)uShortIntVal;
            break;
        case VAL_SHORT_INTEGER:
            errChk( GetCtrlAttribute(panel, ctrl, attribute, &shortIntVal));
            doubleVal = (double)shortIntVal;
            break;
        case VAL_DOUBLE:
            errChk( GetCtrlAttribute(panel, ctrl, attribute, &doubleVal));
            break;
        default:
            AssertMsg(0, "The type of the control passed to GetNumericAttributeAsType\nwas not a numeric type.");
            break;;
        }

    switch (destType)
        {
        case VAL_CHAR:
            *(char *)value = (char)doubleVal;
            break;
        case VAL_UNSIGNED_CHAR:
            *(unsigned char *)value = (unsigned char)doubleVal;
            break;
        case VAL_INTEGER:
            *(int *)value = (int)doubleVal;
            break;
        case VAL_UNSIGNED_INTEGER:
            *(unsigned int *)value = (unsigned int)doubleVal;
            break;
        case VAL_FLOAT:
            *(float *)value = (float)doubleVal;
            break;
        case VAL_UNSIGNED_SHORT_INTEGER:
            *(unsigned short *)value = (unsigned short)doubleVal;
            break;
        case VAL_SHORT_INTEGER:
            *(short *)value = (short)doubleVal;
            break;
        case VAL_DOUBLE:
            *(double *)value = doubleVal;
        default:
            AssertMsg(0, "Parameter 4 passed to GetNumericAttributeAsType\nwas not a numeric data type.");
            break;;
        }

Error:
    return 0;
}

/*********************************************************/

int CVIFUNC MoveBehind(int panel, int ctrl, int ctrlToMove)
{
    int zOfCtrl;
    int zOfCtrlToMove;
    int error = 0;

    errChk( GetCtrlAttribute(panel, ctrl, ATTR_ZPLANE_POSITION, &zOfCtrl));
    errChk( GetCtrlAttribute(panel, ctrlToMove, ATTR_ZPLANE_POSITION, &zOfCtrlToMove));

    if (zOfCtrl > zOfCtrlToMove)
        errChk( SetCtrlAttribute(panel, ctrlToMove, ATTR_ZPLANE_POSITION, zOfCtrl));
    else
        errChk( SetCtrlAttribute(panel, ctrlToMove, ATTR_ZPLANE_POSITION, zOfCtrl + 1));

Error:
    if (error < 0)
        return error;
    else
        return 0;
}

/**********************************************************/

int CVIFUNC MoveInFront(int panel, int ctrl, int ctrlToMove)
{
    int zOfCtrl;
    int zOfCtrlToMove;
    int error = 0;

    errChk( GetCtrlAttribute(panel, ctrl, ATTR_ZPLANE_POSITION, &zOfCtrl));
    errChk( GetCtrlAttribute(panel, ctrlToMove, ATTR_ZPLANE_POSITION, &zOfCtrlToMove));

    if (zOfCtrl > zOfCtrlToMove)
        errChk( SetCtrlAttribute(panel, ctrlToMove, ATTR_ZPLANE_POSITION, zOfCtrl-1));
    else
        errChk( SetCtrlAttribute(panel, ctrlToMove, ATTR_ZPLANE_POSITION, zOfCtrl));

Error:
    if (error < 0)
        return error;
    else
        return 0;
}

/*********************************************************/

    /*  places ctrlToMove after ctrl in the tab order */
int CVIFUNC PlaceAfter(int panel, int ctrl, int ctrlToMove)
{
    int tabPosOfCtrl;
    int tabPosOfCtrlToMove;
    int error = 0;

    errChk( GetCtrlAttribute(panel, ctrl, ATTR_CTRL_TAB_POSITION, &tabPosOfCtrl));
    errChk( GetCtrlAttribute(panel, ctrlToMove, ATTR_CTRL_TAB_POSITION, &tabPosOfCtrlToMove));

    if (tabPosOfCtrl > tabPosOfCtrlToMove)
        errChk( SetCtrlAttribute(panel, ctrlToMove, ATTR_CTRL_TAB_POSITION, tabPosOfCtrl));
    else
        errChk( SetCtrlAttribute(panel, ctrlToMove, ATTR_CTRL_TAB_POSITION, tabPosOfCtrl + 1));

Error:
    if (error < 0)
        return error;
    else
        return 0;
}

/**********************************************************/

    /*  places ctrlToMove before ctrl in the tab order */
int CVIFUNC PlaceBefore(int panel, int ctrl, int ctrlToMove)
{
    int tabPosOfCtrl;
    int tabPosOfCtrlToMove;
    int error = 0;

    errChk( GetCtrlAttribute(panel, ctrl, ATTR_CTRL_TAB_POSITION, &tabPosOfCtrl));
    errChk( GetCtrlAttribute(panel, ctrlToMove, ATTR_CTRL_TAB_POSITION, &tabPosOfCtrlToMove));

    if (tabPosOfCtrl > tabPosOfCtrlToMove)
        errChk( SetCtrlAttribute(panel, ctrlToMove, ATTR_CTRL_TAB_POSITION, tabPosOfCtrl-1));
    else
        errChk( SetCtrlAttribute(panel, ctrlToMove, ATTR_CTRL_TAB_POSITION, tabPosOfCtrl));

Error:
    if (error < 0)
        return error;
    else
        return 0;
}

/***************************************************************************************/

void CVIFUNC SetCtrlRightPosition(int panel, int control, int right)
{
    int width;

    GetCtrlAttribute(panel, control, ATTR_WIDTH, &width);
    SetCtrlAttribute(panel, control, ATTR_LEFT, right - width);
}

/*****************************************************************************************/

int CVIFUNC CenterCtrl(int panel, int ctrl, int top, int left, int height, int width)
{
    int error = 0;
    int cTop, cLeft;        /*  control top, control left */
    int bTop, bLeft;        /*  control bounds top and left */
    int bHeight, bWidth;    /*  control bounds height and width */
    int newBTop, newBLeft;  /*  new control bounds top and left */
    int newTop, newLeft;    /*  new control top and left */

    errChk( GetCtrlBoundingRect(panel, ctrl, &bTop, &bLeft, &bHeight, &bWidth));
    errChk( GetCtrlAttribute(panel, ctrl, ATTR_TOP, &cTop));
    errChk( GetCtrlAttribute(panel, ctrl, ATTR_LEFT, &cLeft));

    newBTop = (height - bHeight) / 2 + top;
    newBLeft = (width - bWidth) / 2 + left;

    newTop = cTop + (newBTop - bTop);
    newLeft = cLeft + (newBLeft - bLeft);

    errChk( SetCtrlAttribute(panel, ctrl, ATTR_TOP, newTop));
    errChk( SetCtrlAttribute(panel, ctrl, ATTR_LEFT, newLeft));

Error:
    return 0;
}

/**********************************************************/

int CVIFUNC SizeRingCtrlToText(int panel, int ring)
{
    int     error  = 0;
    int     index, numItems, maxWidth = 0, width;
    char    *fontName = 0;
    int     fontNameLength;
    int     italic, underLine, strikeOut, pointSize, bold;
    int     style;
    char    label[256];
    int     extraRingWidth;

    errChk( GetCtrlAttribute(panel, ring, ATTR_CTRL_STYLE, &style));

    switch (style)
        {
        case CTRL_RING:
            extraRingWidth = 19;
            break;
        case CTRL_RECESSED_MENU_RING:
            extraRingWidth = 21;
            break;
        default:
            errChk( UIEInvalidControlType);
        }

    errChk( GetCtrlAttribute(panel, ring, ATTR_TEXT_FONT_NAME_LENGTH, &fontNameLength));

    nullChk( fontName = (char *)malloc(fontNameLength + 1));

    errChk( GetCtrlAttribute(panel, ring, ATTR_TEXT_FONT, fontName));
    errChk( GetCtrlAttribute(panel, ring, ATTR_TEXT_ITALIC, &italic));
    errChk( GetCtrlAttribute(panel, ring, ATTR_TEXT_UNDERLINE, &underLine));
    errChk( GetCtrlAttribute(panel, ring, ATTR_TEXT_STRIKEOUT, &strikeOut));
    errChk( GetCtrlAttribute(panel, ring, ATTR_TEXT_POINT_SIZE, &pointSize));
    errChk( GetCtrlAttribute(panel, ring, ATTR_TEXT_BOLD, &bold));

    errChk( CreateMetaFont ("___RingMetaFont", fontName, pointSize, bold, italic, underLine, strikeOut));

    errChk( GetNumListItems (panel, ring, &numItems));

    for (index = 0; index < numItems; index++)
        {
        errChk( GetLabelFromIndex (panel, ring, index, label));
        errChk( GetTextDisplaySize (label, "___RingMetaFont", 0, &width));

        maxWidth = Max(maxWidth, width);
        }

    errChk( SetCtrlAttribute (panel, ring, ATTR_WIDTH, maxWidth + extraRingWidth));

        /*  Since we don't want the user to get back ___RingMetaFont when querying the fontname of their
            controls, we change this metafont to an unreasonable font that no one would ever user (since
            there isn't an explicit DeleteMetaFont() function.
        */
    errChk( CreateMetaFont ("___RingMetaFont", "bogusfont", 6, TRUE, TRUE, TRUE, TRUE));

Error:
    free(fontName);
    return error;
}

/*******************************************************************************/


#define MAX_COLOR_DIFF      15

#define RedComponent(c)     (((c) >> 16) & 0xff)
#define GreenComponent(c)   (((c) >> 8) & 0xff)
#define BlueComponent(c)    ((c) & 0xff)

#define ComputeColorDiff(c1,c2) (abs(RedComponent(c1) - RedComponent(c2)) + \
                                 abs(GreenComponent(c1) - GreenComponent(c2)) + \
                                 abs(BlueComponent(c1) - BlueComponent(c2)))

static int CVIFUNC ConformBitmapDataToSystem (unsigned char *bitmap, int *colorTable, 
                                      int depth, int rowBytes, int width, 
                                      int height, int oldPanelColor, 
                                      int newPanelColor, int oldTextColor, 
                                      int newTextColor)
{
    int error = UIENoError;
    int i, j, index, numColors, color, newColor, srcByte;
    int oldPanelHilite, newPanelHilite, oldPanelShadow, newPanelShadow;
    const int maxDiff = 15;
    
    switch (depth) 
    {
        case 1:
            numColors = 2;
            break;
        case 4:
            numColors = 16;
            break;
        case 8:
            numColors = 256;
            break;
        case 24:
            numColors = 0;
            break;
    }

    errChk (Get3dBorderColors (oldPanelColor, &oldPanelHilite, NULL, 
                               &oldPanelShadow, NULL));
    errChk (Get3dBorderColors (newPanelColor, &newPanelHilite, NULL, 
                               &newPanelShadow, NULL));
    
    if (numColors > 0)
    {
        for (index = 0; index < numColors; index++) 
        {
            newColor = color = *(colorTable + index);
            if (ComputeColorDiff (color, oldPanelColor) <= MAX_COLOR_DIFF)
                newColor = newPanelColor;
            if (ComputeColorDiff (color, oldPanelHilite) <= MAX_COLOR_DIFF)
                newColor = newPanelHilite;
            if (ComputeColorDiff (color, oldPanelShadow) <= MAX_COLOR_DIFF)
                newColor = newPanelShadow;
            if (ComputeColorDiff (color, oldTextColor) <= MAX_COLOR_DIFF)
                newColor = newTextColor;
            if (newColor != color)
                *(colorTable + index) = newColor;
        }
    }
    else                                    /* depth == 24, no colorTable */
    {   
        for (i = 0; i < height ; i++)
        {
            for (j = 0; j < width; j++) 
            {
                srcByte = 3 * j + i * rowBytes;
                newColor = color = MakeColor (bitmap[srcByte],
                                              bitmap[srcByte+1],
                                              bitmap[srcByte+2]);
                if (ComputeColorDiff (color, oldPanelColor) <= MAX_COLOR_DIFF)
                    newColor = newPanelColor;
                if (ComputeColorDiff (color, oldPanelHilite) <= MAX_COLOR_DIFF)
                    newColor = newPanelHilite;
                if (ComputeColorDiff (color, oldPanelShadow) <= MAX_COLOR_DIFF)
                    newColor = newPanelShadow;
                if (ComputeColorDiff (color, oldTextColor) <= MAX_COLOR_DIFF)
                    newColor = newTextColor;
                if (newColor != color) 
                {
                    bitmap[srcByte] = RedComponent(newColor);
                    bitmap[srcByte+1] = GreenComponent(newColor);
                    bitmap[srcByte+2] = BlueComponent(newColor);
                }
            }
        }
    }

Error:
    return error;
}


static int CVIFUNC ConformOneBitmapToPanel (int panel, int control, int index,
                                    int bitmap3dColor, int new3dColor,
                                    int bitmapTextColor, int newTextColor)
{
    int error = UIENoError;
    int *colorTable = NULL, rowBytes, depth, width, height,
        colorTableSize, bitmapSize, maskSize;
    unsigned char *bitmap = NULL, *mask = NULL;

    errChk (GetImageInfo (panel, control, index, &colorTableSize, 
                          &bitmapSize, &maskSize));
    if (colorTableSize)
        nullChk (colorTable = malloc (colorTableSize));
    nullChk (bitmap = malloc (bitmapSize));
    if (maskSize)
        nullChk (mask = malloc (maskSize));
    errChk (GetImageBits (panel, control, index, &rowBytes, &depth, 
                          &width, &height, colorTable, bitmap, mask));

    errChk (ConformBitmapDataToSystem (bitmap, colorTable, depth, rowBytes, 
                                       width, height, bitmap3dColor, new3dColor,
                                       bitmapTextColor, newTextColor));

    errChk (SetImageBits (panel, control, index, rowBytes, depth, 
                          width, height, colorTable, bitmap, mask));

Error :
    if (colorTable)
        free (colorTable);
    if (bitmap)
        free (bitmap);
    if (mask)
        free (mask);
        
    return error;
}


static int CVIFUNC ConformOneCtrlBitmapToPanel (int panel, int control, int index,
                                        int bitmap3dColor, int new3dColor,
                                        int bitmapTextColor, int newTextColor)
{
    int error = UIENoError;
    int ctrlStyle;
    int numItems, i;
    
    errChk (GetCtrlAttribute (panel, control, ATTR_CTRL_STYLE, &ctrlStyle));
    switch (ctrlStyle)
    {
        case CTRL_PICTURE_COMMAND_BUTTON:
        case CTRL_PICTURE_TOGGLE_BUTTON:
        case CTRL_PICTURE:
        case CTRL_PICTURE_RING:
            break;
        default:
            errChk (UIEInvalidControlType);
            break;
    }
    
    if ((ctrlStyle == CTRL_PICTURE_RING) && (index == -1))
    {
        errChk (GetNumListItems (panel, control, &numItems));
        for (i = 0 ; i < numItems ; i++)
        {
            errChk (ConformOneBitmapToPanel (panel, control, i, 
                                             bitmap3dColor, new3dColor, 
                                             bitmapTextColor, newTextColor));
        }
    }
    else
    {
        errChk (ConformOneBitmapToPanel (panel, control, 0, 
                                         bitmap3dColor, new3dColor, 
                                         bitmapTextColor, newTextColor));
    }
    
Error:
    return error;
}


int CVIFUNC ConformCtrlBitmapToPanel (int panel, int control, int index,
                              int bitmap3dColor, int new3dColor,
                              int bitmapTextColor, int newTextColor)
{
    int error = UIENoError;
    int nextCtrl;
    int actual3dColor = new3dColor, actualTextColor = newTextColor;
    
    if (actual3dColor == -1)
    {
        errChk (GetPanelAttribute (panel, ATTR_BACKCOLOR, &actual3dColor));
    }
    
    if (actualTextColor == -1)
    {
        errChk (GetPanelAttribute (panel, ATTR_TITLE_COLOR, &actualTextColor));
    }
    
    if (control == -1)
    {
        errChk (GetPanelAttribute (panel, ATTR_PANEL_FIRST_CTRL, &nextCtrl));
        
        while (nextCtrl != 0)
        {
            errChk (ConformOneCtrlBitmapToPanel (panel, nextCtrl, index, 
                                                 bitmap3dColor, actual3dColor, 
                                                 bitmapTextColor, actualTextColor));
            errChk (GetCtrlAttribute (panel, nextCtrl, ATTR_NEXT_CTRL, 
                                      &nextCtrl));
        }                                                
    }
    else
    {
        errChk (ConformOneCtrlBitmapToPanel (panel, control, index, 
                                             bitmap3dColor, actual3dColor, 
                                             bitmapTextColor, actualTextColor));
    }
    
Error:
    return error;
}

/*******************************************************************************/

#define PROGRESS_BOX_MARGIN         14
#define PROGRESS_BAR_HEIGHT         14
#define PROGRESS_BAR_WIDTH          300
#define PROGRESS_LABEL_CTRL_V_GAP   4

typedef struct
    {
    int panel;
    int progressIndicator;
    int cancelRequested;
    int cancelButton;
    } *ProgressDlg;


void CVIFUNC DiscardProgressDialog(int panel)
{
    int         error = 0;
    ProgressDlg progressDlg = 0;
    int         savedBole;

    savedBole = SetBOLE(FALSE);

    if (panel > 0)
        {
        errChk( GetPanelAttribute(panel, ATTR_CALLBACK_DATA, &progressDlg));
        free(progressDlg);
        DiscardPanel(panel);
        }

Error:
    SetBOLE(savedBole);
}

/*******************************************************************************/

    /*  returns TRUE if the user has tried to cancel the operation */
int CVIFUNC UpdateProgressDialog(int panel, int percentDone, int processEvents)
{
    int         error = 0;
    ProgressDlg progressDlg = 0;
    int         savedBole;
    int         cancelRequested = FALSE;

    savedBole = SetBOLE(FALSE);

    if (panel > 0)
        {
        errChk( GetPanelAttribute(panel, ATTR_CALLBACK_DATA, &progressDlg));

        if (progressDlg)
            {
            percentDone = (int)Pin(percentDone, 0, 100);
            errChk( SetCtrlVal(progressDlg->panel, progressDlg->progressIndicator, percentDone));
            if (processEvents)
                ProcessSystemEvents();
            cancelRequested = progressDlg->cancelRequested;
            if (cancelRequested)
                progressDlg->cancelRequested = FALSE;
            }
        }

Error:
    SetBOLE(savedBole);
    if (error < 0)
        return error;
    else
        return cancelRequested;
}

/*******************************************************************************/

static int CVICALLBACK ProgressCancelCallback(int panel, int ctrl, int event, void *callbackData, int eventData1, int eventData2)
{
    ProgressDlg progressDlg = (ProgressDlg)callbackData;

    UNUSED(panel);
    UNUSED(ctrl);
    UNUSED(eventData1);
    UNUSED(eventData2);

    switch (event)
        {
        case EVENT_COMMIT:
            if (progressDlg)
                progressDlg->cancelRequested = TRUE;
            break;
        }

    return 0;
}

/*******************************************************************************/

    /*  pass 0 for cancelButtonLabel if you don't wan't the dialog to be cancelable.
        pass "" for cancelButtonLabel if you want the dialog to recognized the ESC key but not
        display a cancel button.
    */
int CVIFUNC CreateProgressDialog(char title[], char indicatorLabel[], int modal, int indicatorMarkerStyle, char cancelButtonLabel[])
{
    int         error = 0;
    int         panel = 0;
    ProgressDlg progressDlg = 0;
    int         ctrl;
    int         cancelBtn;
    int         buttonHeight;
    int         buttonWidth;
    int         dlgWidth;
    int         labelHeight;
    int         top, left, height, width;
    int         dlgHeightWithoutCancel;

    dlgWidth = PROGRESS_BAR_WIDTH + 2 * PROGRESS_BOX_MARGIN;

    nullChk( progressDlg = (ProgressDlg)calloc(1, sizeof(*progressDlg)));

    errChk( panel = progressDlg->panel = NewPanel (0, title, 0, 0, 10, dlgWidth));

    errChk( SetPanelAttribute(panel, ATTR_CONFORM_TO_SYSTEM, TRUE));
    errChk( SetPanelAttribute (panel, ATTR_SYSTEM_MENU_VISIBLE, FALSE));
    errChk( SetPanelAttribute(panel, ATTR_CALLBACK_DATA, (void *)progressDlg));

    errChk( ctrl = progressDlg->progressIndicator = NewCtrl(panel, CTRL_NUMERIC_LEVEL_HSLIDE, indicatorLabel, 0, PROGRESS_BOX_MARGIN));

    if (cancelButtonLabel)
        {
        errChk( cancelBtn = progressDlg->cancelButton = NewCtrl(panel, CTRL_SQUARE_COMMAND_BUTTON, cancelButtonLabel, -100, -100));
        errChk( InstallCtrlCallback (panel, cancelBtn, ProgressCancelCallback, (void *)progressDlg));
        SetCtrlAttribute(panel, cancelBtn, ATTR_SHORTCUT_KEY, VAL_ESC_VKEY);
        }

    SetCtrlAttribute(panel, ctrl, ATTR_WIDTH, PROGRESS_BAR_WIDTH);
    SetCtrlAttribute(panel, ctrl, ATTR_HEIGHT, PROGRESS_BAR_HEIGHT);
    SetCtrlAttribute(panel, ctrl, ATTR_DATA_TYPE, VAL_INTEGER);
    SetCtrlAttribute(panel, ctrl, ATTR_MIN_VALUE, 0);
    SetCtrlAttribute(panel, ctrl, ATTR_MAX_VALUE, 100);

    SetCtrlAttribute(panel, ctrl, ATTR_SHOW_DIG_DISP, FALSE);
    SetCtrlAttribute(panel, ctrl, ATTR_SHOW_INCDEC_ARROWS, FALSE);
    SetCtrlAttribute(panel, ctrl, ATTR_MARKER_STYLE, indicatorMarkerStyle);
    if (indicatorMarkerStyle == VAL_NO_MARKERS)
        SetCtrlAttribute (panel, ctrl, ATTR_TICK_STYLE, VAL_NO_TICKS);
    SetCtrlAttribute(panel, ctrl, ATTR_SLIDER_COLOR, VAL_TRANSPARENT);

    GetCtrlAttribute(panel, ctrl, ATTR_LABEL_HEIGHT, &labelHeight);
    SetCtrlAttribute(panel, ctrl, ATTR_TOP, PROGRESS_BOX_MARGIN + labelHeight + PROGRESS_LABEL_CTRL_V_GAP);

    SetCtrlAttribute(panel, ctrl, ATTR_LABEL_LEFT, PROGRESS_BOX_MARGIN);
    SetCtrlAttribute(panel, ctrl, ATTR_LABEL_TOP, PROGRESS_BOX_MARGIN);

    GetCtrlBoundingRect(panel, ctrl, &top, &left, &height, &width);

    dlgHeightWithoutCancel = height + 2 * PROGRESS_BOX_MARGIN;

    errChk( CenterCtrl(panel, ctrl, 0, 0, dlgHeightWithoutCancel, dlgWidth));

    if (cancelButtonLabel && *cancelButtonLabel)
        {
        GetCtrlAttribute(panel, cancelBtn, ATTR_HEIGHT, &buttonHeight);
        GetCtrlAttribute(panel, cancelBtn, ATTR_WIDTH, &buttonWidth);
        SetCtrlAttribute(panel, cancelBtn, ATTR_TOP, dlgHeightWithoutCancel);
        SetCtrlAttribute(panel, cancelBtn, ATTR_LEFT, dlgWidth / 2 - buttonWidth / 2);  /* center the button */
        SetPanelAttribute(panel, ATTR_HEIGHT, dlgHeightWithoutCancel + PROGRESS_BOX_MARGIN + buttonHeight);
        }
    else
        SetPanelAttribute(panel, ATTR_HEIGHT, height + 2*PROGRESS_BOX_MARGIN);

    SetPanelPos(panel, VAL_AUTO_CENTER, VAL_AUTO_CENTER);

    if (modal)
        InstallPopup(panel);
    else
        DisplayPanel(panel);

Error:
    if (error < 0)
        {
        DiscardProgressDialog(panel);
        return error;
        }
    else
        return panel;
}

/*********************************************************************/

int CVIFUNC GetFileWritability (char pathName[], int *isWritable)
{
    FILE    *file;
    long        dummy;
    int     error = 0;
    int     oldBreakPolicy = GetBreakOnLibraryErrors();

    DisableBreakOnLibraryErrors ();
    *isWritable = FALSE;
    switch (GetFileInfo (pathName, &dummy))  /* formatio.h */
        {
        case 1:  /* file exists */
            file = fopen (pathName, "r+b");
            if (file)
                {    /* can open it for read/write */
                fclose (file);
                *isWritable = TRUE;
                }
            break;
        case 0:  /* file does not exist */
            file = fopen (pathName, "wb");
            if (file)
                {    /* can create it and open it for writing */
                fclose (file);
                remove (pathName);
                *isWritable = TRUE;
                }
            break;
        case -1: /* too many file handles open */
            errChk (UIETooManyFilesOpen);
            break;
        default: /* should not occur */
            errChk (UIEIOError);
            break;
        }

Error:
    if (oldBreakPolicy)
        EnableBreakOnLibraryErrors ();
    return error;
}

/**************************************************************************/
/*
 * Create a temporary file in the same directory as the file identified
 * by pathName.  Use the prefix (which should be 5 characters in length),
 * followed by an index of up to three characters, followed by a dot,
 * followed by the extension (which should be from 1 to 3 characters in
 * length).  Returns the file pointer in tempFile, returns the name of the
 * temporary file in tempPathName.
 */

#define MAX_TEMP_PREFIX_LENGTH  5

int CVIFUNC CreateAndOpenTemporaryFile (char *pathName, char *prefix, char *extension,
                                char *openModeString,
                                char tempPathName[], FILE **tempFile)

{
    char    driveName[MAX_DRIVENAME_LEN];
    char    dirName[MAX_DIRNAME_LEN];
    char    fileName[MAX_FILENAME_LEN];
    int     foundNameToUse=FALSE;
    long    dummy;
    int     index;
    FILE    *file;
    int     error = 0;
    int     oldBreakPolicy = GetBreakOnLibraryErrors();
    char    prefixBuf[MAX_TEMP_PREFIX_LENGTH + 1];

    DisableBreakOnLibraryErrors ();

    StringCopyMax(prefixBuf, prefix, MAX_TEMP_PREFIX_LENGTH + 1);

    SplitPath (pathName, driveName, dirName, fileName);

    /*****************************************************************/
    /* Try to create a file in the same directory but with the       */
    /* prefix, followed by an index, followed by the extension.      */
    /* Try up to 1000 different indexes.                             */
    /* For each one, determine if a files exists under that name.    */
    /* If so, increment the index and try again.                     */
    /*****************************************************************/
    for (index=0; index < 1000 && !foundNameToUse; index++)
        {
        sprintf (tempPathName, "%s%s%s%1d.%s", driveName, dirName, 
                                               prefixBuf, index, extension);
        switch (GetFileInfo (tempPathName, &dummy))  /* formatio.h */
            {
            case 1:  /* file exists */
                break;    /* increment index and try again */
            case 0:  /* file does not exist */
                foundNameToUse = TRUE;
                break;
            case -1: /* too many file handles open */
                errChk (UIETooManyFilesOpen);
                break;
            default: /* should not occur */
                errChk (UIEIOError);
                break;
            }
        }

    if (!foundNameToUse)
        errChk (ToolErr_CouldNotFindUnusedTempFileName);

    file = fopen (tempPathName, openModeString);
    if (file)
        *tempFile = file;
    else
        errChk (ToolErr_CouldNotCreateTempFile);

Error:
    if (oldBreakPolicy)
        EnableBreakOnLibraryErrors ();
    return error;
}

/*********************************************************************/

void CVIFUNC RemoveFileIfExists (char pathName[])
{
    int oldBreakPolicy = GetBreakOnLibraryErrors();

    DisableBreakOnLibraryErrors();
    remove (pathName);
    if (oldBreakPolicy)
        EnableBreakOnLibraryErrors();
}

/*********************************************************************/

int CVIFUNC DeleteAndRename (char srcPathName[], char destPathName[])
{
    int error;
    int oldBreakPolicy = GetBreakOnLibraryErrors();

    DisableBreakOnLibraryErrors();
    RemoveFileIfExists (destPathName);
    switch (RenameFile (srcPathName, destPathName))
        {
        case 0:   /* success */
            error = 0;
            break;
        case -1:    /* src file not found */
            error = UIEFileWasNotFound;
            break;
        case -2:
            error = UIEAccessDenied;
            break;
        case -3:
            error = UIEIOError;
            break;
        case -4:
            error = UIEOutOfMemory;
            break;
        default:
            error = UIEIOError;
            break;
        }
    if (oldBreakPolicy)
        EnableBreakOnLibraryErrors();
    return error;
}

/**************************************************************************/

    /* returns 1 if it exists, 0 if it doesn't, or an UIE error */
int CVIFUNC FileExists(char *pathName, int *fileSize)
{
    long    size = -1;
    int     result;
    int     savedBOLE = SetBOLE(FALSE);

    switch (GetFileSize (pathName, &size))
        {
        case 0: /* success */
            result = 1;
            break;
        case -1:    /* file not found */
            result = 0;
            break;
        case -3:    /* general I/O error */
            result = UIEIOError;
            break;
        case -4:    /* out of memory  */
            result = UIEOutOfMemory;
            break;
        default:
            result = UIEIOError;
            break;
        }

    SetBOLE(savedBOLE);
    if (fileSize)
        *fileSize = size;
    return result;
}

/**************************************************************************/

    /*  Retrieves the size of the file in bytes.  Similar to the utility library
        function GetFileSize except that it works from a FILE pointer instead of
        a pathname.  The file must have been opened in binary mode for the results to
        be meaningful.
    */
int CVIFUNC BinaryFileSize(FILE *file, int *size)
{
    int error = 0;
    int currPosition;
    int endPosition;

    currPosition = ftell(file);
    if (currPosition == -1 || errno == 100)
        errChk( UIEIOError);

    if (fseek (file, 0, SEEK_END) != 0)
        errChk( UIEIOError);

    endPosition = ftell(file);
    if (endPosition == -1 || errno == 100)
        errChk( UIEIOError);

    *size = endPosition;

    if (fseek (file, currPosition, SEEK_SET) != 0)  /* restore file position */
        errChk( UIEIOError);

Error:
    return error;
}

/**************************************************************************/

#define SEARCH_BUF_SIZE 128

    /*  returns TRUE if found, or FALSE if not, or a negative error code.
        If the bytesToFind are found, offset is set to how many bytes from the
        initial file position the bytesToFind were located at.
        If the bytesToFind were not found, then offset is set to the distance to
        one byte past the end of the file.
        The position parameter is set to the absolute file position of the
        beginning of the bytes found. If the bytes were not found then position
        is set to one byte past the end of the file.
        Both offset and position are optional and may be NULL.

        The file position is set to the byte after the found string,
        or to one byte past the end of the file.
    */
int CVIFUNC FindBytesInFile(FILE *file, void *bytesToFind, int numBytesToFind, int *offset, int *position)
{
    int     error = 0;
    char    buf[SEARCH_BUF_SIZE];
    char    *bufPtr;
    char    *unMatched;
    int     numLeft = 0;
    int     numBytesFound;
    int     numExamined = 0;
    int     savedBole = SetBOLE(FALSE);
    int     found = FALSE;
    int     startPosition;
    int     offsetToEndOfRead;

    numBytesFound = 0;
    unMatched = (char *)bytesToFind;

    startPosition = ftell(file);
    if (startPosition == -1 || errno == 100)
        errChk( UIEIOError);

    while (TRUE)
        {
        if (numLeft < 1)
            {
            numLeft  = fread (buf, 1, SEARCH_BUF_SIZE, file);
            if (numLeft == 0)
                break;
            else
                bufPtr = buf;
            }

        numExamined++;
        if  (*bufPtr == *unMatched)
            {
            unMatched++;
            numBytesFound++;
            if (numBytesFound == numBytesToFind)
                break;
            }
        else
            {
            unMatched = (char *)bytesToFind;
            numBytesFound = 0;
            }

        bufPtr++;
        numLeft--;
        }

    if (numBytesFound == numBytesToFind)
        {
        offsetToEndOfRead = numExamined - numBytesToFind;   /* back up to start to the bytes to find */
        found = TRUE;
        }
    else
        offsetToEndOfRead = numExamined;

    if (offset)
        *offset = offsetToEndOfRead;

    if (position)
        *position = startPosition + offsetToEndOfRead;

    if (found)
        {
        if (fseek(file, startPosition + offsetToEndOfRead + numBytesToFind, SEEK_SET) != 0)
            errChk( UIEIOError);
        }

    SetBOLE(savedBole);

Error:
    if (error < 0)
        return error;
    else
        return found;
}

/**************************************************************************/

#define READ_BUF_SIZE 128

    /*  returns 0 if successful, or a negative error code..
        The file offset is set to just past the last byte of the string read from the file .
        pass -1 to maxCharactersToRead to have no limit on the string length

        Use endOfFile to find out if the string was terminated by reaching the end of the file.
        Use maxCharactersRead to find out if the string in the file was longer than maxCharactersToRead.
    */
int CVIFUNC ReadStringFromBinaryFile(FILE *file, char **string, int maxCharactersToRead, int *endOfFile, int *maxCharactersRead)
{
    int     error = 0;
    char    readBuf[READ_BUF_SIZE];
    char    *newString = 0;
    char    *stringEnd;
    int     numRead;
    int     numLeftToRead;
    int     numToRead;
    long    startingFilePos;
    int     totalBytesRead = 0;

    if (maxCharactersRead)
        *maxCharactersRead = FALSE;
    if (endOfFile)
        *endOfFile = FALSE;

    nullChk( newString = StrDup(""));

    numLeftToRead = maxCharactersToRead;

    startingFilePos = ftell(file);
    if (startingFilePos == -1 || errno == 100)
        errChk( UIEIOError);

    while (TRUE)
        {
        if (maxCharactersToRead < 0)
            numToRead = READ_BUF_SIZE;
        else
            numToRead =  Min(READ_BUF_SIZE, numLeftToRead);

        if (numToRead <= 0)
            {
            if (maxCharactersRead)
                *maxCharactersRead = TRUE;
            break;
            }

        numRead = fread (readBuf, 1, numToRead, file);

        if (numRead == 0)
            {
            if(endOfFile)
                *endOfFile = TRUE;
            break;
            }
        else
        if (maxCharactersToRead >= 0)
            numLeftToRead -= numRead;

        errChk( AppendString(&newString, readBuf, numRead));

        stringEnd = memchr(readBuf, 0, numRead);
        if (stringEnd)
            {
            totalBytesRead += stringEnd - readBuf + 1;  /* +1 for null terminator */
            break;
            }
        else
            totalBytesRead += numRead;
        }

        /* set file position just after last byte read */
    if (fseek (file , startingFilePos + totalBytesRead, SEEK_SET) != 0)
        errChk( UIEIOError);

Error:
    if (error < 0)
        {
        free(newString);
        *string = 0;
        return error;
        }
    else
        {
        *string = newString;
        return 0;
        }
}

/**************************************************************************/

    /*  returns 0 if successful, or a negative error code..
        The file offset is set to just past the last byte of the string read from the file .
        pass -1 to maxCharactersToRead to have no limit on the string length

        Use endOfFile to find out if the string was terminated by reaching the end of the file.
        Use maxCharactersRead to find out if the string in the file was longer than maxCharactersToRead.
    */
int CVIFUNC ReadLineFromBinaryFile(FILE *file, char **string, int maxCharactersToRead, int *endOfFile, int *maxCharactersRead)
{
    int error = 0;
    int startingFilePos;
    int lineLength;
    int lineTerminatorOffset;
    int found;

    startingFilePos = ftell(file);
    if (startingFilePos == -1 || errno == 100)
        errChk( UIEIOError);

    errChk( found = FindBytesInFile(file, LINE_TERMINATOR, LINE_TERMINATOR_LEN, 0, &lineTerminatorOffset));

    if (fseek(file, startingFilePos, SEEK_SET) != 0)
        errChk( UIEIOError);

    if (found)
        {
        lineLength = lineTerminatorOffset - startingFilePos + LINE_TERMINATOR_LEN;
        if (maxCharactersToRead >= 0)
            maxCharactersToRead = Min(maxCharactersToRead, lineLength);
        else
            maxCharactersToRead = lineLength;
        }

    errChk( ReadStringFromBinaryFile(file, string, maxCharactersToRead, endOfFile, maxCharactersRead));

Error:
    if (error < 0)
        return error;
    else
        return 0;
}

/***********************************************************************************/

#define WRITE_BYTE_BUF_SIZE 64

int CVIFUNC WriteByteToFile(FILE *file, char byte, int numTimes)
{
    int     error = 0;
    char    writeBuf[WRITE_BYTE_BUF_SIZE];
    int     numToWrite;
    int     numWritten;

    if (numTimes > 0)
        {
        memset (writeBuf, byte, Min(numTimes, WRITE_BYTE_BUF_SIZE));

        do
            {
            numToWrite = Min(WRITE_BYTE_BUF_SIZE, numTimes);

            numWritten = fwrite (writeBuf, 1, numToWrite, file);
            if (numWritten != numToWrite)
                errChk( UIEDiskFull);

            numTimes -= numWritten;
            } while (numTimes > 0);
        }

Error:
    return error;
}

/**************************************************************************/

char * CVIFUNC SkipDriveName(char *pathString)
{
#if   defined(_NI_mswin_)
    if (pathString && strlen(pathString) >= 2 && pathString[1] == ':')
        pathString += 2;
#endif

    return pathString;
}

/**************************************************************************/

int CVIFUNC IsAbsolutePath(char *pathString)
{
    int absolute = FALSE;
    int fileSepLength = strlen(FILE_SEPARATOR);
    int pathLen;

    if (pathString && *pathString)
        {
        pathString = SkipDriveName(pathString);
        pathLen = strlen(pathString);

        if (pathLen >= fileSepLength && memcmp(pathString, FILE_SEPARATOR, fileSepLength) == 0)
            absolute = TRUE;
        }

    return absolute;
}

/**************************************************************************/

int CVIFUNC WriteStringToFile (FILE *stream, char *string)
{
    int     error = 0;
    int     len, numWritten;
    char    *newLinePtr;

    while (string)
        {
        newLinePtr = strchr(string, '\n');

        if (newLinePtr)
            len = newLinePtr - string;
        else
            len = strlen(string);

        if (len > 0)
            {
            numWritten = fwrite (string, 1, len, stream);
            if (numWritten != len)
                errChk( UIEDiskFull);
            }

        if (newLinePtr)
            {
            len = strlen (LINE_TERMINATOR);
            numWritten = fwrite (LINE_TERMINATOR, 1, len, stream);
            if (numWritten != len)
                errChk( UIEDiskFull);

            newLinePtr++;   /* skip over newline */
            }

        string = newLinePtr;
        }

Error:
    return error;
}

/**************************************/

void CVIFUNC ChangeFileNameExtension(char *fileName, char *newFileName, char *extension)
    /* fileName may have the same buffer as newFileName */
    /* extension should not include the '.' */
{
    char *dest;

    if ((!fileName) || (!newFileName))
        AssertMsg(0, "Null parameters passed to ChangeFileNameExtension");

    strcpy (newFileName, fileName);
    dest = FindFileDotPosition (newFileName);
    *dest++ = '.';      /* add the '.' in case there was not one already */
    strcpy (dest, extension);
}

/**************************************/

    /* does not preserve the extension, basename should include an extension if needed */
void CVIFUNC ChangeBaseFileName(char *pathString, char *newPathString, char *baseName)
{
    int len;

    len = FindFileName(pathString) - pathString;
    memmove(newPathString, pathString, len);
    strcpy (newPathString + len, baseName);
}

/**************************************/

char * CVIFUNC FindFileDotPosition (char *pathString)
{
 /* This function takes a string and returns a pointer to the last '.' within
    kMaxFileExtension bytes from the end of the string.   If it cannot find a dot, then
    returns a pointer to the terminating nul byte of the string */

    char *p;

    AssertMsg(pathString, "Null pathString parameter passed to FindFileDotPosition");

    p = FindFileExtension (pathString);
    if ((p > pathString) && (*(p-1) == '.'))
        return p - 1;
    else
        return p;
}

/**************************************/

#if defined(_NI_mswin16_)
#define MAX_FILE_EXTENSION_LENGTH   3
#elif defined(_NI_mswin32_)
#define MAX_FILE_EXTENSION_LENGTH   254
#elif defined(_NI_unix_) || defined(_NI_sparc_)
#define MAX_FILE_EXTENSION_LENGTH   253
#elif defined(_NI_mac_)
#define MAX_FILE_EXTENSION_LENGTH   253
#else
#error Undefined platform
#endif

/*  This function takes a string and returns a pointer to the character after
    the last '.' in the string IF the last  '.' is within
    MAX_FILE_EXTENSION_LENGTH characters of the end of the string.
    Else, it returns returns a pointer to the terminating nul byte of the string
*/
char * CVIFUNC FindFileExtension(char *pathString)
{
    int index, count=0;
    char *fileName;
    char *terminatorPtr;

    AssertMsg(pathString, "Null pathString parameter passed to FindFileExtension");

    fileName = FindFileName(pathString);

    if ((index = strlen (fileName)) == 0)
        return fileName;

    terminatorPtr = fileName + index;

    for (; index > 1; index--)     /* do not bother checking when index is 1 because */
         {                         /* if the dot is in position 0, it really is not an extension */
        if (fileName[index-1] == '.')
            return &fileName[index];
        count++;
        if (count > MAX_FILE_EXTENSION_LENGTH)
            return terminatorPtr;
        }

    return terminatorPtr;
}

/*********************************************/

 /* This function takes a string and returns a pointer to the character after last file separator
    in the string or to the string itself if no separators are found */

char * CVIFUNC FindFileName(char *pathString)
{
    int len;
    char *p;

    AssertMsg(pathString, "Null pathString parameter passed to FindFileName");

    if ((len = strlen(pathString)) == 0)
        return pathString;

    for (p= pathString + len - 1; p > pathString; p--)
        {
#if   defined(_NI_mswin_)
        if ( *(p-1) == *FILE_SEPARATOR || *(p-1) == ':')    /* i.e. c:x.c should return x.c on PC */
#else
        if ( *(p-1) == *FILE_SEPARATOR)
#endif
            return p;
        }

    return pathString;
}

/**************************************************************************/

char * CVIFUNC GetToolboxErrorString(int errorCode)
{
    switch (- ABS_VAL(errorCode))
        {
        case ToolErr_CouldNotOpenFileForReading:
            return  "Could Not Open File For Reading";
        case ToolErr_ErrorReadingFile:
            return "Error Reading From File";
        case ToolErr_CouldNotFindUnusedTempFileName :
            return "Could Not Find Unused Temporary File Name";
        case ToolErr_CouldNotCreateTempFile:
            return "Could Not Create Temporary File";
        case ToolErr_IncorrectFileFormat:
            return "Incorrect File Format";
        case ToolErr_UnexpectedEndOfFile:
            return "Unexpected End Of File";
        case ToolErr_InvalidIntNumber:
            return "Invalid Integer Number";
        case ToolErr_InvalidUIntNumber:
            return "Invalid Unsigned Integer Number";
        case ToolErr_InvalidDoubleNumber:
            return "Invalid Floating Point Number";
        case ToolErr_InvalidBooleanValue:
            return "Invalid Boolean Value";
        case ToolErr_MissingSection:
            return "Missing Section";
        case ToolErr_MissingItem:
            return "Missing Item";
        case ToolErr_UnRecognizedValue:
            return "Unrecognized Value";
        case ToolErr_OperationInProgress:
            return "Could not perform requested action because of an operation currently in progress";
        case ToolErr_TooManyItems:
            return "Too many items";
        case ToolErr_DuplicateItemOrValue:
            return "Duplicate item or value";
        case ToolErr_PanelNotAChildOfCorrectPanel:
            return "Panel is not a child panel of the correct panel";
        default:
            return "No description available. ";
        }
}

/**************************************************************************/

char * CVIFUNC GetGeneralErrorString(int errorCode)
{
    int absError = ABS_VAL(errorCode);

    if (absError < 1000)
        return GetUILErrorString(errorCode);
    else
    if (absError >= 5000 && absError < 6000)
        return GetToolboxErrorString(errorCode);
    else
    if (absError >= 10000 && absError < 11000)
        return ToolBox_GetDAQErrorString((short int)errorCode);
    else
        return "No description available. ";
}

/**************************************************************************/

static ListType gExistenceList = 0;

    /*  returns non-zero if success, or 0 if out of memory */
int CVIFUNC RegisterExistence(void *objectPtr)
{
    if (!gExistenceList)
        {
        gExistenceList = ListCreate(sizeof(void *));
        if (gExistenceList)
            {
            atexit(ToolBoxAtExitFunction);
            }
        }

    if (gExistenceList)
        return ListInsertInOrder (gExistenceList, &objectPtr, 0);
    else
        return 0;
}

/**************************************************************************/

void CVIFUNC DeRegisterExistence(void *objectPtr)
{
    if (gExistenceList && objectPtr)
        {
        int index;

        index = ListBinSearch (gExistenceList, &objectPtr, 0);
        if (index > 0)
            ListRemoveItem (gExistenceList, 0, index);
        }
}

/**************************************************************************/

int CVIFUNC ObjectExists(void *objectPtr)
{
    if (gExistenceList && objectPtr)
        {
        int index;

        index = ListBinSearch (gExistenceList, &objectPtr, 0);
        return index > 0;
        }
    else
        return FALSE;
}

/**************************************************************************/

    /*  dispose and resources allocated by the programmers toolbox */
static void CVIANSI ToolBoxAtExitFunction(void)
{
    if (gExistenceList)
        {
        ListDispose(gExistenceList);
        gExistenceList = 0;
        }
}

/**************************************************************************/

    /*  Only one link allowed per type.  If already linked with the same type then UIEInvalidControlType
        is returned.  If successfully linked, 0 is returned, otherwise a negative error
        code is returned.

        Currently this is no support for deleting or replacing a link (could be added in future).
        All links are destroyed when the control is discarded.
    */
int CVIFUNC ChainCtrlCallback(int panel, int ctrl, CtrlCallbackPtr newCallback, void *newCallbackData, char *type)
{
    int                     error = 0;
    CtrlCallbackChainLink   link = 0;
    int                     alreadyLinked = FALSE;

    if (!type)
        errChk( UIENullPointerPassed);

    errChk( alreadyLinked = FindCtrlCallbackLink(panel, ctrl, type, 0));

    if  (!alreadyLinked)
        {
        InstallObjectDeletionProcedure();
        nullChk( link = (CtrlCallbackChainLink)calloc(sizeof(*link),1));
        nullChk( link->type = StrDup(type));
        link->signature = CAT4CHARS('L', 'I', 'N', 'K');

            /*  nothing passed this point should fail if the code is working as intended */
        errChk( GetCtrlAttribute(panel, ctrl, ATTR_CALLBACK_FUNCTION_POINTER, &link->previousCallback));
        errChk( GetCtrlAttribute(panel, ctrl, ATTR_CALLBACK_DATA, &link->previousCallbackData));

        link->callback = newCallback;
        link->callbackData = newCallbackData;

        errChk( InstallCtrlCallback(panel, ctrl, LinkedCtrlCallback, (void *)link));
        }
    else
        errChk( UIEInvalidControlType);

Error:
    if (error < 0)
        {
        if (link)
            {
            free(link->type);
            free(link);
            }
        return error;
        }
    else
        return 0;
}

/*************************************************************************/

    /*  returns  UIEInvalidControlType if the control is not linked with the specified type */
int CVIFUNC GetChainedCallbackData(int panel, int ctrl, char *type, void **callbackData)
{
    int                     error = 0;
    int                     found = FALSE;
    CtrlCallbackChainLink   link;

    errChk( found = FindCtrlCallbackLink(panel, ctrl, type, &link));

    if (found)
        *callbackData = link->callbackData;
    else
        errChk( UIEInvalidControlType);

Error:
    if (error < 0)
        return error;
    else
        return 0;
}

/*************************************************************************/

int CVIFUNC UnchainCtrlCallback(int panel, int ctrl, char *type)
{
    int                     error = 0;
    CtrlCallbackPtr         callbackFunc;
    int                     found = FALSE;
    CtrlCallbackChainLink   currLink;
    CtrlCallbackChainLink   prevLink = NULL;
    int                     savedBole = SetBOLE(FALSE);
    
    if (!type)
        errChk( UIENullPointerPassed);

    errChk( GetCtrlAttribute(panel, ctrl, ATTR_CALLBACK_FUNCTION_POINTER, &callbackFunc));
    errChk( GetCtrlAttribute(panel, ctrl, ATTR_CALLBACK_DATA, &currLink));

    while (!found && (callbackFunc == LinkedCtrlCallback) && currLink)
        {
        if (strcmp(currLink->type, type) == 0)
            found = TRUE;
        else
            {
            callbackFunc = currLink->previousCallback;
            prevLink = currLink;
            currLink = currLink->previousCallbackData;
            }
        }

    if (found)
        {
        if (!prevLink)
            {
            errChk( InstallCtrlCallback(panel, ctrl,  currLink->previousCallback, (void *)currLink->previousCallbackData));
            }
        else
            {
            prevLink->previousCallback = currLink->previousCallback;
            prevLink->previousCallbackData = currLink->previousCallbackData;
            }
        
        free(currLink->type);
        free(currLink);
        }
        
Error:
    SetBOLE(savedBole);
    if (error < 0)
        return error;
    else
        return found;
}

/*************************************************************************/

static int CVIFUNC FindCtrlCallbackLink(int panel, int ctrl, char *type,  CtrlCallbackChainLink *link)
{
    int                     error = 0;
    CtrlCallbackPtr         callbackFunc;
    int                     found = FALSE;
    CtrlCallbackChainLink   currLink;

    if (!type)
        errChk( UIENullPointerPassed);

    errChk( GetCtrlAttribute(panel, ctrl, ATTR_CALLBACK_FUNCTION_POINTER, &callbackFunc));
    errChk( GetCtrlAttribute(panel, ctrl, ATTR_CALLBACK_DATA, &currLink));

    while (!found && (callbackFunc == LinkedCtrlCallback) && currLink)
        {
        if (strcmp(currLink->type, type) == 0)
            {
            found = TRUE;
            if (link)
                *link = currLink;
            }
        else
            {
            callbackFunc = currLink->previousCallback;
            currLink = currLink->previousCallbackData;
            }
        }

Error:
    if (error < 0)
        return error;
    else
        return found;
}

/*************************************************************************/

    /*  When a callback swallows an event by returning a non-zero value, no further callbacks in the
        chain of linked callbacks are called  (with the exception of the EVENT_DISCARD event which is
        alway sent to each callback).  The final return value returned to CVI is TRUE if any ANY of the
        linked callbacks returned TRUE.
    */
static int CVICALLBACK LinkedCtrlCallback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    CtrlCallbackChainLink   link = (CtrlCallbackChainLink)callbackData;
    int                     returnVal = 0;
    CtrlCallbackPtr         previousCallback = link->previousCallback;
    void                    *previousCallbackData = link->previousCallbackData;

        /*  This is being done as a formality, it is only needed if one of the callback functions is
            "bad" and causes an event to be sent to the control's installed callback as a result of
            receiving the discard message.
        */
    if (event == EVENT_DISCARD)
        InstallCtrlCallback(panel, control, previousCallback, previousCallbackData);

        /*  Call the callback passed in to ChainCtrlCallback */
    if (link->callback)
        returnVal =  link->callback(panel, control, event, link->callbackData, eventData1, eventData2);

    if (event == EVENT_DISCARD)   /* can't swallow EVENT_DISCARD */
        returnVal = 0;

    if (returnVal == 0 && previousCallback)
        returnVal = previousCallback(panel, control, event, previousCallbackData, eventData1, eventData2);

    if (event == EVENT_DISCARD)
        {
        free(link->type);
        free(link);
        }

    return returnVal;
}

/*****************************************************************************************/

int CVIFUNC CallCtrlCallback(int panel, int ctrl, int event, int eventData1, int eventData2, int *returnValue)
{
    void                *callbackData;
    CtrlCallbackPtr     callback;
    int                 result;
    int                 calledCallback = FALSE;
    int                 error = 0;

    errChk( GetCtrlAttribute(panel, ctrl, ATTR_CALLBACK_FUNCTION_POINTER, &callback));

    if (callback)
        {
        errChk( GetCtrlAttribute(panel, ctrl, ATTR_CALLBACK_DATA, &callbackData));


        result = (*callback)(panel, ctrl, event, callbackData, eventData1, eventData2);
        if (returnValue)
            *returnValue = result;
        calledCallback = TRUE;
        }

Error:
    if (error < 0)
        return error;
    else
        return calledCallback;
}

/*****************************************************************************************/

static void CVIANSI ObjectDeletionProcedure(void)
{
    extern int CVIFUNC_C DiscardAllUserInterfaceObjects(void);

    DiscardAllUserInterfaceObjects();
}

/*****************************************************************************************/

    /*  When custom controls are deleted by DiscardCtrl() and DiscardPanel(), they have
        a chance to free their associated resources in response to the EVENT_DISCARD
        event.  However, when all remaining UI objects are deleted when the program
        terminates, EVENT_DISCARD is not sent because execution is already terminated.
        This means the memory and other resources allocated for by a custom control will
        not have been freed.  This can be observed in the dynamic memory display after
        exiting a program.  This function causes the deletion of all user interface objects
        to occur while the program is still running so that custom controls can clean up
        after themselves.

        Note: For memory, this is not really a problem, since all unfreed blocks are
        freed before the next time a program is run.
    */
static void CVIFUNC InstallObjectDeletionProcedure(void)
{
    static  int installed = 0;

    if (!installed)
        {
        if (atexit(ObjectDeletionProcedure) == 0)
            installed = 1;
        }
}

/**************************************************************************/

    /*  Only one link allowed per type.  If already linked with the same type then UIEInvalidControlType
        is returned.  If successfully linked, 0 is returned, otherwise a negative error
        code is returned.

        Currently this is no support for deleting or replacing a link (could be added in future).
        All links are destroyed when the control is discarded.
    */
int CVIFUNC ChainPanelCallback(int panel, PanelCallbackPtr newCallback, void *newCallbackData, char *type)
{
    int                     error = 0;
    PanelCallbackChainLink  link = 0;
    int                     alreadyLinked = FALSE;

    if (!type)
        errChk( UIENullPointerPassed);

    errChk( alreadyLinked = FindPanelCallbackLink(panel, type, 0));

    if  (!alreadyLinked)
        {
        InstallObjectDeletionProcedure();
        nullChk( link = (PanelCallbackChainLink)calloc(sizeof(*link),1));
        nullChk( link->type = StrDup(type));
        link->signature = CAT4CHARS('L', 'I', 'N', 'K');

            /*  nothing passed this point should fail if the code is working as intended */
        errChk( GetPanelAttribute(panel, ATTR_CALLBACK_FUNCTION_POINTER, &link->previousCallback));
        errChk( GetPanelAttribute(panel, ATTR_CALLBACK_DATA, &link->previousCallbackData));

        link->callback = newCallback;
        link->callbackData = newCallbackData;

        errChk( InstallPanelCallback(panel, LinkedPanelCallback, (void *)link));
        }
    else
        errChk( UIEInvalidControlType);

Error:
    if (error < 0)
        {
        if (link)
            {
            free(link->type);
            free(link);
            }
        return error;
        }
    else
        return 0;
}

/*************************************************************************/

    /*  returns  UIEInvalidControlType if the panel is not linked with the specified type */
int CVIFUNC GetChainedPanelCallbackData(int panel, char *type, void **callbackData)
{
    int                     error = 0;
    int                     found = FALSE;
    PanelCallbackChainLink  link;

    errChk( found = FindPanelCallbackLink(panel, type, &link));

    if (found)
        *callbackData = link->callbackData;
    else
        errChk( UIEInvalidControlType);

Error:
    if (error < 0)
        return error;
    else
        return 0;
}

/*************************************************************************/

int CVIFUNC UnchainPanelCallback(int panel, char *type)
{
    int                     error = 0;
    PanelCallbackPtr        callbackFunc;
    int                     found = FALSE;
    PanelCallbackChainLink  currLink;
    PanelCallbackChainLink  prevLink = NULL;
    int                     savedBole = SetBOLE(FALSE);

    if (!type)
        errChk( UIENullPointerPassed);

    errChk( GetPanelAttribute(panel, ATTR_CALLBACK_FUNCTION_POINTER, &callbackFunc));
    errChk( GetPanelAttribute(panel, ATTR_CALLBACK_DATA, &currLink));

    while (!found && (callbackFunc == LinkedPanelCallback) && currLink)
        {
        if (strcmp(currLink->type, type) == 0)
            found = TRUE;
        else
            {
            callbackFunc = currLink->previousCallback;
            prevLink = currLink;
            currLink = currLink->previousCallbackData;
            }
        }

    if (found)
        {
        if (!prevLink)
            {
            errChk( InstallPanelCallback(panel,  currLink->previousCallback, (void *)currLink->previousCallbackData));
            }
        else
            {
            prevLink->previousCallback = currLink->previousCallback;
            prevLink->previousCallbackData = currLink->previousCallbackData;
            }
        
        free(currLink->type);
        free(currLink);
        }

Error:
    SetBOLE(savedBole);
    if (error < 0)
        return error;
    else
        return found;
}

/*************************************************************************/

static int CVIFUNC FindPanelCallbackLink(int panel, char *type,  PanelCallbackChainLink *link)
{
    int                     error = 0;
    PanelCallbackPtr        callbackFunc;
    int                     found = FALSE;
    PanelCallbackChainLink  currLink;

    if (!type)
        errChk( UIENullPointerPassed);

    errChk( GetPanelAttribute(panel, ATTR_CALLBACK_FUNCTION_POINTER, &callbackFunc));
    errChk( GetPanelAttribute(panel, ATTR_CALLBACK_DATA, &currLink));

    while (!found && (callbackFunc == LinkedPanelCallback) && currLink)
        {
        if (strcmp(currLink->type, type) == 0)
            {
            found = TRUE;
            if (link)
                *link = currLink;
            }
        else
            {
            callbackFunc = currLink->previousCallback;
            currLink = currLink->previousCallbackData;
            }
        }

Error:
    if (error < 0)
        return error;
    else
        return found;
}

/*************************************************************************/

    /*  When a callback swallows an event by returning a non-zero value, no further callbacks in the
        chain of linked callbacks are called  (with the exception of the EVENT_DISCARD event which is
        alway sent to each callback).  The final return value returned to CVI is TRUE if any ANY of the
        linked callbacks returned TRUE.
    */
static int CVICALLBACK LinkedPanelCallback(int panel, int event, void *callbackData, int eventData1, int eventData2)
{
    PanelCallbackChainLink  link = (PanelCallbackChainLink)callbackData;
    int                     returnVal = 0;
    PanelCallbackPtr        previousCallback = link->previousCallback;
    void                    *previousCallbackData = link->previousCallbackData;

        /*  This is being done as a formality, it is only needed if one of the callback functions is
            "bad" and causes an event to be sent to the panel's installed callback as a result of
            receiving the discard message.
        */
    if (event == EVENT_DISCARD)
        InstallPanelCallback(panel, previousCallback, previousCallbackData);

        /*  Call the callback passed in to ChainPanelCallback */
    if (link->callback)
        returnVal =  link->callback(panel, event, link->callbackData, eventData1, eventData2);

    if (event == EVENT_DISCARD)   /* can't swallow EVENT_DISCARD */
        returnVal = 0;

    if (returnVal == 0 && previousCallback)
        returnVal = previousCallback(panel, event, previousCallbackData, eventData1, eventData2);

    if (event == EVENT_DISCARD)
        {
        free(link->type);
        free(link);
        }

    return returnVal;
}

/*****************************************************************************************/

int CVIFUNC CallPanelCallback(int panel, int event, int eventData1, int eventData2, int *returnValue)
{
    void                *callbackData;
    PanelCallbackPtr    callback;
    int                 result;
    int                 calledCallback = FALSE;
    int                 error = 0;

    errChk( GetPanelAttribute(panel, ATTR_CALLBACK_FUNCTION_POINTER, &callback));

    if (callback)
        {
        errChk( GetPanelAttribute(panel, ATTR_CALLBACK_DATA, &callbackData));


        result = (*callback)(panel, event, callbackData, eventData1, eventData2);
        if (returnValue)
            *returnValue = result;
        calledCallback = TRUE;
        }

Error:
    if (error < 0)
        return error;
    else
        return calledCallback;
}

/*************************************************************/

int CVIFUNC PreviousCtrl(int panel, int ctrl, int wrap)
{
    int currentCtrl = 0;
    int nextCtrl;
    
    GetPanelAttribute(panel, ATTR_PANEL_FIRST_CTRL, &currentCtrl);
    
    if (wrap && ctrl == currentCtrl)
        return LastCtrl(panel);
        
    while (currentCtrl)
        {
        GetCtrlAttribute(panel, currentCtrl, ATTR_NEXT_CTRL, &nextCtrl);
        if (nextCtrl == ctrl)
            return currentCtrl;
        else
            currentCtrl = nextCtrl;
        }
    
    return 0;
}

/*************************************************************/

static int  CVIFUNC NextCtrl(int panel, int ctrl, int wrap)
{
    int nextCtrl;
        
    GetCtrlAttribute(panel, ctrl, ATTR_NEXT_CTRL, &nextCtrl);
    if (wrap && nextCtrl <= 0)
        GetPanelAttribute(panel, ATTR_PANEL_FIRST_CTRL, &nextCtrl);
        
    return nextCtrl;
}

/*************************************************************/

int CVIFUNC CtrlCanBeTabStop(int panel, int ctrl)
{
    int visible, dimmed, mode;

    GetCtrlAttribute(panel, ctrl, ATTR_CTRL_MODE, &mode);
    GetCtrlAttribute(panel, ctrl, ATTR_VISIBLE, &visible);
    GetCtrlAttribute(panel, ctrl, ATTR_DIMMED, &dimmed);

    return (mode != VAL_INDICATOR && !dimmed && visible) ? TRUE : FALSE;
}

/*************************************************************/

static int CVIFUNC LastCtrl(int panel)
{
    int ctrl;
    int nextCtrl;
    
    GetPanelAttribute(panel, ATTR_PANEL_FIRST_CTRL, &ctrl);
    
    while (ctrl)
        {
        GetCtrlAttribute(panel, ctrl, ATTR_NEXT_CTRL, &nextCtrl);
        if (nextCtrl < 0)       
            break;
        ctrl = nextCtrl;
        }
        
    return ctrl;    
}

/*************************************************************/

int CVIFUNC NextCtrlInTabOrder(int panel, int ctrl, int forward, int wrap)
{
    int startCtrl = ctrl;
    
    while (ctrl)
        {
        if (forward)
            ctrl = NextCtrl(panel, ctrl, wrap);
        else
            ctrl = PreviousCtrl(panel, ctrl, wrap);
    
        if (ctrl == startCtrl)
            return 0;       /* wrapped and found nothing */
            
        if (ctrl && CtrlCanBeTabStop(panel, ctrl))
            return ctrl;
        }
        
    return 0;
}

/*************************************************************/

    /*  Returns a boolean indicating whether a control was activated */
int CVIFUNC ActivateFirstCtrl(int panel)
{
    int currentCtrl;
    int controlActivated = FALSE;
    
    GetPanelAttribute(panel, ATTR_PANEL_FIRST_CTRL, &currentCtrl);
    while (currentCtrl)
        {
        if (CtrlCanBeTabStop(panel, currentCtrl))
            {
            controlActivated = TRUE;
            SetActiveCtrl(panel, currentCtrl);
            break;
            }
        else
            GetCtrlAttribute(panel, currentCtrl, ATTR_NEXT_CTRL, &currentCtrl);
        }
        
    return controlActivated;
}

/*************************************************************/

    /* discards a control if it exists, but does not cause an error if it doesn't */
void CVIFUNC DiscardCtrlNoBOLE(int panel, int ctrl, int noDiscardEvent)
{
    int savedBole = SetBOLE(FALSE);

    if (ctrl > 0)
        {
        if (noDiscardEvent)
            SetCtrlAttribute(panel, ctrl, ATTR_CALLBACK_FUNCTION_POINTER, 0);
        DiscardCtrl(panel, ctrl);
        }

    SetBOLE(savedBole);
}

/*****************************************************************************************/

    /* discards a panel if it exists, but does not cause an error if it doesn't */
void CVIFUNC DiscardPanelNoBOLE(int panel, int noDiscardEvent)
{
    int savedBole = SetBOLE(FALSE);

    if (panel > 0)
        {
        if (noDiscardEvent)
            SetCtrlAttribute(panel, ATTR_CALLBACK_FUNCTION_POINTER, 0);
        DiscardPanel(panel);
        }

    SetBOLE(savedBole);
}

/*****************************************************************************************/

static void CVIFUNC MouseEvents_Check(MouseEventsInfo info)
{
    int     leftButtonIsDown;
    int     rightButtonIsDown;
    Point   p;
    int     savedBole;
    int     error;

    savedBole = SetBOLE(FALSE);   
    error = GetRelativeMouseState (info->panel, 0, &p.x, &p.y, &leftButtonIsDown, &rightButtonIsDown, 0);
    SetBOLE(savedBole);
    if (error < 0)      /* in case of hidden panel error */
        return;
    
    if (info->lastX  != p.x || info->lastY != p.y)
        {
        info->lastX = p.x;
        info->lastY = p.y;
        CallCtrlCallback (info->panel, info->ctrl, EVENT_MOUSE_MOVE, p.y, p.x, 0);
        }
        
    if (info->leftButtonDown && !leftButtonIsDown)
        {       
        CallCtrlCallback (info->panel, info->ctrl, EVENT_LEFT_MOUSE_UP, p.y, p.x, 0);
        info->leftButtonDown = FALSE;
        }
        
    if (info->rightButtonDown && !rightButtonIsDown)
        {       
        CallCtrlCallback (info->panel, info->ctrl, EVENT_RIGHT_MOUSE_UP, p.y, p.x, 0);
        info->rightButtonDown = FALSE;
        }
}

/*****************************************************************************************/

static void CVICALLBACK MouseEvents_CheckCallback(void *callbackData)
{
    MouseEventsInfo info = (MouseEventsInfo)callbackData;

    if (info->dispose)
        free(info);
    else
        {
        PostDelayedCall (MouseEvents_CheckCallback, info, info->minPeriod);
        
        if (info->enabled)
            MouseEvents_Check(info);
        }
}

/*****************************************************************************************/

int CVIFUNC DisableExtendedMouseEvents(int panel, int ctrl)
{
    MouseEventsInfo info = 0;
    int             error = 0;
    
    errChk( GetChainedCallbackData (panel, ctrl, "MouseEvents", (void **)&info));
    info->enabled = FALSE;
    
Error:
    if (error < 0)
        return error;
    else
        return 0;
}

/*****************************************************************************************/

int CVIFUNC ConvertMouseCoordinates(int panel, int ctrl, int relativeToCtrl, int clipToControl, int *x, int *y)
{
    Point   p;
    int     top, left;
    Rect    r;
    int     error = 0;
    
    if (x)
        p.x = *x;
    else
        p.x = 0;

    if (y)
        p.y = *y;
    else
        p.y = 0;
        

    if (relativeToCtrl) 
        {
        errChk( GetCtrlAttribute(panel, ctrl, ATTR_TOP, &top));
        errChk( GetCtrlAttribute(panel, ctrl, ATTR_LEFT, &left));
        p.y -= top;
        p.x -= left;
        }

    if (clipToControl)
        {
        errChk( GetCtrlAttribute(panel, ctrl, ATTR_WIDTH, &r.width));
        errChk( GetCtrlAttribute(panel, ctrl, ATTR_HEIGHT, &r.height));

        if(!relativeToCtrl)
            {
            errChk( GetCtrlAttribute(panel, ctrl, ATTR_TOP, &r.top));
            errChk( GetCtrlAttribute(panel, ctrl, ATTR_LEFT, &r.left));
            }
        else
            {
            r.top = 0;
            r.left = 0;
            }
    
        PointPinnedToRect (p, r, &p);
        }
    
    if (x)
        *x = p.x;
    if (y)
        *y = p.y;

Error:
    if (error < 0)
        return error;
    else
        return 0;
}

/*****************************************************************************************/

int CVIFUNC EnableExtendedMouseEvents(int panel, int ctrl, double minPeriod)
{
    int             error = 0;
    MouseEventsInfo newInfo = 0;
    MouseEventsInfo currInfo = 0;
    MouseEventsInfo info = 0;
    int             savedBole;
    
    if ( GetChainedCallbackData (panel, ctrl, "MouseEvents", (void **)&currInfo) >= 0)
        info = currInfo;
    else
        {
        nullChk(newInfo = calloc(sizeof(*info), 1));
        errChk( ChainCtrlCallback (panel, ctrl, MouseEventsCallback, (void *)newInfo, "MouseEvents"));
    
        newInfo->panel = panel;
        newInfo->ctrl = ctrl;
        newInfo->signature = MOUSE_EVENTS_SIGNATURE;
        newInfo->dispose = FALSE;
        PostDelayedCall (MouseEvents_CheckCallback, newInfo, minPeriod);
        info = newInfo;
        }
    
    info->enabled = TRUE;
    info->minPeriod = minPeriod;
    
    savedBole = SetBOLE(FALSE);
    if (GetRelativeMouseState (panel, 0, &info->lastX, &info->lastY, &info->leftButtonDown, &info->rightButtonDown, 0) < 0)
        {
        info->lastX = 0;    /* in case of UIEHiddenPanelError */
        info->lastY = 0;
        }
    SetBOLE(savedBole);
    
Error:
    if (error < 0)
        {
        free(newInfo);
        return error;
        }
    else
        return 0;
}

/*****************************************************************************************/

static int CVICALLBACK MouseEventsCallback (int panel, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    MouseEventsInfo info = (MouseEventsInfo)callbackData;

    switch (event)
        {
        case EVENT_DISCARD:
            info->dispose = TRUE;
            break;
        case EVENT_LEFT_CLICK:
            info->leftButtonDown = TRUE;
            break;
        case EVENT_RIGHT_CLICK:
            info->rightButtonDown = TRUE;
            break;
        }
            
    return 0;
}

/*****************************************************************************************/

int CVIFUNC GetMetaFontInfo(char metaFont[], char typeface[], int *bold, int *underline, int *strikeout, 
    int *italic, int *size)
{
    int         error = 0;
    static  int fontPanel = 0;
    static  int txtMsg = 0;
    
    if (fontPanel <= 0)
        errChk( fontPanel = NewPanel (0, "Metafont info panel", -1000, -1000, 10, 10));
        
    if (txtMsg <= 0)
        errChk( txtMsg = NewCtrl (fontPanel, CTRL_TEXT_MSG, "Sample", 0, 0));
        

    errChk( SetCtrlAttribute(fontPanel, txtMsg, ATTR_TEXT_FONT, metaFont));
    if (bold)
        errChk( GetCtrlAttribute(fontPanel, txtMsg, ATTR_TEXT_BOLD, bold));
    if (underline)
        errChk( GetCtrlAttribute(fontPanel, txtMsg, ATTR_TEXT_UNDERLINE, underline));
    if (strikeout)
        errChk( GetCtrlAttribute(fontPanel, txtMsg, ATTR_TEXT_STRIKEOUT, strikeout));
    if (italic)
        errChk( GetCtrlAttribute(fontPanel, txtMsg, ATTR_TEXT_ITALIC, italic));
    if (size)
        errChk( GetCtrlAttribute(fontPanel, txtMsg, ATTR_TEXT_POINT_SIZE, size));

    if (typeface)
        {
                /* assume no right-minded metafont has a text size of 2, therefore, we will get back the typeface name, not the metafont name */
        errChk( SetCtrlAttribute(fontPanel, txtMsg, ATTR_TEXT_POINT_SIZE, 2));
        errChk( GetCtrlAttribute(fontPanel, txtMsg, ATTR_TEXT_FONT, typeface));
        }

Error:  
    if (error < 0)
        return error;
    else
        return 0;
}

/*****************************************************************************************/

    /*  Returns TRUE, FALSE, or negative error code */
int CVIFUNC IsMetaFont(char fontname[256])
{
    int error = 0;
    int savedBole = SetBOLE(FALSE);
    
    error = GetTextDisplaySize ("sample", fontname, NULL, NULL);
    SetBOLE(savedBole);
    
    if (error == UIEFontNotInFontTable)
        return FALSE;
    
    if (error < 0)
        return error;
    else
        return TRUE;
}

/**************************************************************************/

/*
    Used for finding underline accelerator keys (which are delimited by two underscores)
    returns the zero based position of the first underscore, or -1 if no double underlines.
*/

int CVIFUNC AcceleratorEscapeCodeLocation(char *text)
{
    char    *start = text;
    int     length;
    
    if( text)
        {
        length = strlen(text);

        for (;length > 2; text++, length--)
            if (*text == '_' && *(text+1) == '_')
                if (*(text+2) != '_')                 /* need exactly two in a row */
                    return text - start;
                else
                    for (; length > 2 && *text == '_'; text++, length--);  /* skip current block of underscores */
        }
        
    return -1;
}

/**************************************************************************/

/*
    Removes double underline characters (which are used to signify an accelerator key so that the
    text may be displayed.
*/

void CVIFUNC RemoveAcceleratorEscapeCode(char *text)
{
    int underlineLocation;
    int length;
    
    if (text)
        {
        length = strlen(text);
        underlineLocation = AcceleratorEscapeCodeLocation(text);
        if (underlineLocation >= 0)
            memmove(text + underlineLocation, text + underlineLocation + 2, length - underlineLocation - 1); /* -2 for "__" and +1 for NULL byte == -1 */
        }
}

/**************************************************************************/

    /*  returns the index of the color in the table that most closely matches the specified color */
int CVIFUNC FindClosestColorInTable (int color, int *table, int tableSize)
{
    #define     color_sqr(x)    ((x)*(x))
    int         i;
    unsigned    pytDist, rbDist, rgDist, gbDist, dist;
    unsigned    minDist, index;
    unsigned    palRed, palGreen, palBlue;
    unsigned    red, green, blue;
    unsigned    tableColor;

    red     = (color >> 16) & 0xFF;
    green   = (color >> 8) & 0xFF;
    blue    = color & 0xFF;
    
    for (i = 0, minDist = (unsigned)-1; i < tableSize; i++) 
        {
        tableColor = (unsigned)table[i];
        palRed      = (tableColor >> 16) & 0xFF;
        palGreen    = (tableColor >> 8) & 0xFF;
        palBlue     = tableColor & 0xFF;
        
        pytDist = color_sqr(red - palRed) + color_sqr(green - palGreen) + color_sqr(blue - palBlue);
        rgDist = color_sqr(palRed - palGreen - red + green);
        rbDist = color_sqr(palRed - palBlue - red + blue);
        gbDist = color_sqr(palGreen - palBlue - green + blue);
        dist = 3 * pytDist + rgDist + rbDist + gbDist;
        if (minDist == -1 || dist < minDist) 
            {
            index = i;
            if ((minDist = dist) == 0)
                break;
            }
        }

    return index;
}

/**************************************************************************/

    /*  *ctrlList is allocated and filled with the ID's of the controls on the panel.
        The caller should call ListDispose() when done with the list.
        If an error occurs, *ctrlList is set to NULL.
        If the return value is less than zero, then it is an UIL error code.
        Otherwise, the function succeeded.
    */

int CVIFUNC GetCtrlList(int panel, ListType *ctrlList)
{
    int         error = 0;
    int         ctrl;
    ListType    list = 0;
    
    nullChk( list = ListCreate(sizeof(int)));
    errChk( GetPanelAttribute(panel, ATTR_PANEL_FIRST_CTRL, &ctrl));
    
    while (ctrl)
        {
        nullChk( ListInsertItem (list, &ctrl, END_OF_LIST));
        errChk( GetCtrlAttribute(panel, ctrl, ATTR_NEXT_CTRL, &ctrl));
        }       

Error:
    if (error < 0)
        {
        ListDispose(list);
        return error;
        }
    else
        {
        *ctrlList = list;
        return 0;
        }
}

/**************************************************************************/

static int  gTabComparePanel;  /*   must be set before calling CompareCtrlTabOrder */

static int  CVICALLBACK CompareCtrlTabOrder(void *ctrl1, void *ctrl2)
{
    int tab1, tab2;
    
    GetCtrlAttribute(gTabComparePanel, *(int *)ctrl1, ATTR_CTRL_TAB_POSITION, &tab1);
    GetCtrlAttribute(gTabComparePanel, *(int *)ctrl2, ATTR_CTRL_TAB_POSITION, &tab2);
    
    if (tab1 < tab2)
        return -1;
    else
    if (tab1 > tab2)
        return 1;
    else
        return 0;
}

void CVIFUNC SortCtrlListByTabOrder(int panel, ListType ctrlList)
{
    gTabComparePanel = panel;
    ListQuickSort(ctrlList, CompareCtrlTabOrder);
}

/**************************************************************************/

char * CVIFUNC UILEventString(int event)
{
    static char evtBuf[128];
    
    switch (event)
        {
        case EVENT_COMMIT:
            return "EVENT_COMMIT";
        case EVENT_VAL_CHANGED:
            return "EVENT_VAL_CHANGED";
        case EVENT_LEFT_CLICK:
            return "EVENT_LEFT_CLICK";
        case EVENT_LEFT_DOUBLE_CLICK:
            return "EVENT_LEFT_DOUBLE_CLICK";
        case EVENT_RIGHT_CLICK:
            return "EVENT_RIGHT_CLICK";
        case EVENT_RIGHT_DOUBLE_CLICK:
            return "EVENT_RIGHT_DOUBLE_CLICK";
        case EVENT_KEYPRESS:
            return "EVENT_KEYPRESS";
        case EVENT_GOT_FOCUS:
            return "EVENT_GOT_FOCUS";
        case EVENT_LOST_FOCUS:
            return "EVENT_LOST_FOCUS";
        case EVENT_IDLE: 
            return "EVENT_IDLE";
        case EVENT_CLOSE:
            return "EVENT_CLOSE";
        case EVENT_PANEL_SIZE:
            return "EVENT_PANEL_SIZE";
        case EVENT_PANEL_MOVE:
            return "EVENT_PANEL_MOVE";
        case EVENT_END_TASK:
            return "EVENT_END_TASK";
        case EVENT_TIMER_TICK:
            return "EVENT_TIMER_TICK";
        case EVENT_DISCARD:
            return "EVENT_DISCARD";
        default:
            sprintf(evtBuf, "Event %d", event);
            return evtBuf;
        }
}

/**************************************************************************/

    /*  Multiple should be > 0.   Rounds to next multiple AWAY from zero */
int CVIFUNC RoundToNextMultiple(int number, int multiple)
{
    int r;
    
    r = number % multiple;
    if (r == 0)
        return number;
    else
    if (r < 0)
        return number - multiple - r;
    else
        return number + multiple - r;
}


/**************************************************************************/

static char  * CVIFUNC ToolBox_GetDAQErrorString (short int error)
{
    char * errorStr = NULL;
    
    if ((errorStr = ToolBox_InternalGetDAQErrorString (error)) == NULL)
        {
        /* It may be a warning code that doesn't have a description
           so use error description instead.                      */
           
        errorStr = ToolBox_InternalGetDAQErrorString ((-error));        
        }
    
    if (errorStr == NULL)
        errorStr = "No description available. ";
    
    return errorStr;
}

/**************************************************************************/

static char * CVIFUNC ToolBox_InternalGetDAQErrorString (short int error)
{

    switch (error)
    {
        case 0:
            return ("No error. ");
        case cmdSyntaxErr: /* -9000  */
            return("cmdSyntaxErr: An error was detected in the command string; the arrangement or ordering of the characters in the string is not consistent with the expected ordering. ");
        case cmdLimitSyntaxErr: /* -9001  */
            return("cmdLimitSyntaxErr: A syntax error was detected in the limit specification part of the command string.");
        case cmdInterChannelRateSyntaxErr: /* -9002  */
            return("cmdInterChannelRateSyntaxErr: A syntax error was detected in the interchannel sample rate specification part of the command string.");
        case syntaxErr: /* -10001 */
            return("syntaxErr: An error was detected in the input string; the arrangement or ordering of the characters in the string is not consistent with the expected ordering. ");
        case semanticsErr: /* -10002 */
            return("semanticsErr: An error was detected in the input string; the syntax of the string is correct, but certain values specified in the string are inconsistent with other values specified in the string. ");
        case invalidValueErr: /* -10003 */
            return("invalidValueErr: The value of a numeric parameter is invalid. ");
        case valueConflictErr: /* -10004 */
            return("valueConflictErr: The value of a numeric parameter is inconsistent with another parameter, and the combination is therefore invalid. ");
        case badDeviceErr: /* -10005 */
            return("badDeviceErr: The device or task ID parameter is invalid. ");
        case badLineErr: /* -10006 */
            return("badLineErr: The line parameter is invalid. ");
        case badChanErr: /* -10007 */
            return("badChanErr: A channel is out of range for the device type or input configuration, the combination of channels is not allowed, or you must reverse the scan order so that channel 0 is last. ");
        case badGroupErr: /* -10008 */
            return("badGroupErr: The group parameter is invalid. ");
        case badCounterErr: /* -10009 */
            return("badCounterErr: The counter parameter is invalid. ");
        case badCountErr: /* -10010 */
            return("badCountErr: The count parameter is too small or too large for the specified counter. ");
        case badIntervalErr: /* -10011 */
            return("badIntervalErr: The interval parameter is too small or too large for the associated counter or I/O channel. ");
        case badRangeErr: /* -10012 */
            return("badRangeErr: The analog input or analog output voltage range is invalid for the specified channel. ");
        case badRangeWarn: /* 10012 */
            return("badRangeWarn: The analog input or analog output voltage range is invalid for the specified channel. ");
        case badErrorCodeErr: /* -10013 */
            return("badErrorCodeErr: The driver returned an unrecognized or unlisted error code. ");
        case groupTooLargeErr: /* -10014 */
            return("groupTooLargeErr: The group size is too large for the device. ");
        case badTimeLimitErr: /* -10015 */
            return("badTimeLimitErr: The time limit parameter is invalid. ");
        case badReadCountErr: /* -10016 */
            return("badReadCountErr: The read count parameter is invalid. ");
        case badReadModeErr: /* -10017 */
            return("badReadModeErr: The read mode parameter is invalid. ");
        case badReadOffsetErr: /* -10018 */
            return("badReadOffsetErr: The offset is unreachable. ");
        case badClkFrequencyErr: /* -10019 */
            return("badClkFrequencyErr: The frequency parameter is invalid. ");
        case badTimebaseErr: /* -10020 */
            return("badTimebaseErr: The timebase parameter is invalid. ");
        case badLimitsErr: /* -10021 */
            return("badLimitsErr: The limits are beyond the range of the device. ");
        case badWriteCountErr: /* -10022 */
            return("badWriteCountErr: Your data array contains an incomplete update,"
                   " or you are trying to write past the end of the internal buffer,"
                   " or your output operation is continuous and the length of your array"
                   " is not a multiple of one half of the internal buffer size. ");
        case badWriteModeErr: /* -10023 */
            return("badWriteModeErr: The write mode is out of range or is disallowed. ");
        case badWriteOffsetErr: /* -10024 */
            return("badWriteOffsetErr: The write offset plus the write mark is greater than the internal buffer size or it must be set to 0. ");
        case limitsOutOfRangeWarn: /* 10025 */
            return("limitsOutOfRangeWarn: The voltage limits are out of range for this device in the current configuration.\nAlternate limits were selected. ");
        case limitsOutOfRangeErr: /* -10025 */
            return("limitsOutOfRangeErr: The voltage limits are out of range for this device in the current configuration.\nAlternate limits were selected. ");
        case badInputBufferSpecification: /* -10026 */
            return("badInputBufferSpecification: The input buffer specification is invalid.\nThis error results if, for example, you try to configure a multiple-buffer acquisition for a device that does not support multiple-buffer acquisition. ");
        case badDAQEventErr: /* -10027 */
            return("badDAQEventErr: For DAQEvents 0 and 1, general value A must be greater than 0 and less than the internal buffer size.\nIf DMA is used for DAQEvent 1, general value A must divide the internal buffer size evenly,"
                   " with no remainder.\nIf the TIO-10 is used for DAQEvent 4, general value A must be 1 or 2.\nIf the TIO-10 is used for DAQEvent 4, general value A must be 1 or 2. ");
        case badFilterCutoffErr: /* -10028 */
            return("badFilterCutoffErr: The cutoff frequency is not valid for this device. ");
        case obsoleteFunctionErr: /* -10029 */
            return("obsoleteFunctionErr: The function you are calling is no longer supported in this version of the driver. ");
        case badBaudRateErr: /* -10030 */
            return("badBaudRateErr: The specified baud rate for communicating with the serial port is not valid on this platform. ");
        case badChassisIDErr: /* -10031 */
            return("badChassisIDErr: The SCXI chassis specified does not correspond to a configured SCXI chassis. ");
        case badModuleSlotErr: /* -10032 */
            return("badModuleSlotErr: The SCXI module slot that was specified is invalid or corresponds to an empty slot. ");
        case invalidWinHandleErr: /* -10033 */
            return("invalidWinHandleErr: The window handle passed to the function is invalid. ");
        case noSuchMessageErr: /* -10034 */
            return("noSuchMessageErr: No configured message matches the one you tried to delete. ");
        case badGainErr: /* -10080 */
            return("badGainErr: The gain parameter is invalid. ");
        case badPretrigCountErr: /* -10081 */
            return("badPretrigCountErr: The pretrigger sample count is invalid. ");
        case badPosttrigCountErr: /* -10082 */
            return("badPosttrigCountErr: The posttrigger sample count is invalid. ");
        case badTrigModeErr: /* -10083 */
            return("badTrigModeErr: The trigger mode is invalid. ");
        case badTrigCountErr: /* -10084 */
            return("badTrigCountErr: The trigger count is invalid. ");
        case badTrigRangeErr: /* -10085 */
            return("badTrigRangeErr: The trigger range or trigger hysteresis window is invalid. ");
        case badExtRefErr: /* -10086 */
            return("badExtRefErr: The external reference value is invalid. ");
        case badTrigTypeErr: /* -10087 */
            return("badTrigTypeErr: The trigger type parameter is invalid. ");
        case badTrigLevelErr: /* -10088 */
            return("badTrigLevelErr: The trigger level parameter is invalid. ");
        case badTotalCountErr: /* -10089 */
            return("badTotalCountErr: The total count specified is inconsistent with the buffer configuration and pretrigger scan count or with the device type. ");
        case badRPGErr: /* -10090 */
            return("badRPGErr: The individual range, polarity, and gain settings are valid but the combination specified is not allowed for this device. ");
        case badIterationsErr: /* -10091 */
            return("badIterationsErr: The analog output buffer iterations count is not allowed.\nIt must be 0 (for indefinite iterations) or 1. ");
        case lowScanIntervalErr: /* -10092 */
            return("lowScanIntervalErr: Some devices require a time gap between the last sample in a scan and the start of the next scan.\n"
                   "The scan interval you have specified does not provide a large enough gap for the board.\n"
                   "See the SCAN_Start function in the language interface API for an explanation. ");
        case fifoModeErr: /* -10093 */
            return("fifoModeErr: FIFO mode waveform generation cannot be used because at least one condition is not satisifed. ");
        case badPortWidthErr: /* -10100 */
            return("badPortWidthErr: The requested digital port width is not a multiple of the hardware port width. ");
        case gpctrBadApplicationErr: /* -10120 */
            return("gpctrBadApplicationErr: Invalid application used. ");
        case gpctrBadCounterNumberErr: /* -10121 */
            return("gpctrBadCounterNumberErr: Invalid counterNumber used. ");
        case gpctrBadParamValueErr: /* -10122 */
            return("gpctrBadParamValueErr: Invalid paramValue used. ");
        case gpctrBadParamIdErr: /* -10123 */
            return("gpctrBadParamIdErr: Invalid paramID used. ");
        case gpctrBadEntityIdErr: /* -10124 */
            return("gpctrBadEntityIdErr: Invalid entityID used. ");
        case gpctrBadActionErr: /* -10125 */
            return("gpctrBadActionErr: Invalid action used. ");
        case EEPROMreadErr: /* -10200 */
            return("EEPROMreadErr: Unable to read data from EEPROM. ");
        case EEPROMwriteErr: /* -10201 */
            return("EEPROMwriteErr: Unable to write data to EEPROM. ");
        case noDriverErr: /* -10240 */
            return("noDriverErr: The driver interface could not locate or open the driver. ");
        case oldDriverErr: /* -10241 */
            return("oldDriverErr: The driver is out of date. ");
        case functionNotFoundErr: /* -10242 */
            return("functionNotFoundErr: The specified function is not located in the driver. ");
        case configFileErr: /* -10243 */
            return("configFileErr: The driver could not locate or open the configuration file, or the format of the configuration file is not compatible with the currently installed driver. ");
        case deviceInitErr: /* -10244 */
            return("deviceInitErr: The driver encountered a hardware-initialization error while attempting to configure the specified device. ");
        case osInitErr: /* -10245 */
            return("osInitErr: The driver encountered an operating system error while attempting to perform an operation, or the operating system does not support an operation performed by the driver. ");
        case communicationsErr: /* -10246 */
            return("communicationsErr: The driver is unable to communicate with the specified external device. ");
        case cmosConfigErr: /* -10247 */
            return("cmosConfigErr: The CMOS configuration memory for the computer is empty or invalid, or the configuration specified does not agree with the current configuration of the computer. ");
        case dupAddressErr: /* -10248 */
            return("dupAddressErr: The base addresses for two or more devices are the same; consequently, the driver is unable to access the specified device. ");
        case intConfigErr: /* -10249 */
            return("intConfigErr: The interrupt configuration is incorrect given the capabilities of the computer or device. ");
        case dupIntErr: /* -10250 */
            return("dupIntErr: The interrupt levels for two or more devices are the same. ");
        case dmaConfigErr: /* -10251 */
            return("dmaConfigErr: The DMA configuration is incorrect given the capabilities of the computer/DMA controller or device. ");
        case dupDMAErr: /* -10252 */
            return("dupDMAErr: The DMA channels for two or more devices are the same. ");
        case switchlessBoardErr: /* -10253 */
            return("switchlessBoardErr: NI-DAQ was unable to find one or more switchless boards you have configured using WDAQCONF. ");
        case DAQCardConfigErr: /* -10254 */
            return("DAQCardConfigErr: Cannot configure the DAQCard because:\n"
                   "1) The correct version of card and socket services software is not installed.\n"
                   "2) The card in the PCMCIA socket is not a DAQCard.\n"
                   "3) The base address and/or interrupt level requested are not available according to the card and socket services resource manager.\n"
                   "Try different settings or use AutoAssign in the NIDAQ configuration utility. ");
        case remoteChassisDriverInitErr: /* -10255 */
            return("remoteChassisDriverInitErr: There was an error in initializing the driver for Remote SCXI. ");
        case comPortOpenErr: /* -10256 */
            return("comPortOpenErr: There was an error in opening the specified COM port. ");
        case baseAddressErr: /* -10257 */
            return("baseAddressErr: Bad base address specified in the config utility. ");
        case dmaChannel1Err: /* -10258 */
            return("dmaChannel1Err: Bad dma channel 1 specified in the config utility. ");
        case dmaChannel2Err: /* -10259 */
            return("dmaChannel2Err: Bad dma channel 2 specified in the config utility. ");
        case noConnectErr: /* -10340 */
            return("noConnectErr: No RTSI signal/line is connected, or the specified signal and the specified line are not connected. ");
        case badConnectErr: /* -10341 */
            return("badConnectErr: The RTSI signal/line cannot be connected as specified. ");
        case multConnectErr: /* -10342 */
            return("multConnectErr: The specified RTSI signal is already being driven by a RTSI line, or the specified RTSI line is already being driven by a RTSI signal. ");
        case SCXIConfigErr: /* -10343 */
            return("SCXIConfigErr: The specified SCXI configuration parameters are invalid, or the function cannot be executed given the current SCXI configuration. ");
        case chassisNotSynchedErr: /* -10344 */
            return("chassisNotSynchedErr: The Remote SCXI unit is not synchronized with the host. Reset the chassis again to resynchronize it with the host. ");
        case chassisMemAllocErr: /* -10345 */
            return("chassisMemAllocErr: The required amount of memory cannot be allocated on the Remote SCXI unit for the specified operation. ");
        case badPacketErr: /* -10346 */
            return("badPacketErr: The packet received by the Remote SCXI unit is invalid.  Check your serial port cable connections. ");
        case chassisCommunicationErr: /* -10347 */
            return("chassisCommunicationErr: There was an error in sending a packet to the remote chassis.  Check your serial port cable connections. ");
        case waitingForReprogErr: /* -10348 */
            return("waitingForReprogErr: The Remote SCXI unit is in reprogramming mode and is waiting for reprogramming commands from the host (NI-DAQ Configuration Utility). ");
        case SCXIModuleTypeConflictErr: /* -10349 */
            return("SCXIModuleTypeConflictErr: The module ID read from the SCXI module conflicts with the configured module type. ");
        case DSPInitErr: /* -10360 */
            return("DSPInitErr: The DSP driver was unable to load the kernel for its operating system. ");
        case badScanListErr: /* -10370 */
            return("badScanListErr: The scan list is invalid.\nThis error can result if, for example, you mix AMUX-64T channels and onboard channels, or if you scan multiplexed SCXI channels out of order. ");
        case userOwnedRsrcErr: /* -10400 */
            return("userOwnedRsrcErr: The specified resource is owned by the user and cannot be accessed or modified by the driver. ");
        case unknownDeviceErr: /* -10401 */
        case unknownDeviceErrOld: /* -60 */
            return("unknownDeviceErr: The specified device is not a National Instruments product, or the driver does not support the device (for example, the driver was released before the device was supported). ");
        case deviceNotFoundErr: /* -10402 */
            return("deviceNotFoundErr: No device is located in the specified slot or at the specified address. ");
        case deviceSupportErr: /* -10403 */
            return("deviceSupportErr: The specified device does not support the requested action (the driver recognizes the device, but the action is inappropriate for the device). ");
        case noLineAvailErr: /* -10404 */
            return("noLineAvailErr: No line is available. ");
        case noChanAvailErr: /* -10405 */
            return("noChanAvailErr: No channel is available. ");
        case noGroupAvailErr: /* -10406 */
            return("noGroupAvailErr: No group is available. ");
        case lineBusyErr: /* -10407 */
            return("lineBusyErr: The specified line is in use. ");
        case chanBusyErr: /* -10408 */
            return("chanBusyErr: The specified channel is in use. ");
        case groupBusyErr: /* -10409 */
            return("groupBusyErr: The specified group is in use. ");
        case relatedLCGBusyErr: /* -10410 */
            return("relatedLCGBusyErr: A related line, channel, or group is in use; if the driver configures the specified line, channel, or group,"
                   " the configuration, data, or handshaking lines for the related line, channel, or group will be disturbed. ");
        case counterBusyErr: /* -10411 */
            return("counterBusyErr: The specified counter is in use. ");
        case noGroupAssignErr: /* -10412 */
            return("noGroupAssignErr: No group is assigned, or the specified line or channel cannot be assigned to a group. ");
        case groupAssignErr: /* -10413 */
            return("groupAssignErr: A group is already assigned, or the specified line or channel is already assigned to a group. ");
        case reservedPinErr: /* -10414 */
            return("reservedPinErr: Selected signal indicates a pin reserved by NI-DAQ.\nYou cannot configure this pin yourself. ");
        case externalMuxSupportErr: /* -10415 */
            return("externalMuxSupportErr: This function does not support this device when an external multiplexer (such as an AMUX-64T or SCXI) is connected to it. ");
        case DSPDataPathBusyErr: /* -10416 */
            return("DSPDataPathBusyErr: Either DAQ or WFM can use a PC memory buffer, but not both at the same time. ");
        case SCXIModuleNotSupportedErr: /* -10417 */
            return("SCXIModuleNotSupportedErr: At least one of the SCXI modules specified is not supported for the operation. ");
        case sysOwnedRsrcErr: /* -10440 */
            return("sysOwnedRsrcErr: The specified resource is owned by the driver and cannot be accessed or modified by the user. ");
        case memConfigErr: /* -10441 */
            return("memConfigErr: No memory is configured to support the current data transfer mode, or the configured memory does not support the current data transfer mode."
                   "  (If block transfers are in use, the memory must be capable of performing block transfers. ");
        case memDisabledErr: /* -10442 */
            return("memDisabledErr: The specified memory is disabled or is unavailable given the current addressing mode. ");
        case memAlignmentErr: /* -10443 */
            return("memAlignmentErr: The transfer buffer is not aligned properly for the current data transfer mode.\n"
                   "For example, the memory buffer is at an odd address, is not aligned to a 32-bit boundary, is not aligned to a 512-bit boundary, and so on.\n"
                   "Alternatively, the driver is unable to align the buffer because the buffer is too small. ");
        case memFullErr: /* -10444 */
            return("memFullErr: No more system memory is available on the heap;\n or no more memory is available on the device;\n or part of the DMA DAQ buffer is in an invalid DMA region (above 16M on AT machine)");
        case memLockErr: /* -10445 */
            return("memLockErr: The transfer buffer cannot be locked into physical memory. ");
        case memPageErr: /* -10446 */
            return("memPageErr: The transfer buffer contains a page break; system resources may require reprogramming when the page break is encountered. ");
        case memPageLockErr: /* -10447 */
            return("memPageLockErr: The operating environment is unable to grant a page lock. ");
        case stackMemErr: /* -10448 */
            return("stackMemErr: The driver is unable to continue parsing a string input due to stack limitations. ");
        case cacheMemErr: /* -10449 */
            return("cacheMemErr: A cache-related error occurred, or caching is not supported in the current mode. ");
        case physicalMemErr: /* -10450 */
            return("physicalMemErr: A hardware error occurred in physical memory, or no memory is located at the specified address. ");
        case virtualMemErr: /* -10451 */
            return("virtualMemErr: The driver is unable to make the transfer buffer contiguous in virtual memory and therefore cannot lock the buffer into physical memory; thus, you cannot use the buffer for DMA transfers. ");
        case noIntAvailErr: /* -10452 */
            return("noIntAvailErr: No interrupt level is available for use. ");
        case intInUseErr: /* -10453 */
            return("intInUseErr: The specified interrupt level is already in use by another device. ");
        case noDMACErr: /* -10454 */
            return("noDMACErr: No DMA controller is available in the system. ");
        case noDMAAvailErr: /* -10455 */
            return("noDMAAvailErr: No DMA channel is available for use. ");
        case DMAInUseErr: /* -10456 */
            return("DMAInUseErr: The specified DMA channel is already in use by another device. ");
        case badDMAGroupErr: /* -10457 */
            return("badDMAGroupErr: DMA cannot be configured for the specified group because it is too small, too large, or misaligned.\nConsult the user manual for the device in question to determine group ramifications with respect to DMA. ");
        case diskFullErr: /* -10458 */
            return("diskFullErr: The storage disk you specified is full. ");
        case DLLInterfaceErr: /* -10459 */
            return("DLLInterfaceErr: The DLL could not be called due to an interface error. ");
        case interfaceInteractionErr: /* -10460 */
            return("interfaceInteractionErr: You have attempted to mix LabVIEW 2.2 VIs and LabVIEW 3.0 VIs.\n"
                   "You may run an application consisting only of 2.2 VIs, then run the 2.2 Board Reset VI, before you can run any 3.0 VIs.\n"
                   "You may run an application consisting of only 3.0 VIs, then run the 3.0 Device Reset VI, before you can run any 2.2 VIs. ");
        case muxMemFullErr: /* -10480 */
            return("muxMemFullErr: The scan list is too large to fit into the mux-gain memory of the board. ");
        case bufferNotInterleavedErr: /* -10481 */
            return("bufferNotInterleavedErr: You cannot use DMA to transfer data from two buffers. You may be able to use interrupts. ");
        case invalidDSPhandleErr: /* -10560 */
            return("invalidDSPhandleErr: The DSP handle input to the VI is not a valid handle. ");
        case noSetupErr: /* -10600 */
            return("noSetupErr: No setup operation has been performed for the specified resources. ");
        case multSetupErr: /* -10601 */
            return("multSetupErr: The specified resources have already been configured by a setup operation. ");
        case noWriteErr: /* -10602 */
            return("noWriteErr: No output data has been written into the transfer buffer. ");
        case groupWriteErr: /* -10603 */
            return("groupWriteErr: The output data associated with a group must be for a single channel or must be for consecutive channels. ");
        case activeWriteErr: /* -10604 */
            return("activeWriteErr: Once data generation has started, only the transfer buffers originally written to can be updated.\n"
                   "If DMA is active and a single transfer buffer contains interleaved channel data, new data must be provided for all output channels currently using the DMA channel. ");
        case endWriteErr: /* -10605 */
            return("endWriteErr: No data was written to the transfer buffer because the final data block has already been loaded. ");
        case notArmedErr: /* -10606 */
            return("notArmedErr: The specified resource is not armed. ");
        case armedErr: /* -10607 */
            return("armedErr: The specified resource is already armed. ");
        case noTransferInProgWarn: /* 10608 */
            return("noTransferInProgWarn: No transfer is in progress for the specified resource. ");
        case noTransferInProgErr: /* -10608 */
            return("noTransferInProgErr: No transfer is in progress for the specified resource. ");
        case transferInProgErr: /* -10609 */
            return("transferInProgErr: A transfer is already in progress for the specified resource. ");
        case transferPauseErr: /* -10610 */
            return("transferPauseErr: A single output channel in a group cannot be paused if the output data for the group is interleaved. ");
        case badDirOnSomeLinesErr: /* -10611 */
            return("badDirOnSomeLinesErr: Some of the lines in the specified channel are not configured for the transfer direction specified.\n"
                   "For a write transfer, some lines were configured for input.\n"
                   "For a read transfer, some lines were configured for output. ");
        case badLineDirErr: /* -10612 */
            return("badLineDirErr: The specified line does not support the specified transfer direction. ");
        case badChanDirErr: /* -10613 */
            return("badChanDirErr: The specified channel does not support the specified transfer direction. ");
        case badGroupDirErr: /* -10614 */
            return("badGroupDirErr: The specified group does not support the specified transfer direction. ");
        case masterClkErr: /* -10615 */
            return("masterClkErr: The clock configuration for the clock master is invalid. ");
        case slaveClkErr: /* -10616 */
            return("slaveClkErr: The clock configuration for the clock slave is invalid. ");
        case noClkSrcErr: /* -10617 */
            return("noClkSrcErr: No source signal has been assigned to the clock resource. ");
        case badClkSrcErr: /* -10618 */
            return("badClkSrcErr: The specified source signal cannot be assigned to the clock resource. ");
        case multClkSrcErr: /* -10619 */
            return("multClkSrcErr: A source signal has already been assigned to the clock resource. ");
        case noTrigErr: /* -10620 */
            return("noTrigErr: No trigger signal has been assigned to the trigger resource. ");
        case badTrigErr: /* -10621 */
            return("badTrigErr: The specified trigger signal cannot be assigned to the trigger resource. ");
        case preTrigErr: /* -10622 */
            return("preTrigErr: The pretrigger mode is not supported or is not available in the current configuration, or no pretrigger source has been assigned. ");
        case postTrigErr: /* -10623 */
            return("postTrigErr: No posttrigger source has been assigned. ");
        case delayTrigErr: /* -10624 */
            return("delayTrigErr: The delayed trigger mode is not supported or is not available in the current configuration, or no delay source has been assigned. ");
        case masterTrigErr: /* -10625 */
            return("masterTrigErr: The trigger configuration for the trigger master is invalid. ");
        case slaveTrigErr: /* -10626 */
            return("slaveTrigErr: The trigger configuration for the trigger slave is invalid. ");
        case noTrigDrvErr: /* -10627 */
            return("noTrigDrvErr: No signal has been assigned to the trigger resource. ");
        case multTrigDrvErr: /* -10628 */
            return("multTrigDrvErr: A signal has already been assigned to the trigger resource. ");
        case invalidOpModeErr: /* -10629 */
            return("invalidOpModeErr: The specified operating mode is invalid, or the resources have not been configured for the specified operating mode. ");
        case invalidReadErr: /* -10630 */
            return("invalidReadErr: An attempt was made to read 0 bytes from the transfer buffer, or an attempt was made to read past the end of the transfer buffer. ");
        case noInfiniteModeErr: /* -10631 */
            return("noInfiniteModeErr: Continuous input or output transfers are not allowed in the current operating mode. ");
        case someInputsIgnoredErr: /* -10632 */
            return("someInputsIgnoredErr: Certain inputs were ignored because they are not relevant in the current operating mode. ");
        case invalidRegenModeErr: /* -10633 */
            return("invalidRegenModeErr: This device does not support the specified analog output regeneration mode. ");
        case noContTransferInProgErr: /* -10634 */
            return("noContTransferInProgErr: No continuous (double buffered) transfer is in progress for the specified resource. ");
        case invalidSCXIOpModeErr: /* -10635 */
            return("invalidSCXIOpModeErr:  Either the SCXI operating mode specified in a configuration call is invalid, or a module is in the wrong operating mode to execute the function call.");
        case noContWithSynchErr: /* -10636 */
            return("noContWithSynchErr: You cannot start a continuous (double buffered) operation with a synchronous function call. ");
        case bufferAlreadyConfigErr: /* -10637 */
            return("bufferAlreadyConfigErr: Attempt to configure a buffer after a buffer has already been configured. You can configure a buffer only once. ");
        case badChanGainErr: /* -10680 */
            return("badChanGainErr: All channels must have an identical setting for this device. ");
        case badChanRangeErr: /* -10681 */
            return("badChanRangeErr: All channels of this device must have the same range. ");
        case badChanPolarityErr: /* -10682 */
            return("badChanPolarityErr: All channels of this device must have the same polarity. ");
        case badChanCouplingErr: /* -10683 */
            return("badChanCouplingErr: All channels of this device must have the same coupling. ");
        case badChanInputModeErr: /* -10684 */
            return("badChanInputModeErr: All channels of this device must have the same input range. ");
        case clkExceedsBrdsMaxConvRate: /* -10685 */
            return("clkExceedsBrdsMaxConvRate: The clock rate selected exceeds the recommended maximum rate for this device. ");
        case scanListInvalidErr: /* -10686 */
            return("scanListInvalidErr: A configuration change has invalidated the scan list. ");
        case bufferInvalidErr: /* -10687 */
            return("bufferInvalidErr: A configuration change has invalidated the allocated buffer. ");
        case noTrigEnabledErr: /* -10688 */
            return("noTrigEnabledErr: The total number of scans and pretrigger scans implies that a trigger start is intended, but no trigger is enabled. ");
        case digitalTrigBErr: /* -10689 */
            return("digitalTrigBErr: Digital trigger B is illegal for the total scans and pretrigger scans specified. ");
        case digitalTrigAandBErr: /* -10690 */
            return("digitalTrigAandBErr: This device does not allow digital triggers A and B to be enabled at the same time. ");
        case extConvRestrictionErr: /* -10691 */
            return("extConvRestrictionErr: This device does not allow an external sample clock with an external scan clock, start trigger, or stop trigger. ");
        case chanClockDisabledErr: /* -10692 */
            return("chanClockDisabledErr: Cannot start the acquisition because the channel clock is disabled. ");
        case extScanClockErr: /* -10693 */
            return("extScanClockErr: Cannot use an external scan clock when performing a single scan of a single channel. ");
        case unsafeSamplingFreqErr: /* -10694 */
            return("unsafeSamplingFreqErr: The sampling frequency exceeds the safe maximum rate for the ADC, gains, and filters you are using. ");
        case DMAnotAllowedErr: /* -10695 */
            return("DMAnotAllowedErr: You must use interrupts.\nDMA is not allowed. ");
        case multiRateModeErr: /* -10696 */
            return("multiRateModeErr: Multi-rate scanning can not be used with AMUX-64, SCXI, or pre-triggered acquisitions. ");
        case rateNotSupportedErr: /* -10697 */
            return("rateNotSupportedErr: NI-DAQ was unable to convert your timebase/interval pair to match the actual hardware capabilities of the specified board. ");
        case timebaseConflictErr: /* -10698 */
            return("timebaseConflictErr: You cannot use this combination of scan and sample clock timebases for the specified board. ");
        case polarityConflictErr: /* -10699 */
            return("polarityConflictErr: You cannot use this combination of scan and sample clock source polarities for this operation, for the specified board. ");
        case signalConflictErr: /* -10700 */
            return("signalConflictErr: You cannot use this combination of scan and convert clock signal sources for this operation, for the specified board. ");
        case noLaterUpdateErr: /* -10701 */
            return("noLaterUpdateErr: The call had no effect because the specified channel had not been set for later internal update. ");
        case prePostTriggerErr: /* -10702 */
            return("prePostTriggerErr: Pretriggering and posttriggering cannot be used simultaneously on the Lab-PC+ or 1200 series boards. ");
        case noHandshakeModeErr: /* -10710 */
            return("noHandshakeModeErr: The specified port has not been configured for handshaking. ");
        case noEventCtrErr: /* -10720 */
            return("noEventCtrErr: The specified counter is not configured for event-counting operation. ");
        case SCXITrackHoldErr: /* -10740 */
            return("SCXITrackHoldErr: A signal has already been assigned to the SCXI track-and-hold trigger line, or a control call was inappropriate because the specified module is not configured for one-channel operation. ");
        case sc2040InputModeErr: /* -10780 */
            return("sc2040InputModeErr: When you have an SC2040 attached to your device, all analog input channels must be configured for differential input mode. ");
        case outputTypeMustBeVoltageErr: /* -10781 */
            return("outputTypeMustBeVoltageErr: The polarity of the output channel cannot be bipolar when outputting currents. ");
        case sc2040HoldModeErr: /* -10782 */
            return("sc2040HoldModeErr: The specified operation cannot be performed with the SC-2040 configured in hold mode. ");
        case calConstPolarityConflictErr: /* -10783 */
            return("calConstPolarityConflictErr: Calibration constants in the load area have different polarity than current configuration. Loading constants from factory. ");
        case timeOutErr: /* -10800 */
            return("timeOutErr: The operation could not complete within the time limit. ");
        case calibrationErr: /* -10801 */
            return("calibrationErr: An error occurred during the calibration process. ");
        case dataNotAvailErr: /* -10802 */
            return("dataNotAvailErr: The requested amount of data has not yet been acquired, or the acquisition has completed and no more data is available to read. ");
        case transferStoppedErr: /* -10803 */
            return("transferStoppedErr: The transfer has been stopped to prevent regeneration of output data. ");
        case earlyStopErr: /* -10804 */
            return("earlyStopErr: The transfer stopped prior to reaching the end of the transfer buffer. ");
        case overRunErr: /* -10805 */
            return("overRunErr: The clock source for the input transfer is faster than the maximum input-clock rate; the integrity of the data has been compromised.\n"
                   "Alternatively, the clock source for the output transfer is faster than the maximum output-clock rate; a data point was generated more than once since the update occurred before new data was available. ");
        case noTrigFoundErr: /* -10806 */
            return("noTrigFoundErr: No trigger value was found in the input transfer buffer. ");
        case earlyTrigErr: /* -10807 */
            return("earlyTrigErr: The trigger occurred before sufficient pretrigger data was acquired. ");
        case LPTCommunicationErr: /* -10808 */
            return("LPTCommunicationErr: An error occurred in the parallel port communication with the SCXI-1200. ");
        case gateSignalErr: /* -10809 */
            return("gateSignalErr: Attempted to start a pulse width measurement with the pulse in the active state. ");
        case internalDriverErr: /* -10810 */
            return("internalDriverlErr: An unexpected error occurred inside the driver when performing this operation. ");
        case internalKernelErr: /* -10811 */
            return("internalKernelErr: An unexpected error occurred inside the device kernel when performing this operation. ");
        case softwareErr: /* -10840 */
            return("softwareErr: The contents or the location of the driver file was changed between accesses to the driver. ");
        case firmwareErr: /* -10841 */
            return("firmwareErr: The firmware does not support the specified operation, or the firmware operation could not complete due to a data-integrity problem. ");
        case hardwareErr: /* -10842 */
            return("hardwareErr: The hardware is not responding to the specified operation, or the response from the hardware is not consistent with the functionality of the hardware. ");
        case underFlowErr: /* -10843 */
            return("underFlowErr: The update rate exceeds your system's capacity to supply data to the output channel. ");
        case underWriteErr: /* -10844 */
            return("underWriteErr: At the time of the update for the device-resident memory, insufficient data was present in the output transfer buffer to complete the update. ");
        case overFlowErr: /* -10845 */
            return("overFlowErr: At the time of the update clock for the input channel, the device-resident memory was unable to accept additional data - one or more data points may have been lost. ");
        case overWriteErr: /* -10846 */
            return("overWriteErr: New data was written into the input transfer buffer before the old data was retrieved. ");
        case dmaChainingErr: /* -10847 */
            return("dmaChainingErr: New buffer information was not available at the time of the DMA chaining interrupt; DMA transfers will terminate at the end of the currently active transfer buffer. ");
        case noDMACountAvailErr: /* -10848 */
            return("noDMACountAvailErr: The driver could not obtain a valid reading from the transfer-count register in the DMA controller. ");
        case openFileErr: /* -10849 */
            return("openFileErr: Unable to open a file. ");
        case closeFileErr: /* -10850 */
            return("closeFileErr: Unable to close a file. ");
        case fileSeekErr: /* -10851 */
            return("fileSeekErr: Unable to seek within a file. ");
        case readFileErr: /* -10852 */
            return("readFileErr: Unable to read from a file. ");
        case writeFileErr: /* -10853 */
            return("writeFileErr: Unable to write to a file. ");
        case miscFileErr: /* -10854 */
            return("miscFileErr: An error occurred accessing a file. ");
        case osUnsupportedErr: /* -10855 */
            return("osUnsupportedErr: NI-DAQ does not support the current operation on this particular version of the operating system. ");
        case osErr: /* -10856 */
            return("osErr: An unexpected error occurred from the operating system while performing the given operation. ");
        case updateRateChangeErr: /* -10880 */
            return("updateRateChangeErr: A change to the update rate is not possible at this time because:\n"
                   "1) When waveform generation is in progress, you cannot change the interval timebase.\n"
                   "2) When you make several changes in a row, you must give each change enough time to take effect before requesting further changes. ");
        case partialTransferCompleteErr: /* -10881 */
            return("partialTransferCompleteErr: You cannot do another transfer after a successful partial transfer. ");
        case daqPollDataLossErr: /* -10882 */
            return("daqPollDataLossErr: The data collected on the Remote SCXI unit was overwritten before it could be transferred to the buffer in the host.\n"
                   "Try using a slower data acquisition rate if possible. ");
        case wfmPollDataLossErr: /* -10883 */
            return("wfmPollDataLossErr: New data could not be transferred to the waveform buffer of the Remote SCXI unit to keep up with the waveform update rate.\n"
                   "Try using a slower waveform update rate if possible. ");
        case pretrigReorderErr: /* -10884 */
            return("pretrigReorderErr: Could not rearrange data after a pretrigger acquisition completed. ");
        case gpctrDataLossErr: /* -10920 */
            return("gpctrDataLossErr: One or more data points may have been lost during buffered GPCTR operations due to speed limitations of your system. ");
        case chassisResponseTimeoutErr: /* -10940 */
            return("chassisResponseTimeoutErr: No response was received from the Remote SCXI unit within the specified time limit. ");
        case reprogrammingFailedErr: /* -10941 */
            return("reprogrammingFailedErr: Reprogramming the Remote SCXI unit was unsuccessful.  Please try again. ");
        case invalidResetSignatureErr: /* -10942 */
            return("invalidResetSignatureErr: An invalid reset signature was sent from the host to the Remote SCXI unit. ");
    }

    return NULL;
}

/***************************************************************************/

    /* ConvertDataToText converts raw data into printable ASCII text.  It       */
    /* takes 3 bytes of data at a time and converts it to 4 bytes of text in    */
    /* the following manner.  The two highest order bits of each data byte are  */
    /* written into the lowest order six bits of the first text byte.  The      */
    /* remaining 6 bits of each data byte are written into the lowest order     */
    /* six bits of the remaining 3 text bytes.  This gives 4 text bytes with    */
    /* values from 0-63.  By adding '0' to these values, we have values within  */
    /* the range of the printable ASCII character set.                          */
    /* returns 0 if out of memory, 1 otherwise                                  */

#define CONVERSION_UNIT_MASK (0x03)
#define ASCII_OFFSET '0'      /* start at '0' to avoid '"' (a quote would be misinterpereted as a delimiter at the start or end of an .ini file value) */

int CVIFUNC ConvertDataToText(const unsigned char *dataPtr, long dataSize, unsigned char **textPtr)
{
    int             error = 0;
    long            textSize, numTriples, remainder, i, dataIndex, textIndex;
    unsigned char   temp;
    unsigned char   *textArray;

    numTriples = dataSize / 3;
    remainder = dataSize % 3;

    /* each 3 bytes of data will become 4 bytes of text */
    textSize = 4 * numTriples;

    /* The remainder can be 0, 1, or 2.  If it is 0, no more space is needed. */
    /* If the remainder is 1, then 2 additional bytes will be needed.  If the */
    /* remainder is 2, then 3 additional bytes are needed.                    */
    if(remainder)
        textSize += remainder + 1;

    nullChk(*textPtr = (unsigned char *)malloc(textSize + 1));  /* +1 for NULL */
    textArray = *textPtr;
    for(i = 0; i < numTriples; i++) 
        {
        dataIndex = 3 * i;
        textIndex = 4 * i;
        temp = ((dataPtr[dataIndex] & (CONVERSION_UNIT_MASK << 6)) >> 2) |
                ((dataPtr[dataIndex + 1] & (CONVERSION_UNIT_MASK << 6)) >> 4) |
                ((dataPtr[dataIndex + 2] & (CONVERSION_UNIT_MASK << 6)) >> 6);
        textArray[textIndex] = temp + ASCII_OFFSET;
        temp = (dataPtr[dataIndex] & ~(CONVERSION_UNIT_MASK << 6));
        textArray[textIndex + 1] = temp + ASCII_OFFSET;
        temp = (dataPtr[dataIndex + 1] & ~(CONVERSION_UNIT_MASK << 6));
        textArray[textIndex + 2] = temp + ASCII_OFFSET;
        temp = (dataPtr[dataIndex + 2] & ~(CONVERSION_UNIT_MASK << 6));
        textArray[textIndex + 3] = temp + ASCII_OFFSET;
        }
        
    switch(remainder) 
        {
        case 0:
            break;
        case 1:
            dataIndex = 3 * numTriples;
            textIndex = 4 * numTriples;
            temp = (dataPtr[dataIndex] & (CONVERSION_UNIT_MASK << 6)) >> 2;
            textArray[textIndex] = temp + ASCII_OFFSET;
            temp = dataPtr[dataIndex] & ~(CONVERSION_UNIT_MASK << 6);
            textArray[textIndex + 1] =  temp + ASCII_OFFSET;
            break;
        case 2:
            dataIndex = 3 * numTriples;
            textIndex = 4 * numTriples;
            temp = ((dataPtr[dataIndex] & (CONVERSION_UNIT_MASK << 6)) >> 2) |
                    ((dataPtr[dataIndex + 1] & (CONVERSION_UNIT_MASK << 6)) >> 4);
            textArray[textIndex] = temp + ASCII_OFFSET;
            temp = dataPtr[dataIndex] & ~(CONVERSION_UNIT_MASK << 6);
            textArray[textIndex + 1] = temp + ASCII_OFFSET;
            temp = dataPtr[dataIndex + 1] & ~(CONVERSION_UNIT_MASK << 6);
            textArray[textIndex + 2] = temp + ASCII_OFFSET;
            break;
        default:
            Assert(FALSE);   /* Shouldn't have this value for remainder!!! */
            break;
        }
    textArray[textSize] = NULL;

Error:
    if(error < 0) 
        {
        free(*textPtr);
        *textPtr = NULL;
        }
        
    return error >= 0;
}

/***************************************************************************/

    /* ConvertTextToData converts text obtained from ConvertDataToText back     */
    /* into data.  See ConvertDataToText for more information.                  */
    /* returns 0 if out of memory, 1 otherwise                                  */

int CVIFUNC ConvertTextToData(const unsigned char *textPtr, long *dataSize, unsigned char **dataPtr)
{
    int             error = 0;
    long            size, numQuads, remainder, i, dataIndex, textIndex, textSize;
    unsigned char   temp;
    unsigned char   *dataArray;

    textSize = strlen((char *)textPtr);
    numQuads = textSize / 4;
    remainder = textSize % 4;

    /* each 4 bytes of text will become 3 bytes of data */
    size = numQuads * 3;

    /* The remainder can be 0, 2, or 3.  It cannot be 1, since Convert-      */
    /*  DataToText never leaves a remainder of 1.  If it is 0, no more space     */
    /*  is needed.  If the remainder is 2, 1 byte will be needed.  If it's 3, */
    /* 2 are needed.                                                         */
    if(remainder)
        size += remainder - 1;

    if (size == 0)
        goto Error;
        
    nullChk(*dataPtr = (unsigned char *)malloc(size)); 
    dataArray = *dataPtr;
    for(i = 0; i < numQuads; i++) 
        {
        textIndex = 4 * i;
        dataIndex = 3 * i;
        temp = textPtr[textIndex] - ASCII_OFFSET;
        dataArray[dataIndex] = ((temp & (CONVERSION_UNIT_MASK << 4)) << 2) | (textPtr[textIndex + 1] - ASCII_OFFSET);
        dataArray[dataIndex + 1] = ((temp & (CONVERSION_UNIT_MASK << 2)) << 4) | (textPtr[textIndex + 2] - ASCII_OFFSET);
        dataArray[dataIndex + 2] = ((temp & CONVERSION_UNIT_MASK) << 6) | (textPtr[textIndex + 3] - ASCII_OFFSET);
        }
        
    switch(remainder) 
        {
        case 0:
            break;
        case 2:
            textIndex = 4 * numQuads;
            dataIndex = 3 * numQuads;
            temp = textPtr[textIndex] - ASCII_OFFSET;
            dataArray[dataIndex] = ((temp & (CONVERSION_UNIT_MASK << 4)) << 2) | (textPtr[textIndex + 1] - ASCII_OFFSET);
            break;
        case 3:
            textIndex = 4 * numQuads;
            dataIndex = 3 * numQuads;
            temp = textPtr[textIndex] - ASCII_OFFSET;
            dataArray[dataIndex] = ((temp & (CONVERSION_UNIT_MASK << 4)) << 2) | (textPtr[textIndex + 1] - ASCII_OFFSET);
            dataArray[dataIndex + 1] = ((temp & (CONVERSION_UNIT_MASK << 2)) << 4) | (textPtr[textIndex + 2] - ASCII_OFFSET);
            break;
        default:
            Assert(FALSE);   /* Shouldn't have this value for remainder!!! */
            break;
        }

Error:
    if (dataSize)
        *dataSize = size;
        
    if(error < 0) 
        {
        free(*dataPtr);
        *dataPtr = NULL;
        }
        
    return error >= 0;
}


